# ByFieldSlopeStatistics------------------------------------------------------------------------------------------------------------
#
# Calculates slope in % rise using an input DEM 
# Generates by-field slope statistics using the output slope raster and field boundary feature class
# Required inputs include a field boundary feature class and an input DEM
#
# An output table (slopetable +  inHUC) is created with the following attributes:
#
# MeanSlope: mean slope (% rise) of each field
# 3rd quartile slope: 75th percentile slope (% rise)
# Pct_lt1: PctSlopeClass1 = < 1%
# Pct1_2: PctSlopeClass2 = 1 - 2%
# Pct2_5: PctSlopeClass3 = 2 - 5%
# Pct5_10: PctSlopeClass4 = 5 - 10%
# Pct10_15: PctSlopeClass5 = 10 - 15%
# Pct_gt15: PctSlopeClass6 = > 15%
#
# The slope table can be joined to the field boundary feautre class via the FBndID
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
arcpy.CheckOutExtension("Spatial")


## Modules------------------------------------------------------------------------


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def CreateSlopeTable(FBIn, DEM):
    
    # Calculate slope
    arcpy.AddMessage("Calculating slope in % rise")
    OutSlope = Slope(DEM, "PERCENT_RISE", ZFactor)
    OutSlope.save(OutSlopeRaster)

    # Build Pyramids
    arcpy.BuildPyramids_management(OutSlope)

    # Create slope table ---> output of zonal statistics using FBIn as zones and OutSlope as underlying raster
    SlopeTable = ZonalStatisticsAsTable(FBIn, "FBndID", OutSlope, OutSlopeTable, "#", "MEAN")

    # Add field to slope table
    arcpy.AddField_management(SlopeTable, "MeanSlope", "FLOAT")

    # Calculate mean slope - using value from zonal statistics output
    arcpy.CalculateField_management(SlopeTable, "MeanSlope", "!Mean!", "PYTHON")
    arcpy.DeleteField_management(SlopeTable, ["Mean"])

    return(OutSlope, SlopeTable)


def ThirdQuartileSlope(FBIn, OutSlope, SlopeTable):
    arcpy.AddMessage("Calculating by-field third quartile slope values")
    
    # To find "3rd quartile" slope, the underlying slope raster must be integer ---> Times OutSlope by 100 and convert to integer
    SlopeInt = Int(Times(OutSlope, 100))

    # Zonal statistics to find median slope in each field
    MedianSlope = ZonalStatistics(FBIn, "FBndID", SlopeInt, "MEDIAN")

    # Subtract SlopeInt from MedianSlope
    SlopeMinusMedian = Minus(SlopeInt, MedianSlope)

    # Convert values less than 0 to "NODATA" ---> what remain is values > 50%ile
    AboveMedianSlope = Con(SlopeMinusMedian, SlopeInt, "", "VALUE > 0")

    # Repeat zonal statistics to find median value of > 50%ile values ---> will give 3rd quartile slope in each field
    SlopeStats = ZonalStatisticsAsTable(FBIn, "FBndID", AboveMedianSlope, "in_memory" + "\\SlopeStats", "#", "MEDIAN")

    # Join to SlopeTable and divide by 100 to convert slope from integer back to percent rise 
    arcpy.JoinField_management(SlopeTable, "FBndID", SlopeStats, "FBndID", ["MEDIAN"])
    arcpy.AddField_management(SlopeTable, "Slope75Pct", "FLOAT")
    arcpy.CalculateField_management(SlopeTable, "Slope75Pct", ("!MEDIAN! / 100.00"), "PYTHON")
    arcpy.DeleteField_management(SlopeTable, ["Median"])

    arcpy.Delete_management(SlopeInt)
    arcpy.Delete_management(MedianSlope)
    arcpy.Delete_management(SlopeMinusMedian)
    arcpy.Delete_management(AboveMedianSlope)
    arcpy.Delete_management(SlopeStats)
    del[SlopeInt, MedianSlope, SlopeMinusMedian, AboveMedianSlope, SlopeStats]


def SlopeClassZonalStats(FBIn, OutSlope, SlopeTable):
    # Reclassify slope raster into values that represent slope ranges (i.e. <1% = 1, 1 - 2% = 2, etc.)
    SlopeMax = arcpy.GetRasterProperties_management(OutSlope, "maximum")
    SlopeRec = Reclassify(OutSlope, "Value",
                            RemapRange([[0, 1, "1"], 
                            [1.000000001, 2, "2"], 
                            [2.000000001, 5, "3"], 
                            [5.000000001, 10, "4"], 
                            [10.000000001, 15, "5"], 
                            [15.000000001, SlopeMax, "6"]]))

    # Create a SlopeClassList that lists these values (1 - 6)
    arcpy.AddMessage("Slope class 1: < 1% slope")
    arcpy.AddMessage("Slope class 2: 1 - 2% slope")
    arcpy.AddMessage("Slope class 3: 2 - 5% slope")
    arcpy.AddMessage("Slope class 4: 5 - 10% slope")
    arcpy.AddMessage("Slope class 5: 10 - 15% slope")
    arcpy.AddMessage("Slope class 6: > 15% slope")
    SlopeClassList = [1, 2, 3, 4, 5, 6]

    # For each value in list: 1) Make a "class_ras" raster containing just pixels in that slope class, then 2) Perform zonal statistics with FBIn
    for SlopeClass in SlopeClassList:
        
        arcpy.MakeRasterLayer_management(SlopeRec, "class_tmp", "\"VALUE\" = %s" % SlopeClass)
        class_ras = arcpy.CopyRaster_management("class_tmp", "class_ras")

        arcpy.AddMessage("Finding percent of field in slope class %s" % SlopeClass)
        ClassStats = ZonalStatisticsAsTable(FBIn, "FBndID", class_ras, "in_memory" + "\\ClassStats", "", "SUM")
        arcpy.JoinField_management(SlopeTable, "FBndID", ClassStats, "FBndID", ["COUNT"])

        # Add and populate a class count; equals the # of pixels of each slope class in each field 
        arcpy.AddField_management(SlopeTable, "ClassCount", "FLOAT")
        arcpy.CalculateField_management(SlopeTable, "ClassCount", "!COUNT_1!", "PYTHON")

        # Find % of each field occupied by each slope class: ("CLASSCOUNT" / "COUNT") * 100
        arcpy.AddField_management(SlopeTable, "PctSlope%s" % SlopeClass, "FLOAT")
        arcpy.CalculateField_management(SlopeTable, "PctSlope%s" % SlopeClass, "(!ClassCount!/!Count!) * 100", "PYTHON")
        arcpy.DeleteField_management(SlopeTable, ["COUNT_1", "ClassCount"])

        ## Convert null values to 0 
        arcpy.MakeTableView_management(SlopeTable, "NullValues")
        arcpy.SelectLayerByAttribute_management("NullValues", "NEW_SELECTION", "\"PctSlope%s\" IS NULL" % SlopeClass)
        arcpy.CalculateField_management("NullValues", "PctSlope%s" % SlopeClass, 0, "PYTHON")

    # Relabel fields in SlopeTable
    arcpy.AddField_management(SlopeTable, "Pct_lt1", "FLOAT")
    arcpy.AddField_management(SlopeTable, "Pct1_2", "FLOAT")
    arcpy.AddField_management(SlopeTable, "Pct2_5", "FLOAT")
    arcpy.AddField_management(SlopeTable, "Pct5_10", "FLOAT")
    arcpy.AddField_management(SlopeTable, "Pct10_15", "FLOAT")
    arcpy.AddField_management(SlopeTable, "Pct_gt15", "FLOAT")

    with arcpy.da.UpdateCursor(SlopeTable, ["PctSlope1", "PctSlope2", \
                                            "PctSlope3", "PctSlope4", \
                                            "PctSlope5", "PctSlope6", \
                                            "Pct_lt1", "Pct1_2", \
                                            "Pct2_5", "Pct5_10", \
                                            "Pct10_15", "Pct_gt15"]) as cursor:
        for row in cursor:
            row[6] = row[0]
            row[7] = row[1]
            row[8] = row[2]
            row[9] = row[3]
            row[10] = row[4]
            row[11] = row[5]
            cursor.updateRow(row)

    arcpy.DeleteField_management(SlopeTable, ["COUNT", "AREA", "ZONE_CODE", "PctSlope1", "PctSlope2", "PctSlope3", "PctSlope4", "PctSlope5", "PctSlope6"])
    
    # cleanup on aisle 3
    arcpy.Delete_management(class_ras)
    arcpy.Delete_management(SlopeRec)
    del[class_ras, SlopeRec]
    arcpy.Delete_management("in_memory")

  
##------------------------------------------------------------------------------
##------------------------------------------------------------------------------
if __name__ == "__main__":
    
    # Define Parameters
    FBIn = arcpy.GetParameterAsText(0)
    # Test for a join on FBIn
    basename = arcpy.Describe(FBIn).baseName
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        basename = arcpy.Describe(FBIn).baseName
        arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    DEM = arcpy.GetParameterAsText(1)
    ZFactor = arcpy.GetParameterAsText(2)
    OutSlopeTable = arcpy.GetParameterAsText(3)
    OutSlopeRaster = arcpy.GetParameterAsText(4)

    # Set extent, snap, workspace
    arcpy.env.extent = DEM
    arcpy.env.snapRaster = DEM
    rDesc = arcpy.Describe(DEM)
    arcpy.env.cellSize = rDesc.meanCellHeight
    env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder
    
    # Run Modules
    OutSlope, SlopeTable = CreateSlopeTable(FBIn, DEM)
    
    ThirdQuartileSlope(FBIn, OutSlope, SlopeTable)
    
    SlopeClassZonalStats(FBIn, OutSlope, SlopeTable)
