# The Contour Filter Strip tool identifies locations in agricultural fields where vegetated buffers,
# located along the contour and in areas of high slope, will intercept the largest amount of runoff.  
#
# A slope mask is first created from the input slope raster of slopes between 4 and 15% slope.  This mask is smoothed
# by majority filter and converted to a polygon.  Interior holes smaller than 1 acre are filled in.
#
# A mask of agricultural fields is created, as identified by an "isAG" value of 1.  Pasture is exluded.  If the runoff
# risk table is provided, the mask is further limited to those fields with the runoff risk classification specified.
#
# The agricultural fields mask and slope mask are intersected, and two meter contours are generated on a by-field basis within
# 4 - 15% slopes in ag fields. Contours are sorted by elevation, and the lowest contour in each field (lowest contour) is selected and
# buffered to a width specified by the user.  The next upstream contour that lies a minimum distance away to
# allow for farming operations between filter strips (determined by the implement width and # of passes between filter strips),
# is then selected and buffered to the same buffer extent as the first filter strip.  The process is repeated for all contours.
#
# Regions of collective flow (> 2 acres) are then used to clip the contours.
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 05/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import arcpy.cartography as CA
import sys, string, os, os.path, time
arcpy.CheckOutExtension("Spatial")


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def FieldBoundarymask(FBIn, RunoffRiskTable, SubsetField, SubsetValue):
    arcpy.AddMessage("Identifying agricultural fields...excluding pasture")

    if RunoffRiskTable == '':
        # Create mask of just agricultural fields    
        lyr = arcpy.MakeFeatureLayer_management(FBIn, "lyr", "\"isAG\" = 1")    
        AgFB = arcpy.CopyFeatures_management(lyr, "in_memory" + "\\AgFB")

    elif RunoffRiskTable  <> '':
        # Create mask of agricultural fields that have a runoff risk specified by subset field and subset value
        
        if SubsetField == '':
            arcpy.AddError("Runoff Risk table was provided but no field was selected to subset results...QUITTING!!!")
            sys.exit(0)
            
        elif SubsetField <> '':
            arcpy.AddMessage("Runoff Risk table provided")
            
            # Add field delimiters to the SubsetField
            Field_type = arcpy.ListFields(RunoffRiskTable, "%s" % SubsetField)[0].type
            Field_del = arcpy.AddFieldDelimiters(arcpy.Describe(RunoffRiskTable).path, "%s" % SubsetField)

            # Values from the selected field will be chosen and added to a list
            if str(Field_type) == 'String':
                ValueList = []
                for value in SubsetValue.split(';'):
                    if value not in ValueList:
                        ValueList.append(value)

            # A query is created to select fields that contain values in the list
            ValueList = ["'%s'" % value for value in ValueList]
            whereClause = "\"isAG\" = 1 and %s IN(%s)" % (Field_del, ', '.join(map(str, ValueList)))
            arcpy.AddMessage("Subsetting fields to those with %s values of: %s" % (SubsetField, ValueList))

            # Make a copy of FBIn
            FB_copy = arcpy.CopyFeatures_management(FBIn, "in_memory" + "\\FB_copy")
            
            # Join SubsetField from runoff risk
            arcpy.JoinField_management(FB_copy, "FBndID", RunoffRiskTable, "FBndID", ["%s" % SubsetField])
            
            # Apply query slection to FB_copy
            lyr = arcpy.MakeFeatureLayer_management(FB_copy, "lyr", whereClause)
            AgFB = arcpy.CopyFeatures_management(lyr, "in_memory" + "\\AgFB")
            num_fields = int(arcpy.GetCount_management(AgFB).getOutput(0))
            if num_fields == 0:
                arcpy.AddError("Zero agricultural fields (non pasture) identified with %s value of %s" % (SubsetField, ValueList))
                sys.exit(0)

            arcpy.Delete_management(FB_copy)
            del[FB_copy]

    return(AgFB)


def GenerateContours(OutSlope, D8FlowAcc, AgFB, DEM, SlopeTable, ZFactor):
    arcpy.AddMessage("Identifying regions between 4 and 15% slope within agricultural fields")

    # Join 75th %ile slope value from slope table
    arcpy.JoinField_management(AgFB, "FBndID", SlopeTable, "FBndID", ["Slope75Pct"])  

    # Reclassify slope into 1 for 4 - 15% slope, NULL if not --> run majority filter --> smooth with boundary clean
    SlopeBC = MajorityFilter(Con(OutSlope, 1, "", '"VALUE" > 4 AND "VALUE" < 15'), "FOUR", "HALF")

    # Convert to polygon 
    slope_poly = arcpy.RasterToPolygon_conversion(SlopeBC, "in_memory" + "\\slope_poly", "SIMPLIFY")

    # Add and calculate area_m2 field
    arcpy.AddField_management(slope_poly, "area_m2")
    arcpy.CalculateField_management(slope_poly, "area_m2", "!shape.area@SQUAREMETERS!", "PYTHON")

    # Delete features smaller than 1 acre in size (4046 meters squared)
    erase_poly = arcpy.MakeFeatureLayer_management(slope_poly, "erase_poly", "area_m2 < 4046")
    arcpy.DeleteFeatures_management(erase_poly)

    # Eliminate interior holes smaller than 1/2 acre in size
    slope_mask = arcpy.EliminatePolygonPart_management(slope_poly, "in_memory" + "\\slope_mask", "AREA", "2023", "", "ANY")

    # Convert accumulation raster from m^2 to acres
    D8Acc_Acre = ((Times(D8FlowAcc, cellsize * cellsize)) / 4046)

    # Reclassify flow accumulation raster into a value of 1 if above 2 acres and NODATA if not 
    flownet_ras = Con(D8Acc_Acre, 1, "", "VALUE >= 2")

    # Expand reclassified flow network by ~ 10 meters on each side (# of cells dependent on resolution) 
    exp_value = max(int(10 / cellsize), 1)
    flownet_exp = Expand(flownet_ras, exp_value, ["1"])
    
    # Use conditional statement to change NULL values in flownet_exp into 1 values 
    flownet_con = Con(IsNull(flownet_exp), 1)

    # Convert to polygon for use as a mask
    flow_mask = arcpy.RasterToPolygon_conversion(flownet_con, "flow_mask")

    # Intersect slope mask and ag fields
    ag_slopes = arcpy.Intersect_analysis([slope_mask, AgFB], "in_memory" + "\\ag_slopes")

    arcpy.AddMessage("Generating contours at different intervals based on slope class")

    # Create list of contours to append to ---> will be merged
    contlist = []

    # SLOPE CLASS 1
    arcpy.AddMessage("slope class 1: 12 - 15%...recommmended spacing of 150 feet")
    # Define contour interval to achieve average spacing of 150 feet
    CI12_15 = float(6.13 / float(ZFactor))
    
    # Select fields with 75th percentile slope value in class 1
    ag_12_15slopes = arcpy.MakeFeatureLayer_management(ag_slopes, "ag_12_15slopes", "\"Slope75Pct\" >= 12")
    # Get count
    field_count12_15 = int(arcpy.GetCount_management(ag_12_15slopes).getOutput(0))
    arcpy.AddMessage("%s fields identified" % field_count12_15)
    if field_count12_15 > 0:
        # if count > 0 --> continue
        elev12_15 = ExtractByMask(DEM, ag_12_15slopes)
        cont12_15 = Contour(elev12_15, "in_memory" + "\\cont12_15", "%s" % CI12_15)
        contlist.append(cont12_15)

        # Cleanup
        arcpy.Delete_management(ag_12_15slopes)
        arcpy.Delete_management(elev12_15)
        del[ag_12_15slopes, elev12_15]

    # SLOPE CLASS 2
    arcpy.AddMessage("slope class 2: 8 - 12%...recommmended spacing of 200 feet")
    # Define contour interval to achieve average spacing of 200 feet
    CI8_12 = float(6.04 / float(ZFactor))
    
    # Select fields with 75th percentile slope value in class 2          
    ag_8_12slopes = arcpy.MakeFeatureLayer_management(ag_slopes, "ag_8_12slopes", "\"Slope75Pct\" >= 8 AND \"Slope75Pct\" < 12")
    # Get count
    field_count8_12 = int(arcpy.GetCount_management(ag_8_12slopes).getOutput(0))
    arcpy.AddMessage("%s fields identified" % field_count8_12)
    if field_count8_12 > 0:
        # if count > 0 --> continue
        elev8_12 = ExtractByMask(DEM, ag_8_12slopes)
        cont8_12 = Contour(elev8_12, "in_memory" + "\\cont8_12", "%s" % CI8_12)
        contlist.append(cont8_12)

        # Cleanup
        arcpy.Delete_management(ag_8_12slopes)
        arcpy.Delete_management(elev8_12)
        del[ag_8_12slopes, elev8_12]

    # SLOPE CLASS 3
    arcpy.AddMessage("slope class 2: 4 - 8%...recommmended spacing of 250 feet")
    # Define contour interval to achieve average spacing of 200 feet
    CI4_8 = float(4.5 / float(ZFactor))
    
    # Select fields with 75th percentile slope value in class 3
    ag_4_8slopes = arcpy.MakeFeatureLayer_management(ag_slopes, "ag_4_8slopes", "\"Slope75Pct\" >= 4 AND \"Slope75Pct\" < 8")
    # Get count 
    field_count4_8 = int(arcpy.GetCount_management(ag_4_8slopes).getOutput(0))
    arcpy.AddMessage("%s fields identified" % field_count4_8)
    if field_count4_8 > 0:
        # if count > 0 --> continue
        elev4_8 = ExtractByMask(DEM, ag_4_8slopes)
        cont4_8 = Contour(elev4_8, "in_memory" + "\\cont4_8", "%s" % CI4_8)
        contlist.append(cont4_8)

        # Cleanup
        arcpy.Delete_management(ag_4_8slopes)
        arcpy.Delete_management(elev4_8)
        del[ag_4_8slopes, elev4_8]

    # Count items in list
    count = len(contlist)

    # If at least one slope class resulted in the generation of contours --> continue
    if count > 0:
        arcpy.AddMessage("merging contours from different slope classes")
        cont_merge = arcpy.Merge_management(contlist, "cont_merge")
        
        # Clip by flow mask - will remove concentrated flow greater than 2 acres
        cont_clip = arcpy.Clip_analysis(cont_merge, flow_mask, "in_memory" + "\\cont_clip")
        
        # Convert from multipart to singlepart                                
        cont_single = arcpy.MultipartToSinglepart_management(cont_clip, "in_memory" + "\\cont_single")

        # smooth
        arcpy.AddMessage("smoothing")
        smoothed = CA.SmoothLine(cont_single, "smoothed", "PAEK", "100 meters")

        arcpy.AddField_management(smoothed, "len_met", "FLOAT")
        arcpy.CalculateField_management(smoothed, "len_met", "!shape.length@meters!", "PYTHON")

        # Delete contours less than 100 meters in length
        arcpy.MakeFeatureLayer_management(smoothed, "erase")
        arcpy.SelectLayerByAttribute_management("erase", "NEW_SELECTION", "\"len_met\" < 100")
        arcpy.DeleteFeatures_management("erase")
        arcpy.DeleteField_management(smoothed, ["len_met"])

        # Describe smoothed
        desc = arcpy.Describe(smoothed)
        OID = desc.OIDFieldName 

        # Delete contours that are closed loops
        rows = arcpy.SearchCursor(smoothed)  
        for row in rows:
            geom = row.shape
            cont_ID = row.getValue("%s" % OID)
            if geom.firstPoint.X == geom.lastPoint.X:  
                arcpy.MakeFeatureLayer_management(smoothed, "curr_cont", '\"%s\" = %s' % (OID, cont_ID))
                arcpy.DeleteRows_management("curr_cont")

        arcpy.AddMessage("buffering contours by %s feet" % BufferWidth)
        BuffWidth = float(BufferWidth) / 2

        # Buffer by the specified BuffWidth and output to file location
        CFS = arcpy.Buffer_analysis(smoothed, OutCFS, "%s Feet" % BuffWidth, "FULL", "ROUND", "NONE", "")

        # Describe CFS to get OID 
        desc = arcpy.Describe(CFS)
        OID = desc.OIDFieldName

        # Zonal stats to find mean slope within each buffer strip
        slope_stats = ZonalStatisticsAsTable(CFS, "%s" % OID, OutSlope, "in_memory" + "\\slope_stats", "DATA", "MEAN")
        arcpy.AddField_management(CFS, "MeanSlope", "FLOAT")
        arcpy.JoinField_management(CFS, "%s" % OID, slope_stats, "%s" % OID, ["MEAN"])
        arcpy.CalculateField_management(CFS, "MeanSlope", "!MEAN!", "PYTHON")
        arcpy.DeleteField_management(CFS, ["MEAN"])

        # Create FieldList ----> append each unique "FBndID" to list as string
        rows = arcpy.SearchCursor(AgFB)
        FieldList = []
        for row in rows:
            FBndID = str(row.getValue("FBndID"))
            FieldList.append(FBndID)

        # Loop through FieldList - select contours that have their center in each field - populate with "FBndID"
        arcpy.AddField_management(CFS, "FBndID", "TEXT")
        cont_lyr = arcpy.MakeFeatureLayer_management(CFS, "in_memory" + "\\cont_lyr")
        for Field in FieldList:
            field_lyr = arcpy.MakeFeatureLayer_management(AgFB, "field_lyr", "\"FBndID\" = '%s'" % Field)
            arcpy.SelectLayerByLocation_management(cont_lyr, "HAVE_THEIR_CENTER_IN", field_lyr, "", "NEW_SELECTION")
            arcpy.CalculateField_management(cont_lyr, "FBndID", "'%s'" % Field, "PYTHON")

        # Delete fields
        arcpy.DeleteField_management(CFS, ["BUFF_DIST", "ORIG_FID"])

        arcpy.Delete_management(cont_merge)
        arcpy.Delete_management(cont_clip)
        arcpy.Delete_management(cont_single)
        arcpy.Delete_management(smoothed)
        del[cont_merge, cont_clip, cont_single, smoothed]

    else:
        arcpy.AddMessage("no agricultural fields identified with 75th percentile slope value greater than 4%")

    # Cleanup
    arcpy.Delete_management(SlopeBC)
    arcpy.Delete_management(slope_poly)
    arcpy.Delete_management(slope_mask)
    arcpy.Delete_management(D8Acc_Acre)
    arcpy.Delete_management(flownet_ras)
    arcpy.Delete_management(flownet_exp)
    arcpy.Delete_management(flownet_con)
    arcpy.Delete_management(flow_mask)
    arcpy.Delete_management(ag_slopes)
    del[SlopeBC, slope_poly, slope_mask, D8Acc_Acre, flownet_ras, flownet_exp, flownet_con, flow_mask, ag_slopes]

    arcpy.Delete_management("in_memory")


if __name__ == "__main__":

    # Define Parameters
    FBIn = arcpy.GetParameterAsText(0)
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        basename = arcpy.Describe(FBIn).baseName
        arcpy.AddError("Found a join on %s that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    OutSlope = arcpy.GetParameterAsText(1)
    SlopeTable = arcpy.GetParameterAsText(2)
    DEM = arcpy.GetParameterAsText(3)
    D8FlowAcc = arcpy.GetParameterAsText(4)
    ZFactor = arcpy.GetParameterAsText(5)
    BufferWidth = arcpy.GetParameterAsText(6)
    RunoffRiskTable = arcpy.GetParameterAsText(7)
    if RunoffRiskTable <> "":
        # Test for a join on RunoffRisk table
        basename = arcpy.Describe(RunoffRiskTable).baseName
        hasJoin = joinCheck(RunoffRiskTable)
        if hasJoin:
            arcpy.AddError("Found a join on %s that must be removed before running tool...QUITTING!!!!" % basename)
            sys.exit(0)
    SubsetField = arcpy.GetParameterAsText(8)
    SubsetValue = arcpy.GetParameterAsText(9)
    OutCFS = arcpy.GetParameterAsText(10)

    # Set environments
    arcpy.env.snapRaster = DEM
    arcpy.env.extent = DEM
    rDesc = arcpy.Describe(DEM)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder 

    # Define cellsize of D8FlowAcc
    cellsize = float(arcpy.GetRasterProperties_management(OutSlope, "CELLSIZEX").getOutput(0))

    # Run Modules
    AgFB = FieldBoundarymask(FBIn, RunoffRiskTable, SubsetField, SubsetValue)

    GenerateContours(OutSlope, D8FlowAcc, AgFB, DEM, SlopeTable, ZFactor)

    
