# Controlled Drainage------------------------------------------------------------------------------------------------------------------
#
# the Drainage Water Management tool identifies the largest area within any user-defined contour interval
# within each tile-drained, agricultural field.  The tool works by finding the number of user-defined contour
# intervals (rounded to the nearest integer) that can exist within a given field, based on the range of elevation values.
#
# For example, if the contour interval is defined as 1 meter:
# a field with an elevation range of 432 cm contains approximately four 1-meter contours;
#
# 4.32 / 100 = 4.32
# If < .5 ---> round down = 4
# If > .5 ---> round up = 5
#
# The field is then sliced into that number of equal interval zones (i.e. 4 in example above) and the largest contour is selected.
# If the contour occupies a given percentage (defined by the user) of the field, that field is flagged as a candidate for controlled drainage.
#
# If flagged as a candidate site, each contour is converted to a polygon and added to the output "DrainageMgmt + inHUC" polygon layer in the fgdb.  

##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 05/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time
arcpy.CheckOutExtension("Spatial")


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def ControlledDrainage(FBIn, DrainageTable, DEM, MinPctField):
    
    # Create feature class "DrainageMgmt"  - to append 1 m contour(s) to that are possible controlled drainage sites
    Dmgmt_tmp = arcpy.CreateFeatureclass_management("in_memory", "Dmgmt_tmp", "POLYGON", "", "", "", FBIn)

    # Copy to output file location
    DrainageMgmt = arcpy.CopyFeatures_management(Dmgmt_tmp, OutDrainageMgmt)

    # Cleanup
    arcpy.Delete_management(Dmgmt_tmp)
    del[Dmgmt_tmp]

    # Add fields - to be populated with each append of a suitable contour
    arcpy.AddField_management(DrainageMgmt, "FBndID", "TEXT")
    arcpy.AddField_management(DrainageMgmt, "cont_acres", "FLOAT")
    arcpy.AddField_management(DrainageMgmt, "fld_acres", "FLOAT")
    arcpy.AddField_management(DrainageMgmt, "pct_field", "FLOAT")
        
    # Make a copy of FB feature class
    FBIn_copy = arcpy.CopyFeatures_management(FBIn, "in_memory" + "\\FBIn_copy")

    # Join "DRAINED" field from drainage table to copy of FB feature class
    arcpy.JoinField_management(FBIn_copy, "FBndID", DrainageTable, "FBndID", ["DRAINED"])

    # Make feature layer of tile-drained ag fields - this feature layer will be looped through
    fld_selection = arcpy.MakeFeatureLayer_management(FBIn_copy, "fld_selection", "\"DRAINED\" = 'YES'")

    # Count how many fields will be tested
    FieldCount = int(arcpy.GetCount_management(fld_selection).getOutput(0))
    if FieldCount == 0:
        arcpy.AddError("0 tile drained ag fields identified to test for drainage water management opportunities...QUITTING!!")
        arcpy.Delete_management(DrainageMgmt)
        sys.exit(0)
    elif FieldCount >= 1:
        arcpy.AddMessage("%s tile drained ag fields identified to test for drainage water management opportunities..." % FieldCount)
    
    # Convert fields to raster
    td_ras = arcpy.FeatureToRaster_conversion(fld_selection, "FBndID", "in_memory" + "\\td_ras", cellsize)

    # Find elevation range within each field
    td_range = ZonalStatistics(td_ras, "VALUE", DEM, "RANGE")

    # Identify user-specified contour interval (represents zone of groundwater influence)
    ContourInt = float(ContourInterval)
    arcpy.AddMessage("Contour Interval of %s meters specified" % ContourInt)

    # Divide range of each field by contour interval (accounting for z-factor) 
    td_divide = Divide(td_range, (ContourInt / float(ZFactor)))

    # RoundDown and RoundUp
    td_rounddown = RoundDown(td_divide)
    td_roundup = RoundUp(td_divide)

    # Subtract td_rounddown from td_divide
    td_diff = td_divide - td_rounddown

    # If td_diff is < .5 --> round down, if > .5, roundup; this equals the approximate # of zones (user-defined contours) to divide the field into
    Zones = Con(td_diff, td_rounddown, td_roundup, "VALUE <= .5")

    # Perform zonal statistics to know how many zones to slice each field into
    zone_stats = ZonalStatisticsAsTable(fld_selection, "FBndID", Zones, "in_memory" + "\\zone_stats", "#", "MEAN")
    arcpy.JoinField_management(fld_selection, "FBndID", zone_stats, "FBndID", ["MEAN"])

    # Add and calculate "Zones" field to indicate number of zones in each field
    arcpy.AddField_management(fld_selection, "Zones", "LONG")
    arcpy.CalculateField_management(fld_selection, "Zones", ("!MEAN!"), "PYTHON")

    arcpy.AddMessage("A %s meter contour must occupy %s percent of field to be indentified as a candidate for drainage water management" % (ContourInt, MinPctField))
    arcpy.AddMessage("Testing fields...this may take a few minutes")
 
    # Create a tracker so the user knows how many fields have been processed
    field_count_intervals = [50, 100, 150, 200, 250, 300, 350, 400, 450, 500]
    field_count = 1

    # Add exception; in very low relief watersheds, the # of user defined contour zones might be 0 (total elevation range < .3 m) --- convert to value of 1 (sp - 9/17/14)
    with arcpy.da.UpdateCursor(fld_selection, ["Zones"]) as cursor:
        for row in cursor:
            if row[0] == 0:
                row[0] = 1
            cursor.updateRow(row)
        
    with arcpy.da.SearchCursor(fld_selection, ["FBndID", "Zones", "Acres"]) as cursor:
        for row in cursor:
            arcpy.env.scratchWorkspace = "in_memory"
            
            Field = row[0]
            Zones = int(row[1])
            FieldAcres = row[2]
            
            # Create a counter to know how many fields have been processed
            field_count = int(field_count + 1)
            if field_count in field_count_intervals:
                arcpy.AddMessage("%s fields processed" % field_count)
                
            # Create a feature layer of each field 
            arcpy.MakeFeatureLayer_management(fld_selection, "field", "\"FBndID\" = '%s'" % Field)

            # Copy feature layer to in memory - will become mask
            mask = arcpy.CopyFeatures_management("field", "in_memory" + "\\mask")

            # Set mask as environment mask and extent
            arcpy.env.mask = "in_memory" + "\\mask"
            arcpy.env.extent = "in_memory" + "\\mask"

            # Slice the field into the number of elevation zones identified ---> will only do this for current field because of mask
            fld_slice = Slice(DEM, '%s' % Zones, "Equal_Interval")

            # Make a raster layer of sliced field - to loop through "slice" and test for suitability
            sliced = arcpy.MakeRasterLayer_management(fld_slice, "sliced")

            # For each slice in sliced (which approximates each 1 meter contour)
            with arcpy.da.SearchCursor(sliced, ["COUNT"]) as xcursor:
                for x in xcursor:

                    # "COUNT" equals the # of cells in each contour
                    count = x[0]
                    
                    # Convert the cell count ("COUNT") to a size in acres
                    contour_acres = (count * resolution) / 4046.00

                    # Determine the percent of the field that each contour occupies
                    pct_field = (contour_acres / FieldAcres) * 100

                    # If the percent is greater than the minimum percent as specified by the user
                    if pct_field >= float(MinPctField):
                        
                        # Convert the contour that meets the requirements into a polygon
                        lyr = arcpy.MakeRasterLayer_management(fld_slice, "lyr", "\"COUNT\" = %s" % count)
                        slice_poly = arcpy.RasterToPolygon_conversion(lyr, "in_memory" + "\\slice_poly")

                        # Dissolve by grid code
                        slice_diss = arcpy.Dissolve_management(slice_poly, "in_memory" + "\\slice_diss", "GRIDCODE")                                                 

                        # Add fields to slice_diss: 
                        arcpy.AddField_management(slice_diss, "cont_acres", "FLOAT")
                        arcpy.AddField_management(slice_diss, "FBndID", "TEXT")
                        arcpy.AddField_management(slice_diss, "fld_acres", "FLOAT")
                        arcpy.AddField_management(slice_diss, "pct_field", "FLOAT")

                        # Calculate fields
                        arcpy.CalculateField_management(slice_diss, "cont_acres", "%s" % contour_acres, "PYTHON")
                        arcpy.CalculateField_management(slice_diss, "FBndID", '"%s"' % Field, "PYTHON")
                        arcpy.CalculateField_management(slice_diss, "fld_acres", "%s" % FieldAcres, "PYTHON")
                        arcpy.CalculateField_management(slice_diss, "pct_field", "%s" % pct_field, "PYTHON")

                        # Append to "DrainageMgmt" feature class - fields should match 
                        arcpy.Append_management(slice_diss, DrainageMgmt, "NO_TEST")

                        # Cleanup
                        arcpy.Delete_management(slice_poly)
                        arcpy.Delete_management(slice_diss)
                        del[slice_poly, slice_diss]
                        
                del(xcursor)
                    
            # Clear environments
            arcpy.ClearEnvironment("mask")
            arcpy.ClearEnvironment("extent")

            # Cleanup
            arcpy.Delete_management(mask)
            arcpy.Delete_management(fld_slice)
            arcpy.Delete_management(sliced)
            del[mask, fld_slice, sliced]

    mgmt_fields = int(arcpy.GetCount_management(DrainageMgmt).getOutput(0))
    if mgmt_fields == 0:
        arcpy.AddMessage("All fields processed...%s opportunities for drainage water management" % mgmt_fields)
        arcpy.Delete_management(DrainageMgmt)
        del[DrainageMgmt]
    elif mgmt_fields > 0:
        arcpy.AddMessage("All fields processed...%s opportunities for drainage water management" % mgmt_fields)
        arcpy.AddMessage("Multiple opportunities may exist in a single field")

    # Cleanup
    arcpy.Delete_management(FBIn_copy)
    arcpy.Delete_management(fld_selection)
    arcpy.Delete_management(td_ras)
    arcpy.Delete_management(td_range)
    arcpy.Delete_management(td_divide)
    arcpy.Delete_management(td_rounddown)
    arcpy.Delete_management(td_roundup)
    arcpy.Delete_management(td_diff)
    arcpy.Delete_management(Zones)
    arcpy.Delete_management(zone_stats)
    del[FBIn_copy, fld_selection, td_ras, td_range, td_divide, td_rounddown, td_roundup, td_diff, Zones, zone_stats]
    arcpy.Delete_management("in_memory")

    
if __name__ == "__main__":

    # Define Parameters
    FBIn = arcpy.GetParameterAsText(0)
    DEM = arcpy.GetParameterAsText(1)
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        basename = arcpy.Describe(FBIn).baseName
        arcpy.AddError("Found a join on %s that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    DrainageTable = arcpy.GetParameterAsText(2)
    ZFactor = arcpy.GetParameterAsText(3)
    ContourInterval = arcpy.GetParameterAsText(4)
    MinPctField = arcpy.GetParameterAsText(5)
    OutDrainageMgmt = arcpy.GetParameterAsText(6)

    # Set extent, snap, workspace
    arcpy.env.extent = DEM
    arcpy.env.snapRaster = DEM
    rDesc = arcpy.Describe(DEM)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder             

    # Determine cellsize and resolution of input DEM
    cellsize = float(arcpy.GetRasterProperties_management(DEM, "CELLSIZEX").getOutput(0))
    resolution = float(cellsize * cellsize)
    
    # Run Modules
    ControlledDrainage(FBIn, DrainageTable, DEM, MinPctField)

