# Depression Watersheds-----------------------------------------------------------------------
#
# The drainage area to each depression in the input feature class is
# delineated using D8 flow direction values derived from an unfilled DEM
#
# Drainage areas are converted to polygons and their area (in HA) is calculated
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time
arcpy.CheckOutExtension("Spatial")


# Modules ------------------------------------------------------------------------------------

def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def DelineateDrainageAreas(Depressions, DEM, OutDrainageSheds):

    # Calculate Flow Direction using DEM - should be unfilled
    arcpy.AddMessage("Calculating D8 Flow Direction using an unfilled DEM")
    D8FlowDir = FlowDirection(DEM, "", "")

    # Convert depressions to raster using Depress_ID 
    depress_ras = arcpy.FeatureToRaster_conversion(Depressions, "Depress_ID", "depress_ras", cellsize)

    # Add and calculated Depress_ID field
    arcpy.AddField_management(depress_ras, "Depress_ID", "LONG")
    arcpy.CalculateField_management(depress_ras, "Depress_ID", "!Value!", "PYTHON")

    # Use FlowDir to find watersheds
    arcpy.AddMessage("generating watersheds")
    wsheds = Watershed(D8FlowDir, depress_ras, "Depress_ID")

    # Convert wsheds raster to polygon & dissolve by gridcode values
    wsheds_poly = arcpy.RasterToPolygon_conversion(wsheds, "in_memory" + "\\wsheds_poly", "", "Value") 
    wsheds_diss = arcpy.Dissolve_management(wsheds_poly, OutDrainageSheds, "gridcode")

    # Add ID field to wsheds_diss layer 
    arcpy.AddField_management(wsheds_diss, "Depress_ID")
    arcpy.CalculateField_management(wsheds_diss, "Depress_ID", "!gridcode!", "PYTHON")
    arcpy.DeleteField_management(wsheds_diss, "gridcode")
    
    # Add and calculate a drainage area field to Depress_Wsheds layer
    arcpy.AddField_management(wsheds_diss, "DrainageHA", "FLOAT")
    arcpy.CalculateField_management(wsheds_diss, "DrainageHA", "!shape.area@HECTARES!", "PYTHON")       

    # Cleanup
    arcpy.Delete_management(D8FlowDir)
    arcpy.Delete_management(depress_ras)
    arcpy.Delete_management(wsheds)
    arcpy.Delete_management(wsheds_poly)
    del[D8FlowDir, depress_ras, wsheds, wsheds_poly]
    arcpy.Delete_management("in_memory")


##------------------------------------------------------------------------------
##------------------------------------------------------------------------------
if __name__ == "__main__":

    # Define Parameters
    Depressions = arcpy.GetParameterAsText(0)
    hasJoin = joinCheck(Depressions)
    if hasJoin:
        basename = arcpy.Describe(Depressions).baseName
        arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    DEM = arcpy.GetParameterAsText(1)
    OutDrainageSheds = arcpy.GetParameterAsText(2)

    # Set environments
    arcpy.env.extent = DEM
    arcpy.env.snapRaster = DEM
    rDesc = arcpy.Describe(DEM)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True

    cellsize = float(arcpy.GetRasterProperties_management(DEM, "CELLSIZEX").getOutput(0))

    # Run modules
    DelineateDrainageAreas(Depressions, DEM, OutDrainageSheds)  


        
