# Surface Depressions----------------------------------------------------------------------------------------------------------------------
#
# The Depression identification tool identifies surface depressions
# by subtracting the input DEM from the filled DEM.    
#
# The user has the option to limit sinks based on user specifications ---> any, all, or combination of:
    # 1)hydric soils
    # 2)agricultural fields
    # 3)do not intersect with stream reach
    # 4)minimum depth
    # 5)minimum surface area
#
# Hydric soils are identified as mapunits with a hydric classification of either all hydric or partially hydric
# Polygons are overlaid with gSSURGO to extract the % hydric soils within each depression. 

# Agricultural fields are identified from the "isAG" field (values of 1 or 2) of the field boundary feature class
#
# Polygons are overlaid with the unfilled input DEM to extract the range of elevation
# values within each depression. This range of values represents the maximum depth.
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time
arcpy.CheckOutExtension("Spatial")


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def DepressionIdentification(DEM, gSSURGO, ZFactor, FBIn, StreamReach, MinHydric, LimitByDepth, MinimumSize):

    # Fill DEM
    arcpy.AddMessage("filling DEM")
    DEMFill = Fill(DEM)
    
    # Subtract unfilled DEM from filled DEM
    arcpy.AddMessage("finding sinks")
    FillReg = Minus(DEMFill, DEM)

    # Convert values of 0 or less to NODATA (non-depressional), convert values > 0 to 1 
    AllSinks = Con(FillReg, 1, "", "VALUE > .01")

    # Convert to polygon
    Depressions = arcpy.RasterToPolygon_conversion(AllSinks, OutDepressions)

    arcpy.Delete_management(AllSinks)
    del[AllSinks]

    # Count # of Depressions that exist ---> in the case that none exist
    dep_count = int(arcpy.GetCount_management(Depressions).getOutput(0))
    
    if dep_count >= 1:
        arcpy.AddMessage("%s depressions identified initially" % dep_count)

        if MinimumSize <> '':
            arcpy.AddMessage("Deleting sinks less than %s acres in size" % MinimumSize)
            
            # Convert minimum surface area (defined by user) from acres to meters squared
            MinSize = (float(MinimumSize) * float(4046.00))

            # Select Depressions smaller than minimum size
            arcpy.MakeFeatureLayer_management(Depressions, "small_sinks", "\"Shape_Area\" < %s" % MinSize)
            arcpy.DeleteFeatures_management("small_sinks")

            dep_count = int(arcpy.GetCount_management(Depressions).getOutput(0))
            if dep_count < 1:
                arcpy.AddMessage("No depressions remaining after removing those less than %s acres in size" % MinimumSize)
                arcpy.Delete_management(Depressions)
                sys.exit(0)

        if FBIn <> '':
            # Create mask of just agricultural fields (including pasture)
            AgFB = arcpy.MakeFeatureLayer_management(FBIn, "AgFB", "\"IsAG\" >= 1")

            # Remove Depressions not centered on ag land
            arcpy.AddMessage("Removing sinks not centered on ag land")
            arcpy.MakeFeatureLayer_management(Depressions, "AgSinks")

            # Select those centered on ag fields ---> switch selection
            arcpy.SelectLayerByLocation_management("AgSinks", "HAVE_THEIR_CENTER_IN", AgFB, "", "NEW_SELECTION")
            arcpy.SelectLayerByAttribute_management("AgSinks", "SWITCH_SELECTION")

            # Delete selection
            arcpy.DeleteFeatures_management("AgSinks")

            # Cleanup
            arcpy.Delete_management(AgFB)
            del[AgFB]

            dep_count = int(arcpy.GetCount_management(Depressions).getOutput(0))
            if dep_count < 1:
                arcpy.AddMessage("No depressions remaining after removing those centered on ag land")
                arcpy.Delete_management(Depressions)
                sys.exit(0)

        if StreamReach <> '':
            arcpy.AddMessage("Removing sinks that intersect the stream reach")

            # Select Depressions that intersect the stream reach - delete
            arcpy.MakeFeatureLayer_management(Depressions, "strm_sinks")
            arcpy.SelectLayerByLocation_management("strm_sinks", "INTERSECT", StreamReach)
            arcpy.DeleteFeatures_management("strm_sinks")

            dep_count = int(arcpy.GetCount_management(Depressions).getOutput(0))
            if dep_count < 1:
                arcpy.AddMessage("No depressions remaining after removing those that intersect the stream reach")
                arcpy.Delete_management(Depressions)
                sys.exit(0)          

        # Add and populate a unique "Depress_id" field to Depressions layer ---> populate with OID
        desc = arcpy.Describe(Depressions)
        OID = desc.OIDFieldName
        arcpy.AddField_management(Depressions, "Depress_ID")
        arcpy.CalculateField_management(Depressions, "Depress_ID", "!%s!" % OID, "PYTHON") 
        
        # Calculate the average percent of hydric soils within each depression
        HydricPct = Lookup(gSSURGO, "Hydric")

        arcpy.AddMessage("Calculating mean percent hydric soil in each depression")
        # Zonal mean of hydric percent within each depression
        HydricTable = ZonalStatisticsAsTable(Depressions, "Depress_ID", HydricPct, "in_memory" + "\\HydricTable", "#", "MEAN")

        # Join and populate "PctHydric" field --> hydric percent within each depression
        arcpy.JoinField_management(Depressions, "Depress_ID", HydricTable, "Depress_ID", ["MEAN"])
        arcpy.AddField_management(Depressions, "PctHydric", "FLOAT")
        arcpy.CalculateField_management(Depressions, "PctHydric", "!MEAN!", "PYTHON")
        arcpy.DeleteField_management(Depressions, ["MEAN"])

        # Cleanup
        arcpy.Delete_management(HydricPct)
        arcpy.Delete_management(HydricTable)
        del[HydricPct, HydricTable]

        if MinHydric <> '':
            arcpy.AddMessage("Deleting sinks less than %s pct hydric" % MinHydric)
            
            # Select depression with percent hydric less than specified value - delete
            arcpy.MakeFeatureLayer_management(Depressions, "non_hydric")
            arcpy.SelectLayerByAttribute_management("non_hydric", "NEW_SELECTION", "PctHydric < %s" % MinHydric)
            arcpy.DeleteFeatures_management("non_hydric")

            dep_count = int(arcpy.GetCount_management(Depressions).getOutput(0))
            if dep_count < 1:
                arcpy.AddMessage("No depressions remaining after removing those less than %s pct hydric" % MinHydric)
                arcpy.Delete_management(Depressions)
                sys.exit(0)    

        # Calculate maximum depth of Depressions in cm 
        arcpy.AddMessage("Calculating maximum depth")
        DepthStats = ZonalStatisticsAsTable(Depressions, "Depress_ID", FillReg, "in_memory" + "\\DepthStats", "", "MAXIMUM")

        # Add and calculate MaxDepthCM field --> maximum depth in cm
        arcpy.JoinField_management(Depressions, "Depress_ID", DepthStats, "Depress_ID", ["MAX"])
        arcpy.AddField_management(Depressions, "MaxDepthCM")
        arcpy.CalculateField_management(Depressions, "MaxDepthCM", "(!MAX! * (%s * 100))" % ZFactor, "PYTHON")
        arcpy.DeleteField_management(Depressions, ["gridcode", "MAX", "id"])

        # Cleanup
        arcpy.Delete_management(DepthStats)
        del[DepthStats]
        
        if LimitByDepth <> '':
            # Remove Depressions shallower than minimum depth
            arcpy.AddMessage("Deleting sinks less than %s cm deep" % LimitByDepth)

            # Select those shallower than LimitByDepth - delete
            arcpy.MakeFeatureLayer_management(Depressions, "shallow_sinks")
            arcpy.SelectLayerByAttribute_management("shallow_sinks", "NEW_SELECTION", "MaxDepthCM < %s" % LimitByDepth)
            arcpy.DeleteFeatures_management("shallow_sinks")

            dep_count = int(arcpy.GetCount_management(Depressions).getOutput(0))
            if dep_count < 1:
                arcpy.AddMessage("No depressions remaining after removing those less than %s cm deep" % LimitByDepth)
                arcpy.Delete_management(Depressions)
                sys.exit(0)    

        if DepthRaster == 'true':
            arcpy.AddMessage("Output Depth Raster selected")
            if OutDepthRaster == '':
                arcpy.AddError("Output Depth Raster is selected but no output file location is provided...QUITTING!!!")
                sys.exit(0)
            elif OutDepthRaster <> '':
                arcpy.AddMessage("Generating output depression depth raster")
                
                # Extract the FillReg raster to extent of depressions
                depth_ras = ExtractByMask(FillReg, Depressions)

                # Copy to output file location
                arcpy.CopyRaster_management(depth_ras, OutDepthRaster)

                # Build pyramids
                arcpy.BuildPyramids_management(OutDepthRaster)

                # Cleanup
                arcpy.Delete_management(depth_ras)
                del[depth_ras]

    elif dep_count < 1:
        arcpy.AddMessage("No surface depressions identified")
        arcpy.Delete_management(Depressions)

    # Cleanup
    arcpy.Delete_management(DEMFill)
    arcpy.Delete_management(FillReg)
    del[DEMFill, FillReg]
    arcpy.Delete_management("in_memory")


##------------------------------------------------------------------------------
##------------------------------------------------------------------------------
if __name__ == "__main__":

    # Define Parameters
    DEM = arcpy.GetParameterAsText(0)
    gSSURGO = arcpy.GetParameterAsText(1)
    ZFactor = arcpy.GetParameterAsText(2)
    FBIn = arcpy.GetParameterAsText(3)
    if FBIn <> "":
        # Test for a join on FBIn
        basename = arcpy.Describe(FBIn).baseName
        hasJoin = joinCheck(FBIn)
        if hasJoin:
            arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
            sys.exit(0)
    StreamReach = arcpy.GetParameterAsText(4)
    MinHydric = arcpy.GetParameterAsText(5)
    LimitByDepth = arcpy.GetParameterAsText(6)
    MinimumSize = arcpy.GetParameterAsText(7)
    OutDepressions = arcpy.GetParameterAsText(8)
    DepthRaster = arcpy.GetParameterAsText(9)
    OutDepthRaster = arcpy.GetParameterAsText(10)

    # Set environments
    arcpy.env.extent = DEM
    arcpy.env.snapRaster = DEM
    rDesc = arcpy.Describe(DEM)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True

    # Run modules
    DepressionIdentification(DEM, gSSURGO, ZFactor, FBIn, StreamReach, MinHydric, LimitByDepth, MinimumSize)

