# CALCULATES THE DISTANCE TO STREAM NETWORK FOR EVERY GRID CELL, FOLLOWING D8 FLOWPATHS
#
# The D8 Distance To Stream tool uses an input Stream Reach (polyline) and a D8 Flow Direction grid
# to calculate the horizontal distance (in meters) to the channel from each grid cell, moving downslope
# according to the D8 flow model until a stream grid cell is encountered.
#
# Ideally, the stream reach should represent only perennial flow.
#
# The stream reach polyline is converted to a raster and serves as the input stream raster grid in the
# "D8 Distance To Stream" tool available in TauDEM software.
#
# The output is a D8 Distance To Stream raster, a continuous grid where each cell value is the horizontal distance,
# (in meters) from that cell to the stream reach, following the the flow path defined by the D8 Flow Direction grid.
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")


## Modules---------------------------------------------------------------------------

def DistanceToStreamCalc(D8FlowDir, StreamReach):
    arcpy.AddMessage("Creating Distance To Stream Raster")

    # Reclassify D8FlowDir into taudem flow direction values ---> copy to TDProcDir as uncompressed tiff 
    TDFlowDirRec = Reclassify(D8FlowDir, "VALUE",
                           RemapRange([[1, 1], [128, 2], [64, 3], [32, 4], [16, 5], [8, 6], [4, 7], [2, 8]]))
    TDFlowDir = arcpy.CopyRaster_management(TDFlowDirRec, TDProcDir + "\\TDFlowDir.tif")

    # convert StreamReach to raster ---> StreamNetRas ----> copy to TDProcDir as uncompressed tiff
    desc = arcpy.Describe(StreamReach)
    OID = desc.OIDFieldName
    strm_ras = arcpy.PolylineToRaster_conversion(StreamReach, "%s" % OID, "strm_ras", "", "", cellsize)

    # Convert background values to NODATA
    StreamNetRas = Con(strm_ras, 1, 0, "VALUE >= 0")

    # Copy to TDProcDir as uncompressed tiff
    arcpy.CopyRaster_management(StreamNetRas, TDProcDir + "\\TDStreamNetRas.tif")

    core_cnt = multiprocessing.cpu_count()
    core_no = core_cnt - 1

    # Run taudem "distance to stream" tool
    callstr = 'mpiexec -n %s D8HDistToStrm -p ' % core_no + \
              TDProcDir + "\\TDFlowDir.tif" + ' -src ' + \
              TDProcDir + "\\TDStreamNetRas.tif" + ' -dist ' + \
              TDProcDir + "\\TDDistanceToStream.tif"
    call(callstr, shell = True)

    # Calculate statistics on "TDDistanceToStream.tif"  ---> copy to output file location
    TDDistanceToStream = arcpy.CalculateStatistics_management(TDProcDir + "\\TDDistanceToStream.tif")
    DistanceToStream = arcpy.CopyRaster_management(TDDistanceToStream, OutDistanceToStream)

    # Build pyramids
    arcpy.BuildPyramids_management(OutDistanceToStream)

    # Delete
    arcpy.Delete_management(TDFlowDirRec)
    arcpy.Delete_management(strm_ras)
    arcpy.Delete_management(StreamNetRas)
    del[TDFlowDirRec, strm_ras, StreamNetRas]
    arcpy.Delete_management(TDProcDir)

    return(DistanceToStream)

            
##------------------------------------------------------------------------------
##------------------------------------------------------------------------------
if __name__ == "__main__":
    
    # Define parameters
    StreamReach = arcpy.GetParameterAsText(0)
    D8FlowDir = arcpy.GetParameterAsText(1)
    OutDistanceToStream = arcpy.GetParameterAsText(2)
        
    # Set extent, snap, workspace
    arcpy.env.extent = D8FlowDir
    arcpy.env.snapRaster = D8FlowDir
    arcpy.env.outputCoordinateSystem = D8FlowDir
    rDesc = arcpy.Describe(D8FlowDir)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.compression = "NONE" 

    # Determine resolution of input D8FlowDir
    cellsize = float(arcpy.GetRasterProperties_management(D8FlowDir, "CELLSIZEX").getOutput(0))
    
    # Create and define folder for TauDEM outputs (in ProcDir) - will be deleted at end of script
    scriptpath = sys.path[0]
    if ' ' in scriptpath:
        arcpy.AddError("space exists in %s that must be removed from path name...QUITTING!!!" % scriptpath)
        sys.exit(0)
    else:
        arcpy.CreateFolder_management(scriptpath, "TDProcDir")
        TDProcDir = scriptpath + "\\TDProcDir"

    # Run Modules------------------------------------------------------------------------------
    DistanceToStream = DistanceToStreamCalc(D8FlowDir, StreamReach)

