# Flow Network Definition = Area threshold --------------------------------------------------------------------------------------------------
#
# Defines a stream network using an area threshold in acres
# Required input is a flow accumulation and flow direction raster

# Output will be a stream network "AreaFlowNet" + inHUC
# For use in the ACPF, the next step will be manual revision of the "StreamType" field of
# the flow network to distinguish stream reaches based on permanence
#
# The user will define thresholds in acres of upstream drainage

##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
arcpy.CheckOutExtension("Spatial")


def StreamNetByThreshold(D8FlowAcc, D8FlowDir, AreaThreshold):

    # Convert area threshold from acres (input) to meters: 1 acre = 4046 meters
    thresh_meters = float(AreaThreshold) * 4046
    number_cells = float(thresh_meters / resolution)
    arcpy.AddMessage("Area threshold of %s acres...." % (AreaThreshold))
   
    # Threshold the flow accumulation raster using a CON statement and convert 0 background values to null
    FlowNetRas = Con(D8FlowAcc, 1, "", "VALUE >= %s" % number_cells)

    # Calculate Stream Order using FlowNetRas
    Order = StreamOrder(FlowNetRas, D8FlowDir, "STRAHLER")

    # Use the stream to feature to convert from raster to vector
    FlowNetwork = StreamToFeature(Order, D8FlowDir, OutFlowNet, "NO_SIMPLIFY")

    # Add StreamType field
    arcpy.AddField_management(FlowNetwork, "StreamType", "SHORT")
    arcpy.CalculateField_management(FlowNetwork, "StreamType","0","PYTHON","#")

    # Add Stream Order field
    arcpy.AddField_management(FlowNetwork, "STRAHLOrd", "LONG")
    arcpy.CalculateField_management(FlowNetwork, "STRAHLOrd", "!grid_code!", "PYTHON")
    arcpy.DeleteField_management(FlowNetwork, ["grid_code"])

    # Select only those segments that are centered in the WBD (if the WBD is provided)
    if WBD <> '':
        arcpy.MakeFeatureLayer_management(FlowNetwork, "flownet_fl")
        arcpy.SelectLayerByLocation_management("flownet_fl", "HAVE_THEIR_CENTER_IN", WBD)
        arcpy.SelectLayerByAttribute_management("flownet_fl", "SWITCH_SELECTION")
        arcpy.DeleteFeatures_management("flownet_fl")
 
    # Cleanup
    arcpy.Delete_management(FlowNetRas)
    arcpy.Delete_management(Order)
    del[FlowNetRas, Order]


##------------------------------------------------------------------------------
if __name__ == "__main__":

    # Define Parameters
    D8FlowAcc = arcpy.GetParameterAsText(0)  
    D8FlowDir = arcpy.GetParameterAsText(1)
    AreaThreshold = arcpy.GetParameterAsText(2)
    WBD = arcpy.GetParameterAsText(3)
    OutFlowNet = arcpy.GetParameterAsText(4)

    # Set environments
    arcpy.env.extent = D8FlowAcc
    arcpy.env.snapRaster = D8FlowAcc
    arcpy.env.overwriteOutput = True

    # Determine the cellsize and resolution of the input flow accumulation raster
    cellsize = float(arcpy.GetRasterProperties_management(D8FlowAcc, "CELLSIZEX").getOutput(0))
    resolution = float(cellsize * cellsize)

    # Run Modules
    StreamNetByThreshold(D8FlowAcc, D8FlowDir, AreaThreshold)

