# Grassed Waterways ------------------------------------------------------------------------------------------------------------
#
# Applies a threshold to a flow accumulation raster to define collective flow pathways where a grassed waterway may be suitable
#
# The user will define a threshold between .4 and 6 HA (1 - 15 acres), above which is identified as possible locations for grassed waterways.
#
# The stream reach polyline will be used to remove locations where a perennial flow channel already exists 
#
# The field boundary feature class is used to limit grassed waterways to agricultural fields, as identified by an isAG value of 1.
# If provided, the runoff risk table will be used to further limit grassed waterways to those fields with the ruonff risk classification specified
#
# Remaining collective flow grid cells are converted to a polyline
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 05/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##---------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def FieldBoundarymask(FBIn, RunoffRiskTable, SubsetField, SubsetValue):
    
    if RunoffRiskTable == '':
        # Create mask of just agricultural fields (not pasture)
        lyr = arcpy.MakeFeatureLayer_management(FBIn, "lyr", "\"isAG\" = 1")    
        AgFB = arcpy.CopyFeatures_management(lyr, "in_memory" + "\\AgFB")

    elif RunoffRiskTable  <> '':
        # Create mask of agricultural fields that have a runoff risk specified by subset field and subset value
        
        if SubsetField == '':
            arcpy.AddError("Runoff Risk table was provided but no field was selected to subset results...QUITTING!!!")
            sys.exit(0)
            
        elif SubsetField <> '':
            arcpy.AddMessage("runoff risk table provided")
            
            # Add field delimiters to the SubsetField
            Field_type = arcpy.ListFields(RunoffRiskTable, "%s" % SubsetField)[0].type
            Field_del = arcpy.AddFieldDelimiters(arcpy.Describe(RunoffRiskTable).path, "%s" % SubsetField)

            # Values from the selected field will be chosen and added to a list
            if str(Field_type) == 'String':
                ValueList = []
                for value in SubsetValue.split(';'):
                    if value not in ValueList:
                        ValueList.append(value)

            # A query is created to select fields that contain values in the list
            ValueList = ["'%s'" % value for value in ValueList]
            whereClause = "\"isAG\" = 1 and %s IN(%s)" % (Field_del, ', '.join(map(str, ValueList)))
            arcpy.AddMessage("Subsetting fields to those with %s values of: %s" % (SubsetField, ValueList))

            # Make a copy of FBIn
            FB_copy = arcpy.CopyFeatures_management(FBIn, "in_memory" + "\\FB_copy")
            
            # Join SubsetField from runoff risk
            arcpy.JoinField_management(FB_copy, "FBndID", RunoffRiskTable, "FBndID", ["%s" % SubsetField])
            
            # Apply query slection to FB_copy
            lyr = arcpy.MakeFeatureLayer_management(FB_copy, "lyr", whereClause)
            AgFB = arcpy.CopyFeatures_management(lyr, "in_memory" + "\\AgFB")
            num_fields = int(arcpy.GetCount_management(AgFB).getOutput(0))
            if num_fields == 0:
                arcpy.AddError("Zero agricultural fields (non pasture) identified with %s value of %s" % (SubsetField, ValueList))
                sys.exit(0)

            arcpy.Delete_management(FB_copy)
            del[FB_copy]

    return(AgFB)


def GrassWaterwaySiting(D8FlowAcc, D8FlowDir, StreamReach, AreaThreshold, AgFB):

    # Convert user-defined area threshold from acres to meters squared: 1 acre = 4046 meters
    thresh_meters = float(AreaThreshold) * 4046

    # Convert threshold from meters squared to number of cells
    num_cells = float(thresh_meters / resolution)
    arcpy.AddMessage("Area threshold of %s acres...." % AreaThreshold)
    
    # Convert grid cells to 1 if higher than threshold and NODATA if not
    acc_rec = Con(D8FlowAcc, 1, "", "VALUE >= %s" % num_cells)

    arcpy.AddMessage("Removing stream reach")
    # Dissolve stream reach 
    strm_diss = arcpy.Dissolve_management(StreamReach, "in_memory" + "\\strm_diss", "", "", "", "")

    # Convert to raster 
    desc = arcpy.Describe(strm_diss)
    OID = desc.OIDFieldName 
    strm_ras = arcpy.FeatureToRaster_conversion(strm_diss, "%s" % OID, "in_memory" + "\\strm_ras", cellsize) 

    # Convert stream reach grid cells to 1 and background values to NODATA 
    strm_con = Con(strm_ras, 1, "", "VALUE >= 0")

    # Expand by 1 cell
    strm_exp = Expand(strm_con, 1, 1)

    # Delete cells from acc_rec that correspond to strm_exp
    gw_ras = Con(IsNull(strm_exp), acc_rec, "")

    # Cleanup
    arcpy.Delete_management(strm_diss)
    arcpy.Delete_management(strm_ras)
    arcpy.Delete_management(strm_con)
    del[strm_diss, strm_ras, strm_con]
        
    # Clip gw_ras by ag_fields
    gw_clip = arcpy.Clip_management(gw_ras, "", "in_memory" + "\\gw_clip", AgFB, "", "ClippingGeometry")

    # Convert from raster to polyline using stream to feature
    gw_line = StreamToFeature(gw_clip, D8FlowDir, "gw_line", "NO_SIMPLIFY")

    # Trim dangles shorted than 10 meters
    arcpy.TrimLine_edit(gw_line, "10 meters", "DELETE_SHORT")

    # Convert back to raster - this is done to generate stream link WITHOUT ANY DANGLES 
    desc = arcpy.Describe(gw_line)
    OID = desc.OIDFieldName
    gw_to_ras = arcpy.FeatureToRaster_conversion(gw_line, "%s" % OID, "gw_to_ras", cellsize)

    # Define stream links
    gw_link = StreamLink(gw_to_ras, D8FlowDir)

    # Use the stream to feature tool in ArcGIS to convert from raster to vector
    gw_poly = StreamToFeature(gw_link, D8FlowDir, OutGrassWaterway, "NO_SIMPLIFY")

    # Cleanup
    arcpy.Delete_management(acc_rec)
    arcpy.Delete_management(strm_exp)
    arcpy.Delete_management(gw_ras)
    arcpy.Delete_management(gw_clip)
    arcpy.Delete_management(gw_line)
    arcpy.Delete_management(gw_to_ras)
    arcpy.Delete_management(gw_link)
    arcpy.Delete_management(AgFB)
    del[acc_rec, strm_exp, gw_ras, gw_clip, gw_line, gw_to_ras, gw_link, AgFB]
    arcpy.Delete_management("in_memory")


##------------------------------------------------------------------------------
##------------------------------------------------------------------------------
if __name__ == "__main__":
    
    # Define Parameters
    FBIn = arcpy.GetParameterAsText(0)
    # Test for a join on FBIn
    basename = arcpy.Describe(FBIn).baseName
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    D8FlowAcc = arcpy.GetParameterAsText(1)
    D8FlowDir = arcpy.GetParameterAsText(2)  
    StreamReach = arcpy.GetParameterAsText(3)
    AreaThreshold = arcpy.GetParameterAsText(4)
    RunoffRiskTable = arcpy.GetParameterAsText(5)
    if RunoffRiskTable <> "":
        # Test for a join on RunoffRisk table
        basename = arcpy.Describe(RunoffRiskTable).baseName
        hasJoin = joinCheck(RunoffRiskTable)
        if hasJoin:
            arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
            sys.exit(0)      
    SubsetField = arcpy.GetParameterAsText(6)
    SubsetValue = arcpy.GetParameterAsText(7)
    OutGrassWaterway = arcpy.GetParameterAsText(8)

    # Set environments
    arcpy.env.extent = D8FlowAcc
    arcpy.env.snapRaster = D8FlowAcc
    rDesc = arcpy.Describe(D8FlowAcc)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True

    # Determine cellsize and resolution of the flow accumulation raster
    cellsize = float(arcpy.GetRasterProperties_management(D8FlowAcc, "CELLSIZEX").getOutput(0))
    resolution = float(cellsize * cellsize)

    # Run modules
    AgFB = FieldBoundarymask(FBIn, RunoffRiskTable, SubsetField, SubsetValue)
    
    GrassWaterwaySiting(D8FlowAcc, D8FlowDir, StreamReach, AreaThreshold, AgFB)

