# IMPOUNDMENT SITING - NUTRIENT REMOVAL WETLANDS
#
# NOTE: THIS TOOL WILL TAKE ANYWHERE FROM 15 minutes TO 1 HOUR OF PROCESSING TIME
# Times will increase with # of sample points 
#
# OBJECTIVE: The Impoundment Siting tool allows the user to sample locations along
# collective flow pathways for suitability of Nutrient Removal Wetlands 
#
# Testing locations are generated along collective flow pathways at a specified spacing.  Locations are
# sorted by contributing area and downstream sites are tested first.  At each location, an impoundment
# is simulated in the DEM, creating both a pooled area and a vegetated buffer above the impoundment.
# The drainage area to each impoundment is delineated, and descriptive statistics are generated, including the
# size of the pooled area, the size of the buffer, and the ratio of pooled area to the amount of drainage that it receives.
# If a site is found to be suitable according to predefined criteria, the site is kept and
# added to an output feature class.  The next site upstream of the just created wetland is then tested.
# If not suitable, the site is thrown out, and the next upstream site is tested.
#
# A predefined drainage area range of 60 HA ---> max watershed drainage is defined for Nutrient Removal Wetlands
# A polyline is created for that drainage range - sample points to be tested are created along this polyline.  If the stream
# reach is provided, points are not sampled that fall along the stream reach
#
# A reach file and catchments layer is generated for the drainage range (using TauDEM) and is used to create a dictionary that links downstream watersheds
# to upstream ones.  This greatly reduces the time needed to generated drainage areas to each wetland.  Dictionaries are used to link reaches to upstream reaches,
# and points to upstream points along reaches.  This way the watershed tool only needs to be run once, using all points, and dictionaries are used to merge the
# watershed of the current point to watersheds of upstream points along the same reach, and then with upstream catchments belonging to upstream reaches
#
# The last module in the tool calculates storage volumes (in acre feet) for all suitable Nutrient Removal Wetlands.
# Permanent storage volume is calculated by finding the storage volume of the pooled area
# Variable storage volume is calculated by finding the difference between the storage volume of the pooled area and the storage volume
# of the pooled area "at the height of the buffer", and adding this to the storage volume of the buffer itself
#
# The user has the option to vary the sampling distance, impoundment height, and buffer height
# within a predetermined range unique Nutrient Removal Wetlands 
#
# Optionally, the user can choose to avoid roads by providing a roads layer in the form of a polyline feature class.
# This will consider a site unsuitable if either the pooled area or the buffer area intersects the road layer provided.  
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")



def CreateDrainageNetwork(D8FlowAcc):
    arcpy.AddMessage("Testing watershed for nutrient removal wetland locations")

    # Convert accumulation raster from meters squared to HA
    D8Acc_HA = ((Times(D8FlowAcc, (cellsize * cellsize))) / 10000)
    
    MinDrainage = 60
    arcpy.AddMessage("Creating drainage network to be sampled for wetland suitability: >= %s HA" % MinDrainage)
    flownet = Con(D8Acc_HA, 1, "", "VALUE >= %s" % (MinDrainage))

    return(D8Acc_HA, flownet)


def CatchmentAndReachDelineation(D8FlowDir, DEM, flownet):
    env.compression = "NONE"

    # Fill the DEM
    DEMFill = Fill(DEM)

    # Reclassify D8FlowDir to TauDEM values ----> save to TDProcDir  as uncompressed tiff
    TDFlowDir = Reclassify(D8FlowDir, "VALUE",
                           RemapRange([[1, 1], [128, 2], [64, 3], [32, 4], [16, 5], [8, 6], [4, 7], [2, 8]]))

    # Copy 1)TDFlowDir, 2)DEMFill, and 3)flownet_ras to TDProcDir as uncompressed tiffs 
    arcpy.CopyRaster_management(TDFlowDir, TDProcDir + "\\TDflowdir.tif")
    arcpy.CopyRaster_management(DEMFill, TDProcDir + "\\TDfill.tif")
    arcpy.CopyRaster_management(flownet, TDProcDir + "\\TDflownet.tif")

    core_cnt = multiprocessing.cpu_count()
    core_no = core_cnt - 1
    
    # Create a TauDEM Flow Accumulation raster
    callstr = 'mpiexec -n %s AreaD8 ' % core_no +\
              ' -p ' + TDProcDir + "\\TDflowdir.tif" + \
              ' -ad8 ' + TDProcDir + "\\TDflow_acc.tif -nc"
    call(callstr, shell=True)
    arcpy.CalculateStatistics_management(TDProcDir + "\\TDflow_acc.tif")

    arcpy.AddMessage("Creating reach file and catchments from drainage network")
    # Construct command for Stream Reach, Watersheds, etc.
    callstr = 'mpiexec -n %s StreamNet ' % core_no + \
              ' -fel ' + TDProcDir + "\\TDfill.tif" + \
              ' -p ' + TDProcDir + "\\TDflowdir.tif" + \
              ' -ad8 ' + TDProcDir + "\\TDflow_acc.tif" + \
              ' -src ' + TDProcDir + "\\TDflownet.tif" + \
              ' -ord ' + TDProcDir + "\\TDstreamorder.tif" + \
              ' -tree ' + TDProcDir + "\\TDnetworktree.txt" + \
              ' -coord ' + TDProcDir + "\\TDnetworkcoord.txt" + \
              ' -net ' + TDProcDir + "\\TDreach_file.shp" + \
              ' -w ' + TDProcDir + "\\TDcatchments.tif"
    call(callstr, shell=True)

    # Calculate statistics 
    reach_file = arcpy.CalculateStatistics_management(TDProcDir + "\\TDreach_file.shp")
    catchments = arcpy.CalculateStatistics_management(TDProcDir + "\\TDcatchments.tif")

    # Build attribute table for catchments
    arcpy.BuildRasterAttributeTable_management(catchments)

    # Define projection for stream reach --> same as catchments
    projection = arcpy.Describe(catchments).spatialReference
    arcpy.DefineProjection_management(reach_file, projection)

    # Cleanup
    arcpy.Delete_management(TDFlowDir)
    arcpy.Delete_management(flownet)
    arcpy.Delete_management(DEMFill)
    del[TDFlowDir, flownet, DEMFill]

    # Clear compression environment
    arcpy.ClearEnvironment("compression")

    return(reach_file, catchments)


def CreateOutputFiles(WBD):
    # Create feature class to which to add suitable polygons
    NRW_lyr = arcpy.CreateFeatureclass_management("in_memory", "NRW_lyr", "POLYGON", "", "", "", WBD)
    NRW = arcpy.CopyFeatures_management(NRW_lyr, OutNRW)

    # Add identifier fields - will be populated as wetlands are appended
    arcpy.AddField_management(NRW, "SiteID", "LONG")
    arcpy.AddField_management(NRW, "CoverType", "TEXT")
    arcpy.AddField_management(NRW, "ContAreaHA", "FLOAT")
    arcpy.AddField_management(NRW, "PoolAreaHA", "FLOAT")
    arcpy.AddField_management(NRW, "BuffAreaHA", "FLOAT")
    arcpy.AddField_management(NRW, "StrmElev", "FLOAT")
    arcpy.AddField_management(NRW, "BankHeight", "FLOAT")
    arcpy.AddField_management(NRW, "BankElev", "FLOAT")
    
    # Create feature class to which to add suitable drainage area polygons
    DA_lyr = arcpy.CreateFeatureclass_management("in_memory", "DA_lyr", "POLYGON", "", "", "", WBD)
    NRW_DA = arcpy.CopyFeatures_management(DA_lyr, OutNRWDrainageAreas)

    # Add identifier fields - will be populated as wetland drainage areas are appended
    arcpy.AddField_management(NRW_DA, "SiteID", "LONG")

    # Cleanup
    arcpy.Delete_management(NRW_lyr)
    arcpy.Delete_management(DA_lyr)
    del[NRW_lyr, DA_lyr]
    arcpy.Delete_management("in_memory")

    return(NRW, NRW_DA)
    

def CreatePointsAlongDrainageNetwork(reach_file, spacing):
    arcpy.AddMessage("Creating sample points along drainage network at %s meter spacing" % spacing)
    
    # CREATE POINTS ALONG THE STREAM NETWORK (distance defined by spacing distance)
    create_pts = arcpy.CreateFeatureclass_management("in_memory", "create_pts", "Point", "", "", "", reach_file)

    # Identify the geometry field
    desc = arcpy.Describe(reach_file)
    shapefieldname = desc.ShapeFieldName
    
    # Create search cursor
    rows = arcpy.SearchCursor(reach_file)
    row = rows.next()
    
    iCurPts = arcpy.InsertCursor(create_pts)
    # Enter while loop for each feature/row
    while row:
        # Create the geometry object
        feat = row.getValue(shapefieldname)
        partnum = 0
        partcount = feat.partCount

        # Enter while loop for each part in the feature (if a singlepart feature this will occur only once)
        while partnum < partcount:
            part = feat.getPart(partnum)
            pnt = part.next()
            pntcount = 0

            ## statnPtsSpacing is the distance between Station Points along the line
            statnPtsSpacing = float('%s' % spacing)
            ## statnSctnLeftovers is the unallocated distance between Station Points along the line when you move to a new part
            statnSctnLeftovers = 0.00
                
            newPt = arcpy.CreateObject("Point")
            # Enter while loop for each vertex          
            while pnt:
                if pntcount > 0:
                    ## Calculate the distance (hypotenuse) and angle between previous and current point
                    pnt_dx = pnt.X - prevPtX
                    pnt_dy = pnt.Y - prevPtY
                    hypotenuse = math.sqrt(pnt_dx**2 + pnt_dy**2)
                    angle = math.atan2(pnt_dy, pnt_dx)
                    perp = angle + math.pi
                    
                    statnPtsInSctn = (hypotenuse-(statnSctnLeftovers * statnPtsSpacing))/statnPtsSpacing + 1
                    statnPtsInSctnInt = int(statnPtsInSctn)
                    
                    ## Calculate statnPtsSpacing distance between starting point and next point
                    if angle >= -math.pi/2 and angle < math.pi/2: ## -90 to 90 degrees
                        statnPtDX = statnPtsSpacing*math.cos(angle)
                        statnPtDY = statnPtsSpacing*math.sin(angle)
                    if angle >= 0 and angle < math.pi/2: ## 0 to 90 degrees
                        statnPtDX = statnPtsSpacing*math.cos(angle)
                        statnPtDY = statnPtsSpacing*math.sin(angle)
                    if angle < 0 and angle >= -math.pi/2: ## 0 to -90 degrees
                        statnPtDX = statnPtsSpacing*math.cos(angle)
                        statnPtDY = statnPtsSpacing*math.sin(angle)
                    if angle >= math.pi/2 and angle <= math.pi: ## 90 to 180 degrees
                        statnPtDX = -statnPtsSpacing*math.cos(math.pi-angle)
                        statnPtDY = statnPtsSpacing*math.sin(math.pi-angle)
                    if angle < -math.pi/2 and angle >= -math.pi: ## -90 to -180 degrees
                        statnPtDX = -statnPtsSpacing*math.cos(math.pi-angle)
                        statnPtDY = statnPtsSpacing*math.sin(math.pi-angle)
                        
                    ## Calculate for all sections
                    for i in range(0, statnPtsInSctnInt):
                    ## Set point X & Y
                        newPt.X = prevPtX + (statnSctnLeftovers + i)*statnPtDX
                        newPt.Y = prevPtY + (statnSctnLeftovers + i)*statnPtDY
                        iRowPts = iCurPts.newRow()
                        iRowPts.Shape = newPt
                        iCurPts.insertRow(iRowPts)

                    ## Calculate how much distance was left between last station point in line part and next point in line feature
                    statnSctnRemainingFraction = (hypotenuse-(statnSctnLeftovers * statnPtsSpacing))%statnPtsSpacing/statnPtsSpacing
                    ## Print "Section Remaining Fraction = " + str(statnSctnRemainingFraction)
                    ## Calculate how much distance beetween station points was left over
                    statnSctnLeftovers = 1-statnSctnRemainingFraction 
                    ## Print "Section Leftovers = " + str(statnSctnLeftovers)

                prevPtX = pnt.X
                prevPtY = pnt.Y
                pnt = part.next()
                pntcount += 1
            partnum += 1        
        row = rows.next()

    return(create_pts)


def WetlandLoop(WBD, create_pts, D8Acc_HA, D8FlowDir, DEM, reach_file, catchments, WetlandImpoundmentHeight, WetlandBufferHeight, NRW, NRW_DA, ZFactor):

    # Clip create_pts by WBD so as not to sample outside of WBD ---> snap to D8Acc_HA (within 3 meters --> 1 grid cell) and convert back to points (for attribution)
    clip_pts = arcpy.Clip_analysis(create_pts, WBD, "clip_pts", "")
    snap_ras = SnapPourPoint(clip_pts, D8Acc_HA, 3)
    snap_pts = arcpy.RasterToPoint_conversion(snap_ras, "in_memory" + "\\snap_pts")

    arcpy.AddMessage("Attributing points with site specific information")
    
    # Zonal stats with D8Acc_HA (Contributing Area in HA) 
    accstats = ZonalStatisticsAsTable(snap_pts, "pointid", D8Acc_HA, "in_memory" + "\\accstats", "", "MEAN")
    arcpy.JoinField_management(snap_pts, "pointid", accstats, "pointid", ["MEAN"])
    arcpy.AddField_management(snap_pts, "ContAreaHA", "FLOAT")
    arcpy.CalculateField_management(snap_pts, "ContAreaHA", "!MEAN!", "PYTHON")
    arcpy.DeleteField_management(snap_pts, ["MEAN"])

    # Convert reach_file to a raster based on unique LINKNO - extract LINKNO to snap_pts
    reach_ras = arcpy.PolylineToRaster_conversion(reach_file, "LINKNO", "reach_ras", "", "", cellsize)
    ExtractMultiValuesToPoints(snap_pts, [[reach_ras, "Reach"]], "BILINEAR")

    # Buffer snap_pts by 20 meters ---> run zonal stats with DEM (min, max, and range) - join back to snap_pts
    pts_buff = arcpy.Buffer_analysis(snap_pts, "in_memory" + "\\pts_buff", "20 meters", "FULL")
    pts_zstats = ZonalStatisticsAsTable(pts_buff, "pointid", DEM, "in_memory" + "\\pts_zstats", "", "ALL")
    arcpy.JoinField_management(snap_pts, "pointid", pts_zstats, "pointid", ["MIN", "RANGE", "MAX"])

    # Add and populate "StrmElev" - MIN, "BankHeight" - RANGE, and "BankElev" - MAX, fields
    arcpy.AddField_management(snap_pts, "StrmElev", "FLOAT")
    arcpy.CalculateField_management(snap_pts, "StrmElev", "!MIN!", "PYTHON")
    arcpy.AddField_management(snap_pts, "BankHeight", "FLOAT")
    arcpy.CalculateField_management(snap_pts, "BankHeight", "!RANGE!", "PYTHON")
    arcpy.AddField_management(snap_pts, "BankElev", "FLOAT")
    arcpy.CalculateField_management(snap_pts, "BankElev", "!MAX!", "PYTHON")
    arcpy.DeleteField_management(snap_pts, ["MIN", "RANGE", "MAX"])

    # Sort by ContAreaHA (DESCENDING - larger contributing areas are first)
    sample_pts = arcpy.Sort_management(snap_pts, "in_memory" + "\\sample_pts", [["ContAreaHA", "DESCENDING"]])

    arcpy.AddMessage("Deleting points that have a bank height greater than 4 meters")
    # Delete points that have a contributing area  < MinDrainage or NULL, do not fall on a reach, or have a bank height > 4 meters or NULL
    MaxBankHeight = float(4 / float(ZFactor))
    arcpy.MakeFeatureLayer_management(sample_pts, "erase_pts")
    arcpy.SelectLayerByAttribute_management("erase_pts", "NEW_SELECTION", "\"Reach\" IS NULL OR \"ContAreaHA\" < %s OR \"ContAreaHA\" IS NULL OR \"BankHeight\" >= %s OR \"BankHeight\" IS NULL" % (60, MaxBankHeight))

    # If stream reach is provided, delete points that intersect the stream reach
    if StreamReach <> "":
        arcpy.AddMessage("Removing stream reach from analysis")
        arcpy.SelectLayerByLocation_management("erase_pts", "WITHIN_A_DISTANCE", StreamReach, "%s" % cellsize, "ADD_TO_SELECTION")
    arcpy.DeleteFeatures_management("erase_pts")

    # Give each point a unique ID as determined by its OBJECTID or FID field - label it "SiteID"
    desc = arcpy.Describe(sample_pts)
    idfield = (desc.fields[0].name)
    arcpy.AddField_management(sample_pts, "SiteID", "LONG")
    arcpy.CalculateField_management(sample_pts, "SiteID", '!%s!' % idfield, "PYTHON")

    # Run CreateUpstreamWatershedDictionary and CreateUpstreamPointDictionary modules - links EACH point to upstream points in its drainage area
    wshed_dict = CreateUpstreamWatershedDictionary(reach_file)
    point_dict = CreateUpstreamPointDictionary(sample_pts)

    # Define wetland height and buffer height in same elevation units as DEM
    ImpoundmentHeight = float(WetlandImpoundmentHeight) / float(ZFactor)
    BufferHeight = float(WetlandBufferHeight) / float(ZFactor) 

    # Convert catchments raster to a polygon - dissolve
    catch_to_poly = arcpy.RasterToPolygon_conversion(catchments, "in_memory" + "\\catch_to_poly", "NO_SIMPLIFY")
    catch_poly = arcpy.Dissolve_management(catch_to_poly, "catch_poly", ["GRIDCODE"])

    # Add field delimiters to GRIDCODE field in catch_poly
    catch_fieldType = arcpy.ListFields(catch_poly, "GRIDCODE")[0].type
    catch_del = arcpy.AddFieldDelimiters(arcpy.Describe(catch_poly).path, "GRIDCODE")

    # snap sample points ---> better for running watershed tool
    pts_snap = SnapPourPoint(sample_pts, D8Acc_HA, 0, "SiteID")

    # Delineate watersheds for all pts_snap ---> subcatchments
    subcatchments = Watershed(D8FlowDir, pts_snap, "Value")

    # Convert subcatchments raster to a polygon - dissolve
    subcatch_to_poly = arcpy.RasterToPolygon_conversion(subcatchments, "in_memory" + "\\subcatch_to_poly", "NO_SIMPLIFY")
    subcatch_poly = arcpy.Dissolve_management(subcatch_to_poly, "subcatch_poly", ["GRIDCODE"])

    # Add field delimiters to GRIDCODE field in subcatch_poly
    subcatch_fieldType = arcpy.ListFields(subcatch_poly, "GRIDCODE")[0].type
    subcatch_del = arcpy.AddFieldDelimiters(arcpy.Describe(subcatch_poly).path, "GRIDCODE")

    # Cleanup
    arcpy.Delete_management(clip_pts)
    arcpy.Delete_management(snap_ras)
    arcpy.Delete_management(snap_pts)
    arcpy.Delete_management(accstats)
    arcpy.Delete_management(reach_ras)
    arcpy.Delete_management(pts_buff)
    arcpy.Delete_management(pts_zstats)
    arcpy.Delete_management(create_pts)
    arcpy.Delete_management(catchments)
    arcpy.Delete_management(catch_to_poly)
    arcpy.Delete_management(pts_snap)
    arcpy.Delete_management(subcatchments)
    arcpy.Delete_management(subcatch_to_poly)
    arcpy.Delete_management(D8Acc_HA)
    arcpy.Delete_management(TDProcDir)
    del[clip_pts, snap_ras, snap_pts, accstats, reach_ras, pts_buff, pts_zstats, create_pts, catchments, catch_to_poly, pts_snap, subcatchments, subcatch_to_poly, D8Acc_HA]

    # Create a tracker to count how many points have been processed - a message will be given every 100 points processed
    point_count = int(arcpy.GetCount_management(sample_pts).getOutput(0))
    arcpy.AddMessage("Beginning loop: %s points generated to test for suitability" % point_count)
    point_tracker_intervals = [100, 200, 300, 400, 500, 600, 700, 800, 900, 1000, 1100, 1200, 1300, 1400, 1500, 1600, 1700, 1800, 1900, 2000, 2100, 2200, 2300, 2400, 2500, 2600, 2700, 2800, 2900, 3000]
    point_tracker = 1

    # Multiply input DEM by 1000 and convert to an integer -- requred for Reclassify tool in below loop, which will not work with float inputs
    DEM_1000 = Int(Times(DEM, 1000))
    Minimum = arcpy.GetRasterProperties_management(DEM_1000, "MINIMUM")
    Maximum = arcpy.GetRasterProperties_management(DEM_1000, "MAXIMUM")

    # Begin looping through each point in sample_pts
    with arcpy.da.SearchCursor(sample_pts, ["SiteID", "ContAreaHA", "StrmElev", "BankHeight", "Reach", "BankElev"]) as cursor:
        for row in cursor:
            
            # Determing how many points have been sampled
            point_tracker = int(point_tracker + 1)
            if point_tracker in point_tracker_intervals:
                arcpy.AddMessage("%s points processed" % point_tracker)

            # Define variables
            SiteID = int(float(row[0]))
            ContributingArea = row[1]
            ChannelElevation = row[2]
            BankHeight = row[3]
            Reach = int(float(row[4]))
            BankElevation = row[5]

            # Query wshed_dict to make feature layer from catch_poly of upstream CATCHMENTS
            up_reachlist = wshed_dict[Reach]
            reachcount = len(up_reachlist)
            if reachcount >= 1:
                if str(catch_fieldType) == 'String':
                    up_reachlist = ["'%s'" % up_reach for up_reach in up_reachlist]
                whereClause = "%s IN(%s)" % (catch_del, ', '.join(map(str, up_reachlist)))
                arcpy.MakeFeatureLayer_management(catch_poly, "up_catch", whereClause)

            # Query point_dict to make feature layer from subcatch_poly of upstream SUBCATCHMENTS
            up_pointlist = point_dict[SiteID]
            pointcount = len(up_pointlist)
            if pointcount >= 1:
                if str(subcatch_fieldType) == 'String':
                    up_pointlist = ["'%s'" % up_point for up_point in up_pointlist]
                whereClause = "%s IN(%s)" % (subcatch_del, ', '.join(map(str, up_pointlist)))
                arcpy.MakeFeatureLayer_management(subcatch_poly, "up_subcatch", whereClause)

            # Make feature layer from subcatchments_poly of CURRENT SUBCATCHMENT
            arcpy.MakeFeatureLayer_management(subcatch_poly, "curr_subcatch", "\"GRIDCODE\" = %s" % SiteID)

            # DEFINE WATERSHED FOR CURRENT POINT                                                      
            if pointcount >= 1 and reachcount >= 1:
                Wshed_poly = arcpy.Merge_management(["up_catch", "up_subcatch", "curr_subcatch"], "in_memory" + "\\Wshed_poly")
            elif pointcount < 1 and reachcount >= 1:
                Wshed_poly = arcpy.Merge_management(["up_catch", "curr_subcatch"], "in_memory" + "\\Wshed_poly")
            elif pointcount >= 1 and reachcount < 1:
                Wshed_poly = arcpy.Merge_management(["up_subcatch", "curr_subcatch"], "in_memory" + "\\Wshed_poly")
            else:
                Wshed_poly = arcpy.CopyFeatures_management("curr_subcatch", "in_memory" + "\\Wshed_poly")

            # Set Wshed_poly as environment mask and extent
            arcpy.env.mask = "in_memory" + "\\Wshed_poly"
            arcpy.env.extent = "in_memory" + "\\Wshed_poly"

            # Reclassify DEM_1000 into wetland and buffer - will only do this within the watershed of sample point because of env.mask and env.extent
            DEMRec = Reclassify(DEM_1000, "Value",
                                     RemapRange([[Minimum, Int((BankElevation + ImpoundmentHeight) * 1000), "1"],
                                                 [Int((BankElevation + ImpoundmentHeight) * 1000), Int((BankElevation + ImpoundmentHeight + BufferHeight) * 1000), "2"],\
                                                 [Int((BankElevation + ImpoundmentHeight + BufferHeight) * 1000), Maximum, "NODATA"]]))

            # Verify that the whole DEM is NOT NODATA
            if int(arcpy.GetRasterProperties_management(DEMRec, "ALLNODATA").getOutput(0)) == 0:
            
                # Convert raster to polygon and repair geometry
                structure_poly = arcpy.RasterToPolygon_conversion(DEMRec, "in_memory" + "\\structure_poly")
                arcpy.RepairGeometry_management(structure_poly)

                # Dissolve by grid code (grid code of 1 = wetland pool, 2 = buffer)
                structure = arcpy.Dissolve_management(structure_poly, "in_memory" + "\\structure", ["GRIDCODE"])

                # Verify that both a buffer and a pool poygon exist --> define area (in HA) of each
                if int(arcpy.GetCount_management(structure).getOutput(0)) >= 2:
                    for row in arcpy.da.SearchCursor(structure,["SHAPE@AREA"], "\"GRIDCODE\" = 1"):
                        PoolArea = (row[0] / 10000)
                    for row in arcpy.da.SearchCursor(structure,["SHAPE@AREA"], "\"GRIDCODE\" = 2"):
                        BufferArea = (row[0] / 10000)

                    # Test for suitability
                    if (((float((PoolArea / ContributingArea)* 100)) >= .5) and ((float((PoolArea / ContributingArea)* 100)) <= 2) and ((float(BufferArea / PoolArea)) <= 4)):
                        #Site is suitable
                        
                        if Roads <> '':
                            # Dissolve buffer and wetland into a single polygon
                            desc = arcpy.Describe(structure)
                            OID = desc.OIDFieldName 
                            structure_diss = arcpy.Dissolve_management(structure, "in_memory" + "\\structure_diss", "%s" % OID)

                            # Count the number of intersections with roads layer --> if more than 1, delete structure (i.e. wetland and buffer polyons)
                            arcpy.MakeFeatureLayer_management(structure_diss, "IntersectTest")
                            arcpy.SelectLayerByLocation_management("IntersectTest", "INTERSECT", Roads, "", "NEW_SELECTION")
                            road_count = int(arcpy.GetCount_management("IntersectTest").getOutput(0))
                            if road_count >= 1:
                                arcpy.Delete_management(structure)

                            # Cleanup
                            arcpy.Delete_management(structure_diss)
                            del[structure_diss]

                        if arcpy.Exists(structure):
                            # Either roads were not avoided or the wetland did not intersect a road - site is suitable - add fields to be populated
                            arcpy.AddField_management(structure, "SiteID")
                            arcpy.AddField_management(structure, "CoverType", "TEXT")
                            arcpy.AddField_management(structure, "ContAreaHA", "FLOAT")
                            arcpy.AddField_management(structure, "PoolAreaHA", "FLOAT")
                            arcpy.AddField_management(structure, "BuffAreaHA", "FLOAT")
                            arcpy.AddField_management(structure, "StrmElev", "FLOAT")
                            arcpy.AddField_management(structure, "BankHeight", "FLOAT")
                            arcpy.AddField_management(structure, "BankElev", "FLOAT")
                            
                            # Populate fields with variables
                            with arcpy.da.UpdateCursor(structure, ["GRIDCODE", "CoverType", "SiteID", "ContAreaHA", "PoolAreaHA", "BuffAreaHA", "StrmElev", "BankHeight", "BankElev"]) as cursor:
                                for row in cursor:
                                    row[2] = SiteID
                                    if row[0] == 1:
                                        row[1] = "Wetland"
                                    else:
                                        row[1] = "Buffer"
                                    row[3] = ContributingArea
                                    row[4] = PoolArea
                                    row[5] = BufferArea
                                    row[6] = ChannelElevation
                                    row[7] = BankHeight
                                    row[8] = BankElevation
                                    cursor.updateRow(row)

                            # Append polygons to NRW feature class
                            arcpy.Append_management(structure, NRW, "NO_TEST")

                            # Provide a unique identifier (SiteID) to Wshed_poly 
                            Wshed_diss = arcpy.Dissolve_management(Wshed_poly, "in_memory" + "\\Wshed_diss", "")
                            arcpy.AddField_management(Wshed_diss, "SiteID", "FLOAT")
                            arcpy.CalculateField_management(Wshed_diss, "SiteID", '%s' % SiteID, "PYTHON")

                            # Append drainage areas to NRWDrainageAreas feature class
                            arcpy.Append_management(Wshed_diss, NRW_DA, "NO_TEST")

                            # Delete points from sample_pts that fall "within" the NRW layer, except for the site currently being tested
                            arcpy.MakeFeatureLayer_management(sample_pts, "IntersectSites")
                            arcpy.SelectLayerByLocation_management("IntersectSites", "COMPLETELY_WITHIN", NRW, "", "NEW_SELECTION")
                            arcpy.SelectLayerByAttribute_management("IntersectSites", "SUBSET_SELECTION", '\"SiteID\" <> %s' % SiteID)
                            arcpy.DeleteFeatures_management("IntersectSites")

                            arcpy.Delete_management(Wshed_diss)
                            del[Wshed_diss]

                # Cleanup
                arcpy.Delete_management(structure_poly)
                arcpy.Delete_management(structure)
                del[structure_poly, structure]

            # Clear extent and mask environments
            arcpy.ClearEnvironment("extent")
            arcpy.ClearEnvironment("mask")

            # Cleanup
            arcpy.Delete_management(Wshed_poly)
            del[Wshed_poly, DEMRec]

    # Cleanup
    arcpy.Delete_management(catch_poly)
    arcpy.Delete_management(subcatch_poly)
    arcpy.Delete_management(sample_pts)
    arcpy.Delete_management(DEM_1000)
    del[catch_poly, subcatch_poly, sample_pts, DEM_1000]
    arcpy.Delete_management("in_memory")
        
    return(NRW, NRW_DA)


def CalculateStorageVolumes(NRW, NRW_DA, DEM, WetlandImpoundmentHeight, WetlandBufferHeight, ZFactor):
    
    impoundment_count = (int(arcpy.GetCount_management(NRW).getOutput(0)))
    if impoundment_count == 0:
        arcpy.AddMessage("no suitable Nutrient Removal Wetlands found....exiting")

        arcpy.Delete_management(NRW_DA)
        arcpy.Delete_management(NRW)
        del[NRW_DA, NRW]

    elif impoundment_count >= 1:
        count = int(impoundment_count) / 2
        arcpy.AddMessage("%s Suitable Nutrient Removal Wetlands found: calculating storage volumes" % (count))

        # Define wetland height and buffer height in same elevation units as DEM
        ImpoundmentHeight = float(WetlandImpoundmentHeight) / float(ZFactor)
        BufferHeight = float(WetlandBufferHeight) / float(ZFactor)
        BufferElevation = float(BufferHeight) + float(ImpoundmentHeight)

        # CALCULATE VOLUME OF STORAGE (in acre feet)
        # Create feature layer of all pool polygons (not buffer) 
        pool_lyr = arcpy.MakeFeatureLayer_management(NRW, "pool_lyr", "CoverType = 'Wetland'")
        pool = arcpy.CopyFeatures_management(pool_lyr, "pool")
        
        # Convert pool to raster based on BankElev field ---> each pixel in pool extent will be the top of bank elevation where impoundment was simulated
        BankElev = arcpy.PolygonToRaster_conversion(pool, "BankElev", "BankElev", "", "", cellsize)

        # Add impoundment height to BankElev - each pixel in pool extent is given the constant pool elevation
        PoolElev = Plus(BankElev, ImpoundmentHeight)      

        # If the DEM is subtracted from the raster of pool elevation, this will give us the volume for each pixel (in same units of DEM) of the pool storage 
        PoolDiff = Minus(PoolElev, DEM)
        PoolDiffRec = Con(PoolDiff, "0", PoolDiff, "Value < 0")
  
        # Convert volume to acrefeet --> convert to meters ---> cubic meters ---> acre feet
        PoolDifMet = Times(PoolDiffRec, float(ZFactor))
        PoolVolCM = Times(PoolDifMet, (cellsize * cellsize))
        PoolVolAF = Times(PoolVolCM, .000810713)

        # Zonal Stats
        PoolStat = ZonalStatisticsAsTable(pool, "SiteID", PoolVolAF, "in_memory" + "\\PoolStat", "#", "SUM")

        # Add to impoundment table and populate "PoolStorAf" (i.e pool storage in acre feet)
        arcpy.JoinField_management(NRW, "SiteID", PoolStat, "SiteID", ["SUM"])
        arcpy.AddField_management(NRW, "PoolStorAF", "FLOAT")
        arcpy.CalculateField_management(NRW, "PoolStorAF", "!SUM!", "PYTHON")
        arcpy.DeleteField_management(NRW, ["SUM"])

        # CALCULATE VOLUME OF STORAGE IN BUFFER (IN ACRE FEET)
        # Requires 2 steps: 1)Calculate storage volume "at buffer elevation" over pool extent, and ADD TO 2)Calculate storage volume "at buffer elevation" over buffer extent

        # STEP 1: 
        # Add BufferHeight to PoolElev (i.e. represents the buffer height over the pool extent)
        BufPExt_Elev = Plus(PoolElev, BufferHeight)

        # If the DEM is subtracted from the raster of buffer elevation over the pool exent, this will give the volume (in same units as DEM) of the buffer storage over the pool extent
        BufPExt_Diff = Minus(BufPExt_Elev, DEM)
        BufPExtRec = Con(BufPExt_Diff, "0", BufPExt_Diff, "Value < 0")

        # Convert volume to acrefeet --> convert to meters ---> cubic meters ---> acre feet
        BufPExtDifMet = Times(BufPExtRec, float(ZFactor))
        BufPExt_VolCM = Times(BufPExtDifMet, (cellsize * cellsize))
        BufPExt_VolAF = Times(BufPExt_VolCM, .000810713)

        # Zonal Stats
        BufPExt_Stat = ZonalStatisticsAsTable(pool, "SiteID", BufPExt_VolAF, "in_memory" + "\\BufPExt_Stat", "#", "SUM")

        # STEP 2:
        # Create feature layer of all buffer polygons (not pool)
        buff_lyr = arcpy.MakeFeatureLayer_management(NRW, "buff_lyr", "CoverType = 'Buffer'")
        buff = arcpy.CopyFeatures_management(buff_lyr, "buff")

        # Convert to raster based on BankElev field ---> each pixel in buffer extent will be the top of bank elevation where impoundment was simulated
        BankElevBuff = arcpy.PolygonToRaster_conversion(buff, "BankElev", "BankElevBuff", "", "", cellsize)

        # Add buffer elevation to BankElev - each pixel in buffer extent is given the constant buffer elevation
        BuffElev = Plus(BankElevBuff, BufferElevation)

        # If the DEM is subtracted from the raster representing constant buffer elevation, this will give us the volume (in cm) of the buffer storage over the buffer extent)
        BuffDiff = Minus(BuffElev, DEM)
        BuffDiffRec = Con(BuffDiff, "0", BuffDiff, "Value < 0")

        # Convert volume to acrefeet --> convert to meters ---> cubic meters ---> acre feet
        BuffDifMet = Times(BuffDiffRec, float(ZFactor))
        BuffVolCM = Times(BuffDifMet, (cellsize * cellsize))    
        BuffVolAF = Times(BuffVolCM, .000810713)

        # Zonal Stats
        BuffStat = ZonalStatisticsAsTable(buff, "SiteID", BuffVolAF, "in_memory" + "\\BuffStat", "#", "SUM")
        arcpy.JoinField_management(NRW, "SiteID", BufPExt_Stat, "SiteID", ["SUM"])
        arcpy.JoinField_management(NRW, "SiteID", BuffStat, "SiteID", ["SUM"])
        
        arcpy.AddField_management(NRW, "VarStorAF", "FLOAT")
        arcpy.CalculateField_management(NRW, "VarStorAF", "!SUM! - !PoolStorAF! + !SUM_1!", "PYTHON")

        arcpy.DeleteField_management(NRW, ["SUM", "SUM_1"])

        # Cleanup
        arcpy.Delete_management(pool)
	arcpy.Delete_management(BankElev)
	arcpy.Delete_management(PoolElev)
	arcpy.Delete_management(PoolDiff)
	arcpy.Delete_management(PoolDiffRec)
	arcpy.Delete_management(PoolDifMet)
	arcpy.Delete_management(PoolVolCM)
	arcpy.Delete_management(PoolVolAF)
	arcpy.Delete_management(PoolStat)
	arcpy.Delete_management(BufPExt_Elev)
	arcpy.Delete_management(BufPExt_Diff)
	arcpy.Delete_management(BufPExtRec)
	arcpy.Delete_management(BufPExtDifMet)
	arcpy.Delete_management(BufPExt_VolCM)
	arcpy.Delete_management(BufPExt_VolAF)
	arcpy.Delete_management(BufPExt_Stat)
	arcpy.Delete_management(buff)  
	arcpy.Delete_management(BankElevBuff)
	arcpy.Delete_management(BuffElev)
	arcpy.Delete_management(BuffDiff)
	arcpy.Delete_management(BuffDiffRec)
	arcpy.Delete_management(BuffDifMet)
	arcpy.Delete_management(BuffVolCM)
	arcpy.Delete_management(BuffVolAF)
	arcpy.Delete_management(BuffStat)
        del[pool, BankElev, PoolElev, PoolDiff, PoolDiffRec, PoolDifMet, PoolVolCM, PoolVolAF, PoolStat, BufPExt_Elev, BufPExt_Diff, BufPExtRec, BufPExtDifMet, \
            BufPExt_VolCM, BufPExt_VolAF, BufPExt_Stat, buff, BankElevBuff, BuffElev, BuffDiff, BuffDiffRec, BuffDifMet, BuffVolCM, BuffVolAF, BuffStat]
    arcpy.Delete_management("in_memory")



def CreateUpstreamWatershedDictionary(reach_file):
    arcpy.AddMessage("Creating dictionary to link reaches to upstream reaches")
    # CREATE A DICTIONARY ---> wshed_dict - will link each reach to upstream reaches; structure is: {LINKNO: upstream_reachlist}
    # this will enable watersheds to be genereated for all reaches at once (i.e. catchments)

    # Create 2 lists to contain headwaters and non-headwater reaches ---> if the magnitude field is 1 (magnitude refers to Shreve stream order), then the reach is a headwaters reach 
    HeadwaterReachList = []
    LowerOrderReachList = []
    for row in arcpy.da.SearchCursor(reach_file, ["LINKNO", "Magnitude"]):
        LINKNO = row[0]
        if row[1] == 1:
            HeadwaterReachList.append(LINKNO)
        elif row[1] > 1:
            LowerOrderReachList.append(LINKNO)
    del row

    # Create empty wshed_dict 
    wshed_dict = {}

    # Loop through each reach in LowerOrderReachList ----> use USLINKNO1 and USLINKNO2 to define the 2 upstream LINKNOs for each reach
    # Perform recursive looping (i.e. family tree style) to append each upstream LINKNO to the UpstreamReachList - repeat for those links..and so on
    # At the end of one loop (i.e. one reach) ----> create and update a reach_dict (i.e. {reach: UpstreamReachList} and append to the wshed_dict
    for reach in LowerOrderReachList:
        CurrentReach = arcpy.MakeFeatureLayer_management(reach_file, "CurrentReach", "\"LINKNO\" = %s" % reach)
        for row in arcpy.da.SearchCursor(CurrentReach, ["USLINKNO1", "USLINKNO2"]):
            UpLink1 = row[0]
            UpLink2 = row[1]
        
        UpstreamReachList = [UpLink1, UpLink2]
        for UpstreamReach in UpstreamReachList:
            UpReach = arcpy.MakeFeatureLayer_management(reach_file, "UpReach", "\"LINKNO\" = %s" % UpstreamReach)
            for row in arcpy.da.SearchCursor(UpReach, ["LINKNO", "USLINKNO1", "USLINKNO2"]):
                UpLink1 = row[1]
                UpLink2 = row[2]
                if UpLink1 <> -1 and UpLink2 <> -1:
                    UpstreamReachList.append(UpLink1)
                    UpstreamReachList.append(UpLink2)
                        
        reach_dict = {int(reach): UpstreamReachList}
        wshed_dict.update(reach_dict)

    # Loop through each reach in HeadwaterReachList ----> there is no USLINKNO1 and USLINKNO2
    # At the end of one loop (i.e. one reach) ----> update the reach dict and append to the wshed_dict
    for reach in HeadwaterReachList:
        UpstreamReachList = []
        reach_dict = {int(reach):UpstreamReachList}
        wshed_dict.update(reach_dict)

    arcpy.Delete_management(reach_file)
    del[reach_file]

    return(wshed_dict)


def CreateUpstreamPointDictionary(sample_pts):
    arcpy.AddMessage("Creating dictionary to link sample points to upstream subcatchments")
    # CREATE A DICTIONARY ---> point_dict - will link each point to upstream points along the same reach; structure is: {pointid: up_point_list}
    # this will enable watersheds to be generated for all points at once (i.e. subcatchments)
    # For each point, if another point is along the same reach with a Contributing Area less than the current point, append to list

    # Create empty dict 
    point_dict = {}

    # Loop through each point in sample_pts
    for row in arcpy.da.SearchCursor(sample_pts, ["SiteID", "ContAreaHA", "Reach"]):
        row_id = row[0]
        up_point_list = []
        for point in arcpy.da.SearchCursor(sample_pts, ["SiteID", "ContAreaHA", "Reach"]):
            point_ID = point[0]
            if point[1] < row[1] and point[2] == row[2]:
                up_point_list.append(point_ID)
        reach_dict = {int(row_id): up_point_list}
        point_dict.update(reach_dict)
    del row

    return(point_dict)


if __name__ == "__main__":

    # Define Parameters
    DEM = arcpy.GetParameterAsText(0)
    D8FlowDir = arcpy.GetParameterAsText(1)
    D8FlowAcc = arcpy.GetParameterAsText(2)
    WBD = arcpy.GetParameterAsText(3)
    StreamReach = arcpy.GetParameterAsText(4)
    Roads = arcpy.GetParameterAsText(5)
    ZFactor = arcpy.GetParameterAsText(6)
    spacing = arcpy.GetParameterAsText(7)
    WetlandImpoundmentHeight = arcpy.GetParameterAsText(8)
    WetlandBufferHeight = arcpy.GetParameterAsText(9)
    OutNRW = arcpy.GetParameterAsText(10)
    OutNRWDrainageAreas = arcpy.GetParameterAsText(11)

    # Set extent, snap, workspace
    arcpy.env.extent = DEM
    arcpy.env.snapRaster = DEM
    arcpy.env.outputCoordinateSystem = DEM
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder
    
    # Determine the cellsize of the input DEM
    cellsize = float(arcpy.GetRasterProperties_management(DEM, "CELLSIZEX").getOutput(0))
    
    # Create and define folder for TauDEM outputs (in ProcDir) - will be deleted at end of script
    scriptpath = sys.path[0]
    if ' ' in scriptpath:
        arcpy.AddError("space exists in %s that must be removed from path name...QUITTING!!!" % scriptpath)
        sys.exit(0)
    else:
        arcpy.CreateFolder_management(scriptpath, "TDProcDir")
        TDProcDir = scriptpath + "\\TDProcDir"

    # Run Modules
    D8Acc_HA, flownet = CreateDrainageNetwork(D8FlowAcc)
    
    reach_file, catchments = CatchmentAndReachDelineation(D8FlowDir, DEM, flownet)

    NRW, NRW_DA = CreateOutputFiles(WBD)
    
    create_pts = CreatePointsAlongDrainageNetwork(reach_file, spacing)
    
    NRW, NRW_DA = WetlandLoop(WBD, create_pts, D8Acc_HA, D8FlowDir, DEM, reach_file, catchments, WetlandImpoundmentHeight, WetlandBufferHeight, NRW, NRW_DA, ZFactor)

    CalculateStorageVolumes(NRW, NRW_DA, DEM, WetlandImpoundmentHeight, WetlandBufferHeight, ZFactor)


                

        
