# Peuker Douglas Stream Definition--------------------------------------------------------------------------------------------------
#
# Defines a stream network using the Peuker Douglas Stream Definition
# Required inputs include:
#   2 terrain rasters: a flow direction raster and a filled DEM
#   1 feature class: NHD-derived watershed boundary (WBD)
# Both terrain inputs were derived from ArcGIS processing and will be converted to uncompressed .tifs for use with TauDEM software

# Output will be a stream network "PDFlowNetwork" + inHUC
# For use in the ACPF, the next step will be manual revision of the "StreamType" field of the flow network 
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")


# Modules------------------------------------------------------------------------

def CreatePeukerDouglasInputs(WBD, DEMFill, D8FlowDir):
    arcpy.AddMessage("Creating Peuker Douglas Inputs")

    # Reclassify D8FlowDir raster to TauDEM flow direction values
    TDFlowDirRec = Reclassify(D8FlowDir, "VALUE",
                           RemapRange([[1, 1], [128, 2], [64, 3], [32, 4], [16, 5], [8, 6], [4, 7], [2, 8]]))

    # Copy TDFlowDirRec and DEMFill rasters to TDProcDir as uncompressed tiffs (set in env settings)
    TDFlowDir = arcpy.CopyRaster_management(TDFlowDirRec, TDProcDir + "\\TDFlowDir.tif")
    TDFill = arcpy.CopyRaster_management(DEMFill, TDProcDir + "\\TDFill.tif")

    # Calculate TauDEM D8 flow accumulation
    core_cnt = multiprocessing.cpu_count()
    core_no = core_cnt - 1
    
    callstr = 'mpiexec -n %s AreaD8 ' % core_no + \
              ' -p ' + TDProcDir + "\\TDFlowDir.tif" + \
              ' -ad8 ' + TDProcDir + "\\TDFlowAccumulation.tif -nc"
    call(callstr, shell=True)
    TDFlowAccumulation = arcpy.CalculateStatistics_management(TDProcDir + "\\TDFlowAccumulation.tif")

    # FIND POUR POINTS
    if PourPointFile <> '':
        arcpy.AddMessage("Pour Points provided")
        
        # Select pour points within 100 meters of the watershed boundary ---> count
	arcpy.MakeFeatureLayer_management(PourPointFile, "PPts_lyr")
	arcpy.SelectLayerByLocation_management("PPts_lyr","WITHIN_A_DISTANCE", WBD, "100 Meters","NEW_SELECTION")
	pointCount = int(arcpy.GetCount_management("PPts_lyr").getOutput(0))

	# If there is at least 1 pour point within 100 m of WBD ---> continue; if not---> give error message and quit
	if pointCount > 0:
		arcpy.AddMessage("PourPoints: " + str(pointCount))
		arcpy.CopyFeatures_management("PPts_lyr", TDProcDir + "\\PourPts.shp")
        else:
        	arcpy.AddError("No points found within 100m of watershed boundary...QUITTING!")
        	sys.exit(0)

    elif PourPointFile == '':
        # Create pour points automatically
        arcpy.AddMessage("Finding Pour Points")
        
        # Convert WBD to line - then to raster (will contain grid cells along the edge of WBD boundary) --> Extract flow accumulation values to each grid cell 
        WBDLine = arcpy.FeatureToLine_management(WBD, "in_memory" + "\\WBDLine")

        desc = arcpy.Describe(WBDLine)
        OID = desc.OIDFieldName
        WBDRaster = arcpy.PolylineToRaster_conversion(WBDLine, "%s" % OID, "in_memory" + "\\WBDRaster", "", "", cellsize)

        # Extract flow accumulation values to just those grid cells along border of WBD
        WBDAccum = ExtractByMask(TDFlowAccumulation, WBDRaster)

        # Find mean and standard deviation of accumulation values along WBD boundary
        WBDAccumMean = WBDAccum.mean
        WBDAccumStdDev4x = WBDAccum.standardDeviation * 4

        # Define threshold -> values above this will become pour points
        threshold = int(WBDAccumMean + WBDAccumStdDev4x)
            
        # Extract grid cells as those that are > than threshold
        PourPointsRas = Con(WBDAccum >= threshold, WBDAccum)

        # Convert to points
        PourPointsTmp = arcpy.RasterToPoint_conversion(PourPointsRas, TDProcDir + "\\PourPts.shp")
        if OutPourPoint <> '':
            PourPoints = arcpy.CopyFeatures_management(PourPointsTmp, OutPourPoint)

        arcpy.Delete_management(WBDLine)
        arcpy.Delete_management(WBDRaster)
        arcpy.Delete_management(WBDAccum)
        arcpy.Delete_management(PourPointsRas)
        del[WBDLine, WBDRaster, WBDAccum, PourPointsRas]
        arcpy.Delete_management("in_memory")

    arcpy.Delete_management(TDFlowDirRec)
    del[TDFlowDirRec]
    

def PeukerDouglasFlowNetwork(TDProcDir, D8FlowDir):
    
    core_cnt = multiprocessing.cpu_count()
    core_no = core_cnt - 1
    
    arcpy.AddMessage("Creating Peuker Douglas Flow Network")
    # Stream Source Grid
    callstr = 'mpiexec -n %s PeukerDouglas ' % core_no + \
              ' -fel ' + TDProcDir + "\\TDFill.tif" + \
              ' -ss ' + TDProcDir + "\\StreamSourceGrid.tif"
    call(callstr, shell=True)
    StreamSourceGrid = arcpy.CalculateStatistics_management(TDProcDir + "\\StreamSourceGrid.tif")

    # Flow Accumulation with Weight Raster
    callstr = 'mpiexec -n %s Aread8 ' % core_no +\
              ' -p ' + TDProcDir + "\\TDFlowDir.tif" + \
              ' -o ' + TDProcDir + "\\PourPts.shp" + \
              ' -ad8 ' + TDProcDir + "\\AccumulatedStreamSourceGrid.tif" + \
              ' -wg ' + TDProcDir + "\\StreamSourceGrid.tif -nc"
    call(callstr, shell=True)
    AccumulatedStreamSourceGrid = arcpy.CalculateStatistics_management(TDProcDir + "\\AccumulatedStreamSourceGrid.tif")

    # DROP ANALYSIS
    callstr = ' mpiexec -n %s Dropanalysis ' % core_no + \
              ' -p ' + TDProcDir + "\\TDFlowDir.tif" + \
              ' -fel ' + TDProcDir + "\\TDFill.tif" + \
              ' -ad8 ' + TDProcDir + "\\TDFlowAccumulation.tif" + \
              ' -ssa ' + TDProcDir + "\\AccumulatedStreamSourceGrid.tif" + \
              ' -drp ' + TDProcDir + "\\demdrp.txt " + \
              ' -o ' + TDProcDir + "\\PourPts.shp -par 500 100000 50 0"
    call(callstr, shell=True)

    # get drop initiation threshold 
    chThresh = getThresh(TDProcDir)

    # Generate Peuker Douglas flow network using threshold
    callstr = 'mpiexec -n %s Threshold ' % core_no +\
              ' -ssa ' + TDProcDir + "\\AccumulatedStreamSourceGrid.tif" + \
              ' -src ' + TDProcDir + "\\PDStreamNetRas.tif" + \
              ' -thresh ' + str(chThresh)
    call(callstr, shell=True)
    PDStreamNetRas = arcpy.CalculateStatistics_management(TDProcDir + "\\PDStreamNetRas.tif")

    # Convert background  values to null
    FlowNetRas = Con(PDStreamNetRas, 1, "", "VALUE = 1")

    # Calculate Stream Order
    Order = StreamOrder(FlowNetRas, D8FlowDir, "STRAHLER")

    # Use the stream to feature tool to convert from raster to vector
    FlowNetwork = StreamToFeature(Order, D8FlowDir, OutPDFlowNet, "NO_SIMPLIFY")

    # Add Stream field
    arcpy.AddField_management(FlowNetwork, "StreamType", "SHORT")
    arcpy.CalculateField_management(FlowNetwork, "StreamType","0","PYTHON","#")

    # Add Stream Order field
    arcpy.AddField_management(FlowNetwork, "STRAHLOrd", "LONG")
    arcpy.CalculateField_management(FlowNetwork, "STRAHLOrd", "!grid_code!", "PYTHON")
    arcpy.DeleteField_management(FlowNetwork, ["grid_code"])

    # Cleanup
    arcpy.Delete_management(TDProcDir)
    arcpy.Delete_management(Order)
    arcpy.Delete_management(FlowNetRas)
    del[FlowNetRas, Order]


def getThresh(TDProcDir):
    inf = open(TDProcDir + "\\demdrp.txt")
    lineList = inf.readlines()
    last = lineList[len(lineList)-1].split(':')
    chThresh = int(float(last[1]))
    arcpy.AddMessage("Constant stream drop threshold of %s grid cells" % chThresh)
    
    if chThresh == 0:
	chThresh = 5000

    if OutDropFile <> '':
        dropfile = TDProcDir + "\\demdrp.txt"
        drop_copy = arcpy.TableToTable_conversion(dropfile, "in_memory", "dropfile")
        dfile = arcpy.CopyRows_management(drop_copy, OutDropFile)

    return(chThresh)
    


##------------------------------------------------------------------------------
##------------------------------------------------------------------------------
if __name__ == "__main__":

    # Define Parameters
    DEMFill = arcpy.GetParameterAsText(0)
    D8FlowDir = arcpy.GetParameterAsText(1)
    WBD = arcpy.GetParameterAsText(2)
    PourPointFile = arcpy.GetParameterAsText(3)
    OutPDFlowNet = arcpy.GetParameterAsText(4)
    OutPourPoint = arcpy.GetParameterAsText(5)
    OutDropFile = arcpy.GetParameterAsText(6)

    # Set extent, snap 
    arcpy.env.extent = DEMFill
    arcpy.env.snapRaster = DEMFill
    arcpy.env.outputCoordinateSystem = DEMFill
    rDesc = arcpy.Describe(DEMFill)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.compression = "NONE"
    arcpy.env.scratchWorkspace = env.scratchFolder    

    # Determine the resolution of the input data
    cellsize = float(arcpy.GetRasterProperties_management(DEMFill, "CELLSIZEX").getOutput(0))

    # Create and define folder for TauDEM outputs (in ProcDir) - will be deleted at end of script
    scriptpath = sys.path[0]
    if ' ' in scriptpath:
        arcpy.AddError("space exists in %s that must be removed from path name...QUITTING!!!" % scriptpath)
        sys.exit(0)
    else:
        arcpy.CreateFolder_management(scriptpath, "TDProcDir")
        TDProcDir = scriptpath + "\\TDProcDir"

    # Run Modules
    CreatePeukerDouglasInputs(WBD, DEMFill, D8FlowDir)
    
    PeukerDouglasFlowNetwork(TDProcDir, D8FlowDir)
