# MANUAL CUTTER--------------------------------------------------------------------------------------------------
#
# OBJECTIVE: Modifies DEM by "burning" in manually created cut and/or dam lines
#
# Elevation values along each cut line will be replaced with the minimum elevation value along the line.
# Elevation values along each dam line will be replaced with the maximum elevation value along the line.
# All other elevation values will remain the same
# Output will be a new DEM called "NewDEM" - this will be used as the input DEM for terrain analysis
#
# The user will provide a feature class containing cut lines or dam lines
# This feature class will be manually created by the user as a polyline feature class
# Can contain multiple cut and/or dam lines to be processed at once
#
# Terrain processing will automatically be rerun using the NewDEM
# Output layers will replace any D8FlowDir, DEMFill, D8FlowAcc, and Hillshade rasters
#
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 05/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
arcpy.CheckOutExtension("Spatial")


def ManualCutter(CutLines, DEM):
    
    if CutLines == "":
        if DamLines == "":
            arcpy.AddError("Must provide either cut lines OR dam lines OR both...QUITTING!!!")
            sys.exit(0)

    if CutLines <> "" and DamLines <> "":
        
        # Get count of lines to process
        numCut = int(arcpy.GetCount_management(CutLines).getOutput(0))
        numDam = int(arcpy.GetCount_management(DamLines).getOutput(0))
        arcpy.AddMessage("Processing %s Cut features and %s Dam features"  %(numCut, numDam))

        # Convert cut lines to raster
        desc = arcpy.Describe(CutLines)
        OID = desc.OIDFieldName
        cut_ras = arcpy.PolylineToRaster_conversion(CutLines, "%s" % OID, "in_memory" + "\\cut_ras", "", "", cellsize)

        # Convert dam lines to raster
        desc = arcpy.Describe(DamLines)
        OID = desc.OIDFieldName
        dam_ras = arcpy.PolylineToRaster_conversion(DamLines, "%s" % OID, "in_memory" + "\\dam_ras", "", "", cellsize)

        # Perform zonal statistics on the DEM with cut_line_ras
        cut_ext = ZonalStatistics(cut_ras, "VALUE", DEM, "MINIMUM")

        # Perform zonal statistics on the DEM with dam_line_ras
        dam_ext = ZonalStatistics(dam_ras, "VALUE", DEM, "MAXIMUM")

        # 2 Con operations; 1) burn cut lines into DEM 2) burn dam lines into output from step 1
        NewDEM_cut = Con(IsNull(cut_ext), DEM, cut_ext)
        
        NewDEM_dam = Con(IsNull(dam_ext), NewDEM_cut, dam_ext)

        # Output to file location specified by user
        arcpy.AddMessage("New DEM created with altered elevation values")
        arcpy.CopyRaster_management(NewDEM_dam, NewDEM)

        # Build pyramids
        arcpy.BuildPyramids_management(NewDEM)

        arcpy.Delete_management(cut_ras)
        arcpy.Delete_management(dam_ras)
        arcpy.Delete_management(cut_ext)
        arcpy.Delete_management(dam_ext)
        arcpy.Delete_management(NewDEM_cut)
        arcpy.Delete_management(NewDEM_dam)
        del[cut_ras, dam_ras, cut_ext, dam_ext, NewDEM_cut, NewDEM_dam]
        arcpy.Delete_management("in_memory")

    if CutLines == "" and DamLines <> "":
        
        # Get count of # of lines to process
        numCut = int(arcpy.GetCount_management(DamLines).getOutput(0))
        arcpy.AddMessage("Processing %s Dam features"  %(numCut))

        # Convert dam lines to raster
        desc = arcpy.Describe(DamLines)
        OID = desc.OIDFieldName
        dam_ras = arcpy.PolylineToRaster_conversion(DamLines, "%s" % OID, "in_memory" + "\\dam_ras", "", "", cellsize)

        # Perform zonal statistics on the DEM with dam_line_ras
        dam_ext = ZonalStatistics(dam_ras, "VALUE", DEM, "MAXIMUM")

        # Con operation to burn dam_line_ext into DEM
        NewDEM_dam = Con(IsNull(dam_ext), DEM, dam_ext)

        # Output to file location specified by user
        arcpy.AddMessage("New DEM created with altered elevation values")
        arcpy.CopyRaster_management(NewDEM_dam, NewDEM)

        # Build pyramids
        arcpy.BuildPyramids_management(NewDEM)
        
        arcpy.Delete_management(dam_ras)
        arcpy.Delete_management(dam_ext)
        arcpy.Delete_management(NewDEM_dam)
        del[dam_ras, dam_ext, NewDEM_dam]
        arcpy.Delete_management("in_memory")
        
    if CutLines <> "" and DamLines == "":

        # Get count of lines to process
        numCut = int(arcpy.GetCount_management(CutLines).getOutput(0))
        arcpy.AddMessage("Processing %s Cut features"  %(numCut))

        # Convert cut lines to raster
        desc = arcpy.Describe(CutLines)
        OID = desc.OIDFieldName
        cut_ras = arcpy.PolylineToRaster_conversion(CutLines, "%s" % OID, "in_memory" + "\\cut_ras", "", "", cellsize)

        # Perform zonal statistics on the DEM with cut_line_ras
        cut_ext = ZonalStatistics(cut_ras, "VALUE", DEM, "MINIMUM")

        # Con operation to burn cut_line_ext into DEM
        NewDEM_cut = Con(IsNull(cut_ext), DEM, cut_ext)

        # Output to file location specified by user
        arcpy.AddMessage("New DEM created with altered elevation values")
        arcpy.CopyRaster_management(NewDEM_cut, NewDEM)

        # Build pyramids
        arcpy.BuildPyramids_management(NewDEM)

        arcpy.Delete_management(cut_ras)
        arcpy.Delete_management(cut_ext)
        arcpy.Delete_management(NewDEM_cut)
        del[cut_ras, cut_ext, NewDEM_cut]
        arcpy.Delete_management("in_memory")

    return(NewDEM)

def TerrainProcessing(NewDEM):

    # Fill the cut DEM
    arcpy.AddMessage("Filling DEM")
    DEMFill = Fill(NewDEM)
    DEMFill.save(OutputFillDEM)

    # Build pyramids
    arcpy.BuildPyramids_management(OutputFillDEM)

    # Calculate Flow Direction
    arcpy.AddMessage("Calculating D8 Flow Direction")
    D8FlowDir = FlowDirection(DEMFill, "", "")
    D8FlowDir.save(OutputFlowDir)

    # Build pyramids
    arcpy.BuildPyramids_management(OutputFlowDir)

    # Calculate Flow Accumulation
    arcpy.AddMessage("Calculating Flow Accumulation")
    D8Accumulation = FlowAccumulation(D8FlowDir)
    D8Accumulation.save(OutputFlowAcc)

    # Build pyramids
    arcpy.BuildPyramids_management(OutputFlowAcc)

    # Calculate Hillshade
    arcpy.AddMessage("Generating Hillshade")
    HillshadeOut = Hillshade(NewDEM, "", "", "", ZFactor)
    HillshadeOut.save(OutputHshd)

    # Build pyramids
    arcpy.BuildPyramids_management(OutputHshd)


if __name__ == "__main__":

    # Define Parameters
    CutLines = arcpy.GetParameterAsText(0)
    DamLines = arcpy.GetParameterAsText(1)
    DEM = arcpy.GetParameterAsText(2)
    NewDEM = arcpy.GetParameterAsText(3)
    OutputFillDEM = arcpy.GetParameterAsText(4)
    OutputFlowDir = arcpy.GetParameterAsText(5)
    OutputFlowAcc = arcpy.GetParameterAsText(6)
    OutputHshd = arcpy.GetParameterAsText(7)
    ZFactor = arcpy.GetParameterAsText(8)

    # Set extent, snap
    arcpy.env.extent = DEM
    arcpy.env.snapRaster = DEM
    rDesc = arcpy.Describe(DEM)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder

    # Determine the resolution of the DEM
    cellsize = float(arcpy.GetRasterProperties_management(DEM, "CELLSIZEX").getOutput(0))

    # Run modules
    NewDEM = ManualCutter(CutLines, DEM)
    
    TerrainProcessing(NewDEM)
    
