# RIPARIAN ANALYSIS--------------------------------------------------------------------------------------------------
#
# NOTE: THIS TOOL WILL TAKE ON AVERAGE 15 min --> 1 HOUR OF PROCESSING TIME
# Times will increase with size of watershed and # of reaches
#
# OBJECTIVE: Perform riparian analysis on a stream reach file
# The stream reach file was created using TauDEM software, and as such has necessary fields that are referred to in this tool
#
# The tool first creates inputs to the riparian analysis --- > rasters to be used in zonal statistics with RAP (riparian analysis polygons)
# Inputs include:
    # RelativeElevation - Original coding for RelativeElevation module: Kathy Boomer - Nature Conservancy 06/2012       
    # WaterTableDepth -  raster that represents the estimated depth to the water table.  This is reclassified into the following rasters for riparian analysis:
        # 2)  ShallowWaterTable (0 - 1.5 meters)
        # 3)  SemiShallowWaterTable (1.5 - 3 meters)
        # 4)  MediumWaterTable (3 - 5 meters)
    # Using an input gSSURGO raster and Soil Value Added table, the average soil organic carbon (in g/m^2) is found within the
    # shallow water table and semi-shallow water table depth zones within each RAP at 2 depths; 20 - 50cm and 50 - 100cm.
    # WeightedFlowAccumulationMasks - uses input field boundaries and drainage table to make tile-drained and non tile-drained mask
        # 1) AdjD8Acc raster: flow accumulation raster using AdjFlowDir to determine flow direction and no weight raster
        # 2) TileDrainedAcc raster: flow accumulation raster using AdjFlowDir to determine flow direction and tile-drained fields mask as weight raster
        # 3) NonTileDrainedAcc raster: flow accumulation raster using AdjFlowDir to determine flow direction and non tile-drained fields mask as weight raster
        
    # The strip map index feature is used to create polygons (250 meters long, 180 meters wide) along each stream reach independently
    # The reach polyline file itself is then used to split the polygons into two RAP polygons (one on each side of the stream)
    # Zonal statistics are performed separately for each RAP polygon
    # Statistics are generated for each RAP polygon and populated in the RiparianTable, along with the RAPID, the unique identifier for each RAP polygon
        # Accumulation masks are extracted to either
        # 1) bank cells - for local accumulation, local tile-drained accumulation, and local non-tile drained accumulation
                # These bank cells are then summed within each RAP
        # 2) channel cells - for watershed accumulation, defined as the maximum flow accumulation among all channel cells in a RAP
        # Areas (in m^2) of water table depth zones and HighSOC are found for each RAP:
                # area in m^2 divided by 250 gives average width of each characteristic within the RAP
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def CreateInputsToRiparianAnalysis(FBIn, DrainageTable, StreamReach, AdjFlowDir, SoilValueTable, gSSURGO, WaterTableDepth):
    arcpy.AddMessage("Creating Weighted Flow Accumulation Masks")
    arcpy.env.compression = "NONE"

    # Dissolve stream reach --> convert to raster
    rch_diss = arcpy.Dissolve_management(StreamReach, "rch_diss", "", "", "", "")
    desc = arcpy.Describe(rch_diss)
    OID = desc.OIDFieldName
    reach_ras = arcpy.PolylineToRaster_conversion(rch_diss, "%s" % OID, "reach_ras", "", "", cellsize)

    # Convert stream values to 1 and background to NODATA
    reach_con = Con(reach_ras, 1, "", "VALUE >= 0")

    # Join "soc20_50" (soil organic carbon at 20 - 50 cm depth) and "soc50_100" (soil organic carbon at 50 - 100 cm depth) to a copy of gSSURGO 
    gSSURGO_copy = arcpy.CopyRaster_management(gSSURGO, "gSSURGO_copy")
    arcpy.JoinField_management(gSSURGO_copy, "mukey", SoilValueTable, "mukey", ["soc20_50", "soc50_100"])
    
    # Create two lookup rasters from gSSURGO_copy; 1) SOC20_50 and 2) SOC50_100 --> cellsize of 3 set in environments
    SOC20_50 = Lookup(gSSURGO_copy, "soc20_50")
    SOC50_100 = Lookup(gSSURGO_copy, "soc50_100")

    # Create Water Table Depth masks --> mask has values of 1 and background is NODATA (SWT: < 1.5 m or channel, SSWT: 1.5-3 m, MWT: 3-5 m) 
    arcpy.AddMessage("Creating Water Table Depth masks")
    SWT = Con(WaterTableDepth, 1, "", "\"WATERTABLEDEPTH\" = 'lt_1_5m' or \"WATERTABLEDEPTH\" = 'Channel'")
    SSWT = Con(WaterTableDepth, 1, "", "\"WATERTABLEDEPTH\" = '1_5to3m'")
    MWT = Con(WaterTableDepth, 1, "", "\"WATERTABLEDEPTH\" = '3to5m'")

    # Create SOC Masks at 2 depths (20 - 50 cm and 50 - 100 cm) within the SWT (< 1.5 m) and SSWT (1.5-3 m) zones
    SOC20_50SWT = Con(SWT, SOC20_50, "", "Value = 1")
    SOC20_50SSWT = Con(SSWT, SOC20_50, "", "Value = 1")
    SOC50_100SWT = Con(SWT, SOC50_100, "", "Value = 1")
    SOC50_100SSWT = Con(SSWT, SOC50_100, "", "Value = 1")

    # Join "drained" field from DrainageTable to a copy of FBIn
    FBIn_copy = arcpy.CopyFeatures_management(FBIn, "in_memory" + "\\FBIn_copy")
    arcpy.JoinField_management(FBIn_copy, "FBndID", DrainageTable, "FBndID", ["DRAINED"])
    
    # CREATE 2 MASKS - 1) tile drained ag fields and 2) non tile-drained ag fields ---> convert to raster based on OID
    arcpy.MakeFeatureLayer_management(FBIn_copy, "td_fields", "\"DRAINED\" = 'YES'")
    desc = arcpy.Describe("td_fields")
    OID = desc.OIDFieldName
    td_mask = arcpy.FeatureToRaster_conversion("td_fields", "%s" % OID, "td_mask", cellsize)

    arcpy.MakeFeatureLayer_management(FBIn_copy, "ntd_fields", "\"DRAINED\" = 'NO' or \"DRAINED\" = 'NonAg'")
    desc = arcpy.Describe("ntd_fields")
    OID = desc.OIDFieldName
    ntd_mask = arcpy.FeatureToRaster_conversion("ntd_fields", "%s" % OID, "ntd_mask", cellsize)

    # Convert NODATA values to 0 
    TileDrainedMask = Con(IsNull(td_mask), 0, 1)
    NonTileDrainedMask = Con(IsNull(ntd_mask), 0, 1)

    # Reclassify AdjFlowDir into taudem flow direction values ---> copy to TDProcDir as uncompressed tiff
    TDFlowDirTemp = Reclassify(AdjFlowDir, "VALUE",
                           RemapRange([[1, 1], [128, 2], [64, 3], [32, 4], [16, 5], [8, 6], [4, 7], [2, 8]]))
    TDFlowDir = arcpy.CopyRaster_management(TDFlowDirTemp, TDProcDir + "\\TDFlowDir.tif")

    # Copy tile drained and non tile-drained mask to TDProcDir as uncompressed tiff
    TDMask = arcpy.CopyRaster_management(TileDrainedMask, TDProcDir + "\\TDMask.tif")
    NTDMask = arcpy.CopyRaster_management(NonTileDrainedMask, TDProcDir + "\\NTDMask.tif")

    core_cnt = multiprocessing.cpu_count()
    core_no = core_cnt - 1

    # Run flow accumulation using adjusted flow direction values and 1)tile drained mask as weight raster 2)non-tiled drained mask as weight raster, and 3)no weight raster 
    callstr = "mpiexec -n %s Aread8 -p " % core_no + \
              TDProcDir + "\\TDFlowDir.tif -wg " + \
              TDProcDir + "\\TDMask.tif -ad8 " + \
              TDProcDir + "\\TDTileDrainedAcc.tif" + \
              ' -nc '
    call(callstr, shell=True)

    callstr = "mpiexec -n %s Aread8 -p " % core_no + \
              TDProcDir + "\\TDFlowDir.tif -wg " + \
              TDProcDir + "\\NTDMask.tif -ad8 " + \
              TDProcDir + "\\TDNonTileDrainedAcc.tif" + \
              ' -nc '
    call(callstr, shell=True)

    callstr = "mpiexec -n %s Aread8 -p " % core_no + \
              TDProcDir + "\\TDFlowDir.tif -ad8 " + \
              TDProcDir + "\\TDFlowAcc.tif" + \
              ' -nc '
    call(callstr, shell=True)

    # Calculate Statistics
    AccTmp = arcpy.CalculateStatistics_management(TDProcDir + "\\TDFlowAcc.tif")
    TileDrainedAccTmp = arcpy.CalculateStatistics_management(TDProcDir + "\\TDTileDrainedAcc.tif")
    NonTileDrainedAccTmp = arcpy.CalculateStatistics_management(TDProcDir + "\\TDNonTileDrainedAcc.tif")

    # Remove stream network from flow accumulation rasters --> save to pass on to future modules
    Acc = Con(IsNull(reach_con), AccTmp, "")
    Acc.save("Acc")
    TDAcc = Con(IsNull(reach_con), TileDrainedAccTmp, "")
    TDAcc.save("TDAcc")
    NTDAcc = Con(IsNull(reach_con), NonTileDrainedAccTmp, "")
    NTDAcc.save("NTDAcc")

    # Define StrmAcc as AccTmp without stream network removed
    StrmAcc = AccTmp

    # Cleanup
    arcpy.Delete_management(rch_diss)
    arcpy.Delete_management(reach_ras)
    arcpy.Delete_management(reach_con)
    arcpy.Delete_management(FBIn_copy)
    arcpy.Delete_management(td_mask)
    arcpy.Delete_management(ntd_mask)
    arcpy.Delete_management(TileDrainedMask)
    arcpy.Delete_management(NonTileDrainedMask)
    arcpy.Delete_management(TDFlowDirTemp)
    arcpy.Delete_management(gSSURGO_copy)
    arcpy.Delete_management(SOC20_50)
    arcpy.Delete_management(SOC50_100)
    del[rch_diss, reach_ras, reach_con, FBIn_copy, td_mask, ntd_mask, TileDrainedMask, NonTileDrainedMask, TDFlowDirTemp, gSSURGO_copy, SOC20_50, SOC50_100]
    arcpy.Delete_management("in_memory")

    # Clear compression and extent environment 
    arcpy.ClearEnvironment("compression")
    arcpy.ClearEnvironment("extent")

    return(StrmAcc, Acc, TDAcc, NTDAcc, SWT, SSWT, MWT, SOC20_50SWT, SOC20_50SSWT, SOC50_100SWT, SOC50_100SSWT)


def RiparianAnalysis(StreamReach, StrmAcc, Acc, TDAcc, NTDAcc, SWT, SSWT, MWT, SOC20_50SWT, SOC20_50SSWT, SOC50_100SWT, SOC50_100SSWT):
    arcpy.AddMessage("Perform Riparian Analysis")
    
    # Create a FEATURE CLASS that will contain all of the riparian boxes - add fields that will be required during each append process
    RAP = arcpy.CreateFeatureclass_management("in_memory", "RAP", "POLYGON", "", "", "", StreamReach)
    RiparianPolygons = arcpy.CopyFeatures_management(RAP, OutRAP)
    arcpy.AddField_management(RiparianPolygons, "RAPID", "TEXT")
    arcpy.AddField_management(RiparianPolygons, "Strm_lngth", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "Wshed_Run", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SUM", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SUM_1", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "SUM_12", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "AREA", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "AREA_1", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "AREA_12", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "MEAN", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "MEAN_1", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "MEAN_12", "FLOAT", "", "", 25)
    arcpy.AddField_management(RiparianPolygons, "MEAN_12_13", "FLOAT", "", "", 25)
    arcpy.Delete_management(RAP)
    del[RAP]

    # Create reach list to loop through --> the straight line distance ("Straight_L") must be >= the length of the riparian box that will be created (250m)
    rows = arcpy.SearchCursor(StreamReach)
    ReachList = []
    for row in rows:
        if row.Straight_L >= 250:
            ReachID = str(row.getValue("LINKNO"))
            ReachList.append(ReachID)
    ReachCount = len(ReachList)
    arcpy.AddMessage("Processing %s Reaches" % (ReachCount))

    # Begin riparian analysis loop
    arcpy.AddMessage("Beginning riparian analysis loop")
    
    for Reach in ReachList:
        arcpy.AddMessage("Reach" + str(Reach))

        # Select just the current reach polyline ---> convert to raster
        reach_line = arcpy.FeatureClassToFeatureClass_conversion(StreamReach, "in_memory", "reach_line", '\"LINKNO\" = %s' % Reach)
        reach_ras = arcpy.FeatureToRaster_conversion(reach_line, "LINKNO", "in_memory" + "\\reach_ras", cellsize)
        
        # Create "reach_bank_id": bank cells "along current reach only" receive a value of 1, all other cells receive NODATA
        bank_ID = Con(IsNull(reach_ras), Expand(Con(reach_ras, 1, "", "VALUE >= 0"), 1, 1), "", "")

        # Extract Acc, TDAcc, and NTDAcc to the "bank" cells of the current reach
        BankTDAcc = ExtractByMask(TDAcc, bank_ID)
        BankNTDAcc = ExtractByMask(NTDAcc, bank_ID)
        BankAcc = ExtractByMask(Acc, bank_ID)

        # Extract StrmAcc to the "channel" cells of the current reach
        RchAcc = ExtractByMask(StrmAcc, reach_ras)

        # Create riparian polygons along each reach (250 meters long, 180 meters wide, 0% overlap) - pagenumber is unique identifier of each box
        rip_boxes = arcpy.StripMapIndexFeatures_cartography(reach_line, "rip_boxes", "", "", "250", "180", "", "0", "", "WE_SN")

        # Create 2 feature classes; odd "pagenumber" rip_boxes will be appended to odd_rch; even "pagenumber" rip_boxes will be appended to even_rch
        # Allows zonal stats of all odd RAPs and even RAPs along a reach to be processed together - significantly reduces processing time
        odd_rch = arcpy.CreateFeatureclass_management("in_memory", "odd_rch", "POLYGON", "", "", "", StreamReach)
        arcpy.AddField_management(odd_rch, "RAPID", "TEXT", "", "", 12)
        arcpy.AddField_management(odd_rch, "Strm_lngth", "FLOAT")
        arcpy.AddField_management(odd_rch, "Wshed_Run", "FLOAT")
        even_rch = arcpy.FeatureClassToFeatureClass_conversion(odd_rch, "in_memory", "even_rch")

        # For EACH box (strip map polygon) that was created along the stream reach:
        with arcpy.da.SearchCursor(rip_boxes, ["PageNumber"]) as cursor:
            for row in cursor:
                # Define PageNumber as odd or even
                PageNumber = str(row[0])
                Type = int(PageNumber) % 2

                # select one polygon at a time (will not yet be split by stream) ---> define variables that will be the same for each side of stream)
                riparian_box = arcpy.FeatureClassToFeatureClass_conversion(rip_boxes, "in_memory", "riparian_box", "\"PageNumber\" = %s" % PageNumber)
                
                # Watershed Runoff (max accumulation value of all "channel" cells in riparian_box)      
                AccStats = ZonalStatisticsAsTable(riparian_box, "PageNumber", RchAcc, "in_memory" + "\\AccStats", "#", "MAXIMUM")
                with arcpy.da.SearchCursor(AccStats, ["MAX"]) as runoffcursor:
                    for x in runoffcursor:
                        WatershedRunoff = float((x[0] * (cellsize * cellsize)) / 10000.00)

                # StreamLength (intersect riparian_box with reach_line ---> dissolve to define stream length)
                box_reach_sect = arcpy.Intersect_analysis([riparian_box, reach_line], "in_memory" + "\\box_reach_sect", "", "", "LINE")
                box_reach_diss = arcpy.Dissolve_management(box_reach_sect, "in_memory" + "\\box_reach_diss")
                with arcpy.da.SearchCursor(box_reach_diss, ["SHAPE@LENGTH"]) as lengthcursor:
                    for y in lengthcursor:
                        StreamLength = float(y[0])

                # The following steps use the stream network to split the riparian box into 2; 1 box for each side of the stream, while accounting for topological issues
                # 1) Convert riparian polygon to line and merge with "reach_line"
                # 2) Extend lines by 25 meters to completely bisect each box
                # 3) Convert back to polygon - box_split (will now contain two boxes for every one - split by streamnet)
                box_to_line = arcpy.PolygonToLine_management(riparian_box, "in_memory" + "\\box_to_line")
                box_merge = arcpy.Merge_management([box_to_line, reach_line], "in_memory" + "\\box_merge")
                arcpy.ExtendLine_edit(box_merge, "25 meters")
                box_split = arcpy.FeatureToPolygon_management(box_merge, "in_memory" + "\\box_split")
                
                # Add fields for Watershed runoff and Stream Length - populate with variables defined above
                arcpy.AddField_management(box_split, "Strm_lngth", "float")
                arcpy.CalculateField_management(box_split, "Strm_lngth", '%s' % StreamLength, "PYTHON")
                arcpy.AddField_management(box_split, "Wshed_Run", "float")
                arcpy.CalculateField_management(box_split, "Wshed_Run", WatershedRunoff, "PYTHON")

                # Ensure that the length of each box is > 250 and that the area is < 40000 m^2 --> delete if not
                with arcpy.da.SearchCursor(box_split, ["SHAPE@AREA", "Strm_lngth"]) as boxcursor:
                    for box in boxcursor:
                        if box[0] > 40000 or box[1] < 250:
                            arcpy.Delete_management(box_split)

                # If box still exists (i.e. it met the above dimension criteria) --> get count of # of polygons (or rows) that exist
                if arcpy.Exists(box_split):
                    box_count = int(arcpy.GetCount_management(box_split).getOutput(0))

                    # If more than 2 boxes exist, there were some topological issues ---> select the smallest boxes and perform an eliminate
                    if box_count > 2:
                        arcpy.AddField_management(box_split, "Area", "FLOAT")
                        arcpy.CalculateField_management(box_split, "Area", "!shape.area@squaremeters!", "PYTHON")
                        AreaSort = arcpy.Sort_management(box_split, "in_memory" + "\\AreaSort", [["Area", "DESCENDING"]])

                        desc = arcpy.Describe(AreaSort)
                        FID = desc.OIDFieldName
                        arcpy.MakeFeatureLayer_management(AreaSort, "box_eliminate")
                        with arcpy.da.SearchCursor("box_eliminate", ["%s" % FID]) as eliminatecursor:
                            for eliminate in eliminatecursor:
                                boxID = eliminate[0]
                                arcpy.SelectLayerByAttribute_management("box_eliminate", "ADD_TO_SELECTION", '\"%s\" = %s' % (FID, boxID))
                                count = int(arcpy.GetCount_management("box_eliminate").getOutput(0))
                                if count > 2:
                                    arcpy.SelectLayerByAttribute_management("box_eliminate", "REMOVE_FROM_SELECTION", '\"%s\" = %s' % (FID, boxID))
                                    arcpy.SelectLayerByAttribute_management("box_eliminate", "SWITCH_SELECTION")
                                    box_edit = arcpy.Eliminate_management("box_eliminate", "in_memory" + "\\box_edit", "LENGTH")
                                    break
                        
                        # Re-add fields for Watershed runoff and Stream Length --> repopulate
                        arcpy.AddField_management(box_edit, "Strm_lngth", "float")
                        arcpy.CalculateField_management(box_edit, "Strm_lngth", '%s' % StreamLength, "PYTHON")
                        arcpy.AddField_management(box_edit, "Wshed_Run", "float")
                        arcpy.CalculateField_management(box_edit, "Wshed_Run", WatershedRunoff, "PYTHON")

                        # Add unique identifier field - RAPID - structure is: Reach_PageNumber_OID
                        arcpy.AddField_management(box_edit, "RAPID", "TEXT")
                        desc = arcpy.Describe(box_edit)
                        FID = desc.OIDFieldName
                        arcpy.CalculateField_management(box_edit, "RAPID", "('%s' + '_' + '%s' + '_' + '!%s!')" % (Reach, PageNumber, FID), "PYTHON")

                        # Append to appropriate feature class (odd_rch OR even_rch)
                        if Type == 0:
                            arcpy.Append_management(box_edit, even_rch, "NO_TEST")
                        elif Type == 1:
                            arcpy.Append_management(box_edit, odd_rch, "NO_TEST")

                        arcpy.Delete_management(AreaSort)
                        arcpy.Delete_management(box_edit)
                        del[AreaSort, box_edit]
                        
                    elif box_count <= 2:
                        # Add unique identifier field - RAPID - structure is: Reach_PageNumber_OID
                        desc = arcpy.Describe(box_split)
                        FID = desc.OIDFieldName
                        arcpy.AddField_management(box_split, "RAPID", "TEXT")
                        arcpy.CalculateField_management(box_split, "RAPID", "('%s' + '_' + '%s' + '_' + '!%s!')" % (Reach, PageNumber, FID), "PYTHON")
                        
                        # Append to the appropriate feature class (odd_rch OR even_rch)
                        if Type == 0:
                            arcpy.Append_management(box_split, even_rch, "NO_TEST")
                        elif Type == 1:
                            arcpy.Append_management(box_split, odd_rch, "NO_TEST")
                    
                arcpy.Delete_management(riparian_box)
                arcpy.Delete_management(AccStats)
                arcpy.Delete_management(box_reach_sect)
                arcpy.Delete_management(box_reach_diss)
                arcpy.Delete_management(box_to_line)
                arcpy.Delete_management(box_merge)
                arcpy.Delete_management(box_split)
                del[riparian_box, AccStats, box_reach_sect, box_reach_diss, box_to_line, box_merge, box_split]
            
        # Get the count of odd_rch and even_rch
        EvenCount = int(arcpy.GetCount_management(even_rch).getOutput(0))
        OddCount = int(arcpy.GetCount_management(odd_rch).getOutput(0))

        # If there is at least one box in odd_rch OR even_rch --> append to a RAP_reach_list (to loop through for zonal stats)
        RAP_reach_list = []
        if EvenCount >= 1:
            RAP_reach_list.append(even_rch)
        if OddCount >= 1:
            RAP_reach_list.append(odd_rch)

        # Loop through RAP_reach_list to perform zonal stats - each RAP_reach represents ALL odd OR even boxes along a given reach 
        for RAP_reach in RAP_reach_list:

            # zone: RAP_even; underlying raster: "BankAcc"  ----> # Result is sum of local accumulation (lateral runoff) for each polygon   
            BankAccStats = ZonalStatisticsAsTable(RAP_reach, "RAPID", BankAcc, "in_memory" + "\\BankAccStats", "#", "SUM")
            arcpy.JoinField_management(RAP_reach, "RAPID", BankAccStats, "RAPID", ["SUM"])

            # zone: RAP_even; underlying raster: "BankTDAcc"  ----> # Result is sum of local tile-drained accumulation (lateral runoff) for each polygon  
            BankTDAccStats = ZonalStatisticsAsTable(RAP_reach, "RAPID", BankTDAcc, "in_memory" + "\\BankTDAccStats", "#", "SUM")
            arcpy.JoinField_management(RAP_reach, "RAPID", BankTDAccStats, "RAPID", ["SUM"])

            # zone: RAP_even; underlying raster: "BankTDAcc"  ----> # Result is sum of local non tile-drained accumulation (lateral runoff) for each polygon
            BankNTDAccStats = ZonalStatisticsAsTable(RAP_reach, "RAPID", BankNTDAcc, "in_memory" + "\\BankNTDAccStats", "#", "SUM")
            arcpy.JoinField_management(RAP_reach, "RAPID", BankNTDAccStats, "RAPID", ["SUM"])

            # zone: RAP_even; underlying raster: "SWT"  ----> # Result is area of shallow water table (<1.5 m) in each polygon
            SWTStats = ZonalStatisticsAsTable(RAP_reach, "RAPID", SWT, "in_memory" + "\\SWTStats", "#", "SUM")
            arcpy.JoinField_management(RAP_reach, "RAPID", SWTStats, "RAPID", ["AREA"])

            # zone: RAP_even; underlying raster: "SemiShallowWaterTable"  ----> # Result is area of semi shallow water table (1.5 - 3 m) in each polygon
            SSWTStats = ZonalStatisticsAsTable(RAP_reach, "RAPID", SSWT, "in_memory" + "\\SSWTStats", "#", "SUM")
            arcpy.JoinField_management(RAP_reach, "RAPID", SSWTStats, "RAPID", ["AREA"])

            # zone: RAP_even; underlying raster: "MediumWaterTable"  ----> # Result is area of medium water table (3 - 5 m) in each polygon
            MWTStats = ZonalStatisticsAsTable(RAP_reach, "RAPID", MWT, "in_memory" + "\\MWTStats", "#", "SUM")
            arcpy.JoinField_management(RAP_reach, "RAPID", MWTStats, "RAPID", ["AREA"])

            # zone: RAP_even; underlying raster: "SOC20_50SWT"  ----> # Result is mean SOC (g/m2) at a depth of 20 - 50 cm in SWT zone of each polygon
            SOC20_50SWT_Stats = ZonalStatisticsAsTable(RAP_reach, "RAPID", SOC20_50SWT, "in_memory" + "\\SOC20_50SWTStats", "#", "MEAN")
            arcpy.JoinField_management(RAP_reach, "RAPID", SOC20_50SWT_Stats, "RAPID", ["MEAN"])

            # zone: RAP_even; underlying raster: "SOC20_50SSWT"  ----> # Result is mean SOC (g/m2) at a depth of 20 - 50 cm in SSWT zone of each polygon
            SOC20_50SSWT_Stats = ZonalStatisticsAsTable(RAP_reach, "RAPID", SOC20_50SSWT, "in_memory" + "\\SOC20_50SSWTStats", "#", "MEAN")
            arcpy.JoinField_management(RAP_reach, "RAPID", SOC20_50SSWT_Stats, "RAPID", ["MEAN"])            

            # zone: RAP_even; underlying raster: "SOC50_100SWT"  ----> # Result is mean SOC (g/m2) at a depth of 50 - 100 cm in SWT zone of each polygon
            SOC50_100SWT_Stats = ZonalStatisticsAsTable(RAP_reach, "RAPID", SOC50_100SWT, "in_memory" + "\\SOC50_100SWTStats", "#", "MEAN")
            arcpy.JoinField_management(RAP_reach, "RAPID", SOC50_100SWT_Stats, "RAPID", ["MEAN"])

            # zone: RAP_even; underlying raster: "SOC50_100SSWT"  ----> # Result is mean SOC (g/m2) at a depth of 50 - 100 cm in SSWT zone of each polygon
            SOC50_100SSWT_Stats = ZonalStatisticsAsTable(RAP_reach, "RAPID", SOC50_100SSWT, "in_memory" + "\\SOC50_100SSWTStats", "#", "MEAN")
            arcpy.JoinField_management(RAP_reach, "RAPID", SOC50_100SSWT_Stats, "RAPID", ["MEAN"])  

            # Append RAP_reach to RiparianPolygons ---> fields should match as they were added to RiparianPolygons feature class at beginning of module
            arcpy.Append_management(RAP_reach, RiparianPolygons, "NO_TEST")

            # Cleanup
            arcpy.Delete_management(BankAccStats)
            arcpy.Delete_management(BankTDAccStats)
            arcpy.Delete_management(BankNTDAccStats)
            arcpy.Delete_management(SWTStats)
            arcpy.Delete_management(SSWTStats)
            arcpy.Delete_management(MWTStats)
            arcpy.Delete_management(SOC20_50SWT_Stats)
            arcpy.Delete_management(SOC20_50SSWT_Stats)
            arcpy.Delete_management(SOC50_100SWT_Stats)
            arcpy.Delete_management(SOC50_100SSWT_Stats)
            del[BankAccStats, BankTDAccStats, BankNTDAccStats, SWTStats, SSWTStats, MWTStats, SOC20_50SWT_Stats, SOC20_50SSWT_Stats, SOC50_100SWT_Stats, SOC50_100SSWT_Stats]    

        # Cleanup
        arcpy.Delete_management(odd_rch)
        arcpy.Delete_management(even_rch)
        arcpy.Delete_management(reach_line)
        arcpy.Delete_management(reach_ras)
        arcpy.Delete_management(bank_ID)
        arcpy.Delete_management(rip_boxes)
        del[odd_rch, even_rch, reach_line, reach_ras, BankTDAcc, BankNTDAcc, BankAcc, RchAcc, bank_ID, rip_boxes]     
        arcpy.Delete_management("in_memory")

    # Add and relabel fields in RiparianPolygons
    arcpy.AddField_management(RiparianPolygons, "Local_Run", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "TD_Run", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "NTD_Run", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SWT_Width", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SSWT_Width", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "MWT_Width", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SWT20_50", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SSWT20_50", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SWT50_100", "FLOAT")
    arcpy.AddField_management(RiparianPolygons, "SSWT50_100", "FLOAT")

    with arcpy.da.UpdateCursor(RiparianPolygons, ["SUM", "SUM_1", "SUM_12", "AREA", "AREA_1", "AREA_12",  "MEAN", "MEAN_1", "MEAN_12", "MEAN_12_13", \
                                               "Local_Run", "TD_Run", "NTD_Run", "SWT_Width", "SSWT_Width", "MWT_Width", "SWT20_50", "SSWT20_50", "SWT50_100", "SSWT50_100"]) as cursor:
        for row in cursor:
            if row[0] == None:
                row[10] = 0
            else:
                row[10] = ((row[0] * (cellsize * cellsize)) / 10000.00)
            if row[1] == None:
                row[11] = 0
            else:
                row[11] = ((row[1] * (cellsize * cellsize)) / 10000.00)
            if row[2] == None:
                row[12] = 0
            else:
                row[12] = ((row[2] * (cellsize * cellsize)) / 10000.00)
            if row[3] == None:
                row[13] = 0
            else:
                row[13] = (row[3] / 250)
            if row[4] == None:
                row[14] = 0
            else:
                row[14] = (row[4] / 250)
            if row[5] == None:
                row[15] = 0
            else:
                row[15] = (row[5] / 250)
            if row[6] == None:
                row[16] = 0
            else:
                row[16] = row[6]
            if row[7] == None:
                row[17] = 0
            else:
                row[17] = row[7]
            if row[8] == None:
                row[18] = 0
            else:
                row[18] = row[8]
            if row[9] == None:
                row[19] = 0
            else:
                row[19] = row[9]
            cursor.updateRow(row)

    # Delete fields
    arcpy.DeleteField_management(RiparianPolygons, ["SUM", "SUM_1", "SUM_12", "AREA", "AREA_1", "AREA_12", "MEAN", "MEAN_1", "MEAN_12", "MEAN_12_13"])

    # Cleanup
    arcpy.Delete_management(SWT)
    arcpy.Delete_management(SSWT)
    arcpy.Delete_management(MWT)
    arcpy.Delete_management(StrmAcc)
    arcpy.Delete_management(Acc)
    arcpy.Delete_management(TDAcc)
    arcpy.Delete_management(NTDAcc)
    arcpy.Delete_management(SOC20_50SWT)
    arcpy.Delete_management(SOC20_50SSWT)
    arcpy.Delete_management(SOC50_100SWT)
    arcpy.Delete_management(SOC50_100SSWT)
    arcpy.Delete_management(TDProcDir)
    del[SWT, SSWT, MWT, StrmAcc, Acc, TDAcc, NTDAcc, SOC20_50SWT, SOC20_50SSWT, SOC50_100SWT, SOC50_100SSWT]

    return(RiparianPolygons)


def RiparianFunctionClassification(RiparianPolygons):

    # Find total runoff to the stream "through riparian boxes" by summing the local runoff for all riparian boxes 
    LocalRunoffStats = arcpy.Statistics_analysis(RiparianPolygons, "in_memory" + "\\LocalRunoffStats", [["Local_Run", "SUM"]])
    with arcpy.da.SearchCursor(LocalRunoffStats,["SUM_Local_Run"]) as cursor:
        for row in cursor:
            TotalRunoffToStream = row[0]

    # Find 50% of the total runoff through all boxes (those boxes with the highest runoff that add up to this value will be ranked "HIGH"
        FiftyPctRunoffToStream = float(TotalRunoffToStream / 2)

    # convert any 0 value runoff to .0001 
    arcpy.MakeTableView_management(RiparianPolygons, "NullValues")
    arcpy.SelectLayerByAttribute_management("NullValues", "NEW_SELECTION", "\"Local_Run\" = 0")
    arcpy.CalculateField_management("NullValues", "Local_Run", .0001, "PYTHON")

    # Make table view of riparian table - sort by local runoff in descending order (boxes with highest runoff will be top rows)
    RiparianTable = arcpy.MakeTableView_management(RiparianPolygons, "RiparianTable")
    Run_sort = arcpy.Sort_management(RiparianTable, "in_memory" + "\\Run_sort", [["Local_Run", "DESCENDING"]])

    # Add and populate a cumulative "Local_Run" field to the sorted table (each row will add its local runoff to the running total)
    arcpy.AddField_management(Run_sort, "Accum_CA", "FLOAT")
    AccumulationValue = 0
    with arcpy.da.UpdateCursor(Run_sort, ["Local_Run", "Accum_CA"]) as cursor:
        for row in cursor:      
            AccumulationValue = row[0] + AccumulationValue
            row[1] = AccumulationValue
            cursor.updateRow(row)

    # Join this accumulatedCA field to the riparian table based on the RAPID
    arcpy.JoinField_management(RiparianPolygons, "RAPID", Run_sort, "RAPID", "Accum_CA")

    # Add and populate a BuffCARat10 field; = ratio of the width of a 10 meter buffer to the average slope length (local runoff (in m^2) divided by length of box)
    arcpy.AddField_management(RiparianPolygons, "BufCARat10", "FLOAT")
    with arcpy.da.UpdateCursor(RiparianPolygons,["Local_Run", "BufCARat10"]) as cursor:
        for row in cursor:
            row[1] = (10 / ((row[0] * 10000.00) / 250))
            cursor.updateRow(row)

    # Add fields to classify each RAP into high, med, or low for runoff and water table interaction rank
    arcpy.AddField_management(RiparianPolygons, "RunoffRank", "TEXT")
    arcpy.AddField_management(RiparianPolygons, "WTInt_Rank", "TEXT")

    # High RunoffRank - runoff of RAPs with highest runoff adds up to 50% of runoff to stream
    # Med RunoffRank - the ratio of the width of a 10 meter buffer to the average slope length is < .02 (a wider buffer is needed to meet NRCS 2% guidelines)
    # Low RunoffRank - the ratio of the width of a 10 meter buffer to the average slope length is > .02 (a 10 meter buffer is sufficient to meet NRCS 2% guidelines)
    
    # High WTIRank - average width of shallow water table (SWT) zone is > 50 meters
    # Med WTIRank - average width of SWT zone is 25 - 50 meters
    # Low WTIRank - average width of SWT zone is < 25 meters
    with arcpy.da.UpdateCursor(RiparianPolygons,["BufCARat10", "Accum_CA", "SWT_Width", "RunoffRank", "WTInt_Rank",]) as cursor:
        for row in cursor:
            if row[1] < FiftyPctRunoffToStream:
                row[3] = 'High'
            elif row[0] < .02:
                row[3] = 'Med'
            elif row[0] >= .02:
                row[3] = 'Low'
            if row[2] > 50:
                row[4] = 'High'
            elif row[2] <= 50 and row[2] > 25:
                row[4] = 'Med'
            elif row[2] <= 25:
                row[4] = 'Low'
            cursor.updateRow(row)
        arcpy.DeleteField_management(RiparianPolygons, ["Accum_CA"])
            
    # Populate riparian function field by the matrix of local runoff rank and water table interaction rank
    arcpy.AddField_management(RiparianPolygons, "RAP_Func", "TEXT")
    with arcpy.da.UpdateCursor(RiparianPolygons, ["RunoffRank", "WTInt_Rank", "RAP_Func"]) as cursor:
        for row in cursor:
            if row[0] == 'High' and row[1] == 'High':
                row[2] = 'Critical Zone'
            elif (row[0] == 'High' and row[1] == 'Med') or \
                 (row[0] == 'Med' and row[1] == 'High') or \
                 (row[0] == 'Med' and row[1] == 'Med'):
                row[2] = 'Multi Species Buffer'
            elif (row[0] == 'High' and row[1] == 'Low') or \
                 (row[0] == 'Med' and row[1] == 'Low'):
                row[2] = 'Stiff Stemmed Grasses'
            elif (row[0] == 'Low' and row[1] == 'High') or \
                 (row[0] == 'Low' and row[1] == 'Med'):
                row[2] = 'Deep Rooted Vegetation'
            elif (row[0] == 'Low' and row[1] == 'Low'):
                row[2] = 'Stream Bank Stabilization'
            cursor.updateRow(row)

    arcpy.AddField_management(RiparianPolygons, "BuffWidth", "FLOAT")
    with arcpy.da.UpdateCursor(RiparianPolygons, ["RAP_Func", "Local_Run", "SWT_Width", "BuffWidth"]) as cursor:
        for row in cursor:
            NRCSwidth = (row[1] * 10000 * .02) / 250
            if row[0] == 'Multi Species Buffer' and NRCSwidth > 90:
                row[3] = 90
            elif row[0] == 'Multi Species Buffer' and NRCSwidth < 25:
                row[3] = 25
            elif row[0] == 'Multi Species Buffer' and NRCSwidth >= 25 and NRCSwidth <= 90:
                row[3] = NRCSwidth
            if row[0] == 'Critical Zone' and NRCSwidth > 90:
                row[3] = 90
            elif row[0] == 'Critical Zone' and NRCSwidth < 25:
                row[3] = 25
            elif row[0] == 'Critical Zone' and NRCSwidth >= 25 and NRCSwidth <= 90:
                row[3] = NRCSwidth
            if row[0] == 'Stiff Stemmed Grasses' and NRCSwidth > 90:
                row[3] = 90
            elif row[0] == 'Stiff Stemmed Grasses' and NRCSwidth <= 90:
                row[3] = NRCSwidth            
            if row[0] == 'Deep Rooted Vegetation' and row[2] > 50:
                row[3] = 50
            elif row[0] == 'Deep Rooted Vegetation' and row[2] <= 50:
                row[3] = row[2]
            if row[0] == 'Stream Bank Stabilization' and NRCSwidth < 6:
                row[3] = 6
            elif row[0] == 'Stream Bank Stabilization' and NRCSwidth >= 6:
                row[3] = NRCSwidth
            cursor.updateRow(row)

    arcpy.Delete_management(RiparianTable)
    arcpy.Delete_management(LocalRunoffStats)
    arcpy.Delete_management(Run_sort)
    del[RiparianTable, LocalRunoffStats, Run_sort]
    arcpy.Delete_management("in_memory")
    

if __name__ == "__main__":
    
    # Define Parameters
    FBIn = arcpy.GetParameterAsText(0)
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        basename = arcpy.Describe(FBIn).baseName
        arcpy.AddError("Found a join on %s that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    StreamReach = arcpy.GetParameterAsText(1)
    gSSURGO = arcpy.GetParameterAsText(2)
    SoilValueTable = arcpy.GetParameterAsText(3)
    DrainageTable = arcpy.GetParameterAsText(4)
    AdjFlowDir = arcpy.GetParameterAsText(5)
    WaterTableDepth = arcpy.GetParameterAsText(6)
    OutRAP = arcpy.GetParameterAsText(7)
   
    # Set snap
    arcpy.env.snapRaster = AdjFlowDir
    arcpy.env.extent = AdjFlowDir
    rDesc = arcpy.Describe(AdjFlowDir)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder

    # Define cellsize
    cellsize = float(arcpy.GetRasterProperties_management(AdjFlowDir, "CELLSIZEX").getOutput(0))
    
    # Create and define folder for TauDEM outputs (in ProcDir) - will be deleted at end of script
    scriptpath = sys.path[0]
    if ' ' in scriptpath:
        arcpy.AddError("space exists in %s that must be removed from path name...QUITTING!!!" % scriptpath)
        sys.exit(0)
    else:
        arcpy.CreateFolder_management(scriptpath, "TDProcDir")
        TDProcDir = scriptpath + "\\TDProcDir"

    # Run Modules 
    StrmAcc, Acc, TDAcc, NTDAcc, SWT, SSWT, MWT, SOC20_50SWT, SOC20_50SSWT, SOC50_100SWT, SOC50_100SSWT = \
             CreateInputsToRiparianAnalysis(FBIn, DrainageTable, StreamReach, AdjFlowDir, SoilValueTable, gSSURGO, WaterTableDepth)

    RiparianPolygons = RiparianAnalysis(StreamReach, StrmAcc, Acc, TDAcc, NTDAcc, SWT, SSWT, MWT, SOC20_50SWT, SOC20_50SSWT, SOC50_100SWT, SOC50_100SSWT)
        
    RiparianFunctionClassification(RiparianPolygons)


 

