# Runoff Risk Assessment------------------------------------------------------------------------------------------------------------
#
# Performs a cross classification of two variables within EACH AGRICULTURAL FIELD:
# 1) Distance To Stream (converted to a sediment delivery ratio using the MN Phosphorus Index), and
# 2) Slope (using the 3rd quartile - 75% slope value - of each field)
# ******** the distance to stream is the MINIMUM distance from each field to the stream *********
#
# For each variable, classify into a high, med, or low classification using either of 2 OPTIONS:
# 1) User-defined values for breaks
# ****(user will specify a high slope break (slopes above this considered high) and a medium slope break (slopes below this considered low) ****
# 2) Automatic breaks generation: high - top 20% of fields, med = middle 40% of fields, and low = lowest 40% of fields
# 
# These two variables are used in a cross classification to calculate RUNOFF RISK of each field (Critical, VeryHigh, High, and Present)
# **** all agricultural fields are classified into a runoff risk classification (i.e. lowest risk fields get classified as Present) *****
# 
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------

import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
arcpy.CheckOutExtension("Spatial")


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def RunoffRiskAssessment(FBIn, SlopeTable, DistanceToStream):
    
    # Create mask of just agricultural fields    
    lyr = arcpy.MakeFeatureLayer_management(FBIn, "lyr", "\"isAG\" >= 1") 
    AgFB = arcpy.CopyFeatures_management(lyr, "in_memory" + "\\AgFB")

    # -------------------------SEDIMENT DELIVERY RATIO -----------------------------------------
    arcpy.AddMessage("Calculating by-field sediment delivery ratio")

    # Use zonal statistics to find the minimum distance to stream for each ag field (in feet)
    DTSStats = ZonalStatisticsAsTable(AgFB, "FBndID", DistanceToStream, "in_memory" + "\\DTSStats", "#", "MINIMUM")
    arcpy.JoinField_management(AgFB, "FBndID", DTSStats, "FBndID", ["MIN"])

    # Delete fields that drain out of watershed (the result of above zonal statistics is null)
    arcpy.MakeFeatureLayer_management(AgFB, "NullValues")
    arcpy.SelectLayerByAttribute_management("NullValues", "NEW_SELECTION", "\"MIN\" IS NULL")
    arcpy.DeleteFeatures_management("NullValues")

    # Convert minimum distance to stream ("MIN" field) from meters to feet
    arcpy.AddField_management(AgFB, "DTS_ft", "FLOAT")
    with arcpy.da.UpdateCursor(AgFB, ["MIN", "DTS_ft"]) as cursor:
        for row in cursor:
            row[1] = row[0] * 3.28084
            cursor.updateRow(row)

    # Delete min field
    arcpy.DeleteField_management(AgFB, ["MIN"])

    # Calculate adjusted distance to stream;
    # If minimum distance (in feet) is < 10 ft, convert to 0; otherwise, subtract distance of 10 feet from each field
    arcpy.AddField_management(AgFB, "AdjDTS", "FLOAT")
    with arcpy.da.UpdateCursor(AgFB, ["DTS_ft", "AdjDTS"]) as cursor:
        for row in cursor:
            if row[0] <= 10:
                row[1] = 1
            elif row[0] > 10:
                row[1] = row[0]
            cursor.updateRow(row)

    # Calculate Sediment Delivery Ratio - equation derived from MN phosphorus index
    arcpy.AddField_management(AgFB, "SDR", "FLOAT")
    with arcpy.da.UpdateCursor(AgFB, ["AdjDTS", "SDR"]) as cursor:
        for row in cursor:
            if row[0] >= 5000:
                row[1] = .17
            else:
                row[1] = row[0]**(-.2069)
            cursor.updateRow(row)


    # -------------------------- 3RD QUARTILE SLOPE----------------------------------------
    arcpy.AddMessage("Calculating by-field third quartile slope values")

    # Join "Slope75Pct" field from slope table
    arcpy.JoinField_management(AgFB, "FBndID", SlopeTable, "FBndID", ["Slope75Pct"])

    # Copy AgFB to runoff risk table
    RRTable = arcpy.TableToTable_conversion(AgFB, "in_memory", "RRTable", "")
    RunoffRiskTable = arcpy.CopyRows_management(RRTable, OutRunoffRisk)

    # Add fields for slope rank and SDR rank
    arcpy.AddField_management(RunoffRiskTable, "SlopeRank", "TEXT")
    arcpy.AddField_management(RunoffRiskTable, "SDRRank", "TEXT")

    # Perform a field count to determine how many fields exist
    FieldCount = int(arcpy.GetCount_management(RunoffRiskTable).getOutput(0)) 

    # How many fields represent 20% of fields
    Field20Count = int(.20 * FieldCount)

    # How many fields represent 60% of fields
    Field60Count = int(.60 * FieldCount)

    arcpy.Delete_management(AgFB)
    arcpy.Delete_management(DTSStats)
    arcpy.Delete_management(RRTable)
    del[AgFB, DTSStats, RRTable]
    arcpy.Delete_management("in_memory")

  
    # -----------------------------SLOPE RANK----------------------------------------------
    # Calculate breaks for High, Med, and Low using:
    arcpy.AddMessage("Classifying each field by slope rank")

    # 1) user-defined values for breaks
    if HighSlope <> '':
        if MediumSlope == '':
            arcpy.AddError("Must provide both medium and high slope values or neither...QUITTING!")
    if MediumSlope <> '':
        if HighSlope == '':
            arcpy.AddError("Must provide both medium and high slope values or neither...QUITTING!")
    if HighSlope <> '' and MediumSlope <> '':
        arcpy.AddMessage("User defined slope breaks provided")
        HighSlope_fl = float(HighSlope)
        MedSlope_fl = float(MediumSlope)
        with arcpy.da.UpdateCursor(RunoffRiskTable, ["Slope75Pct", "SlopeRank"]) as cursor:
            for row in cursor:
                if row[0] >= HighSlope_fl:
                    row[1] = 'High'
                elif row[0] < HighSlope_fl and row[0] >= MedSlope_fl:
                    row[1] = 'Med'
                elif row[0] < MedSlope_fl:
                    row[1] = 'Low'
                cursor.updateRow(row)

    # 2) automatically generated breaks (using a 20%-40%-40% breakdown of fields)
    if HighSlope == '' and MediumSlope == '':
        arcpy.AddMessage("Automatically generated slope breaks...calculating 20-40-40 breakdown of fields")
        
        # Sort RunoffRiskTable by Slope75Pct in a descending order
        SlopeSort = arcpy.Sort_management(RunoffRiskTable, "in_memory" + "\\SlopeSort", [["Slope75Pct", "DESCENDING"]])
        desc = arcpy.Describe(SlopeSort)
        OID = desc.OIDFieldName
    
        # Identify the 3rd quartile slope value that 20% of fields are steeper than
        Field20 = arcpy.MakeTableView_management(SlopeSort, "Field20", '\"%s\" = %s' % (OID, Field20Count))
        for row in arcpy.da.SearchCursor(Field20, ["Slope75Pct"]):
            Slope20Value = row[0]

        # Identify the 3rd quartile slope value above which 60% of fields are steeper than 
        Field60 = arcpy.MakeTableView_management(SlopeSort, "Field60", '\"%s\" = %s' % (OID, Field60Count))
        for row in arcpy.da.SearchCursor(Field60, ["Slope75Pct"]):
            Slope60Value = row[0]

        # Calculate SlopeRank using automatically generated breaks (20-40-40)                               
        with arcpy.da.UpdateCursor(RunoffRiskTable, ["Slope75Pct", "SlopeRank"]) as cursor:
            for row in cursor:
                if row[0] >= Slope20Value:
                    row[1] = 'High'
                elif row[0] < Slope20Value and row[0] >= Slope60Value:
                    row[1] = 'Med'
                elif row[0] < Slope60Value:
                    row[1] = 'Low'
                cursor.updateRow(row)
                
        arcpy.Delete_management(SlopeSort)  
        del[SlopeSort]

    # -----------------------------SEDIMENT DELIVERY RANK------------------------------------------
    # Calculate breaks for High, Med, and Low using:
    arcpy.AddMessage("Classifying each field by sediment delivery ratio rank")

    # 1) user-defined values for breaks
    if HighSDR <> '':
        if MediumSDR == '':
            arcpy.AddError("Must provide both medium and high SDR values or neither...QUITTING!")
    if MediumSDR <> '':
        if HighSDR == '':
            arcpy.AddError("Must provide both medium and high SDR values or neither...QUITTING!")
    if HighSDR <> '' and MediumSDR <> '':     
        arcpy.AddMessage("User defined SDR breaks provided")
        HighSDR_fl = float(HighSDR)
        MedSDR_fl = float(MediumSDR)
        with arcpy.da.UpdateCursor(RunoffRiskTable, ["SDR", "SDRRank"]) as cursor:
            for row in cursor:
                if row[0] >= HighSDR_fl:
                    row[1] = 'High'
                elif row[0] < HighSDR_fl and row[0] >= MedSDR_fl:
                    row[1] = 'Med'
                elif row[0] < MedSDR_fl:
                    row[1] = 'Low'
                cursor.updateRow(row)

    # 2) automatically generated (using a 20-40-40 breakdown of fields)                       
    if HighSDR == '' and MediumSDR == '':
        arcpy.AddMessage("Automatically generated sediment delivery ratio breaks...calculating 20-40-40 breakdown of fields")
        
        # Sort RunoffRiskTable by SDR in a descending order
        SDRSorted = arcpy.Sort_management(RunoffRiskTable, "in_memory" + "\\SDRSorted", [["SDR", "DESCENDING"]])
        desc = arcpy.Describe(SDRSorted)
        OID = desc.OIDFieldName

        # Identify the 3rd quartile sediment delivery ratio value that 20% of fields are higher than
        Field20 = arcpy.MakeTableView_management(SDRSorted, "Field20", '\"%s\" = %s' % (OID, Field20Count))
        for row in arcpy.da.SearchCursor(Field20, ["SDR"]):
            SDR20Value = row[0]

        # Identify the 3rd quartile sediment delivery ratio value that 60% of fields are higher than
        Field60 = arcpy.MakeTableView_management(SDRSorted, "Field60", '\"%s\" = %s' % (OID, Field60Count))
        for row in arcpy.da.SearchCursor(Field60, ["SDR"]):
            SDR60Value = row[0]

        # Calculate SedimentDeliveryRank using automatically generated breaks (20-40-40)
        with arcpy.da.UpdateCursor(RunoffRiskTable, ["SDR", "AdjDTS", "SDRRank"]) as cursor:                            
            for row in cursor:
                if row[0] >=  SDR20Value:
                    row[2] = 'High'
                elif row[0] < SDR20Value and row[0] >= SDR60Value:
                    row[2] = 'Med'
                elif row[0] < SDR60Value:
                    row[2] = 'Low'
                if row[1] >= 5000:
                    row[2] = 'Low'
                if row[0] == 9999999:
                    row[2] = 'NA'
                cursor.updateRow(row)
                
        arcpy.Delete_management(SDRSorted)   
        del[SDRSorted]

    # --------------------------CALCULATE RUNOFF RISK-------------------------------------------
    # based on cross classification of SDRRank and SlopeRank
    arcpy.AddMessage("Calculating Runoff Risk")
    
    arcpy.AddField_management(RunoffRiskTable, "RunoffRisk", "TEXT")
    with arcpy.da.UpdateCursor(RunoffRiskTable, ["SDRRank", "SlopeRank", "RunoffRisk"]) as cursor:
        for row in cursor:
            if row[0] == 'High' and row[1] == 'High':
                row[2] = 'A-Critical'
            elif row[0] == 'High' and row[1] == 'Med' or \
                 row[0] == 'Med' and row[1] == 'High':
                row[2] = 'B-VeryHigh'
            elif row[0] == 'Med' and row[1] == 'Med' or \
                          row[0] == 'Low' and row[1] == 'High' or \
                          row[0] == 'High' and row[1] == 'Low':
                row[2] = 'C-High'
            elif (row[0] == 'Low' and row[1] == 'Med') or \
                 (row[0] == 'Med' and row[1] == 'Low') or \
                 (row[0] == 'Low' and row[1] == 'Low'):
                row[2] = 'D-Present'
            else:
                row[2] = 'NA'
            cursor.updateRow(row)
            
    arcpy.Delete_management("in_memory")


                    
##------------------------------------------------------------------------------
##------------------------------------------------------------------------------
if __name__ == "__main__":
    
    # Define parameters
    FBIn = arcpy.GetParameterAsText(0)
    # Test for a join on FBIn
    basename = arcpy.Describe(FBIn).baseName
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        basename = arcpy.Describe(FBIn).baseName
        arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    SlopeTable = arcpy.GetParameterAsText(1)
    DistanceToStream = arcpy.GetParameterAsText(2)
    HighSlope = arcpy.GetParameterAsText(3)
    MediumSlope = arcpy.GetParameterAsText(4)
    HighSDR = arcpy.GetParameterAsText(5)
    MediumSDR = arcpy.GetParameterAsText(6)
    OutRunoffRisk = arcpy.GetParameterAsText(7)
        
    # Set extent, snap, workspace
    arcpy.env.extent = DistanceToStream
    arcpy.env.snapRaster = DistanceToStream
    rDesc = arcpy.Describe(DistanceToStream)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True

    # Run Modules------------------------------------------------------------------------------
    RunoffRiskAssessment(FBIn, SlopeTable, DistanceToStream)


