# Stream Reach and Catchments------------------------------------------------------------------------------------------------------------
#
# Define stream reach and catchments in a HUC12 watershed
# Input is a flow network polyline, a flow direction raster and a DEMFill raster
#
# Input feature class should have the "StreamType" field attributed - this field can be
# used to limit the segements that are identified from which to generate stream reach and catchments
#
# The stream network is converted to a raster (maintaining snap raster and extent of orginal DEM) for TauDEM processing
# TauDEM processing is performed to generate stream reaches and watersheds
# watersheds (catchments) are converted to vector format (polygon)
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")


# Modules ------------------------------------------------------------------------


def LimitByField(FlowNetwork, LimitField, ClassValue):
    # Define the criteria for limiting stream reach and catchment generation - based on a field input FlowNetwork
    
    if LimitField <> '':
        # There is a field that will be used to refine stream reach and catchment generation
        if ClassValue == '':
            
            # A field has been specified but no values were chosen --- QUIT!
            arcpy.AddError("Must select classification value...QUITTING!!!")
            sys.exit(0)
            
        elif ClassValue <> '':
            arcpy.AddMessage("Stream reach and catchments to be generated for %s values of %s" % (LimitField, ClassValue))
            
            # Add field delimiters to the LimitField
            Field_type = arcpy.ListFields(FlowNetwork, "%s" % LimitField)[0].type
            Field_del = arcpy.AddFieldDelimiters(arcpy.Describe(FlowNetwork).path, "%s" % LimitField)

            # If classification values are text
            if str(Field_type) == 'String':
                ValueList = []
                for value in ClassValue.split(';'):
                    if value not in ValueList:
                        ValueList.append(value)

                arcpy.AddMessage("field is string")
                ValueList = ["'%s'" % value for value in ValueList]
                whereClause = "%s IN(%s)" % (Field_del, ', '.join(map(str, ValueList)))
                
                # Apply query slection to flow network - reach
                reach = arcpy.MakeFeatureLayer_management(FlowNetwork, "reach", whereClause)

            # If classification values are not text
            elif Field_type <> 'String':
                ValueList = []
                for value in ClassValue.split(';'):
                    value_int = int(value)
                    if value_int not in ValueList:
                        ValueList.append(value_int)
                        
                ValueList = ["%s" % value for value in ValueList]
                whereClause = "%s IN(%s)" % (Field_del, ', '.join(map(str, ValueList)))

                # Create feature layer from Flow Network - selection applied
                reach = arcpy.MakeFeatureLayer_management(FlowNetwork, "reach", whereClause)

    elif LimitField == '':
        # No field was provided from which to limit catchment and reach generation
        reach = arcpy.MakeFeatureLayer_management(FlowNetwork, "reach")

    return(reach)


def StreamNetworkAndCatchments(reach, D8FlowDir, DEMFill):

    # Convert reach to a raster
    desc = arcpy.Describe(reach)
    OID = desc.OIDFieldName
    strm_ras = arcpy.PolylineToRaster_conversion(reach, "%s" % OID, "strm_ras", "", "", cellsize)

    # Convert stream cells to 1 and background values to 0
    StrmNetRas = Con(strm_ras, 1, 0, "VALUE >= 0")

    # Reclassify D8 flow direction raster to TauDEM values ----> save to TDProcDir 
    TDFlowDir = Reclassify(D8FlowDir, "VALUE",
                           RemapRange([[1, 1], [128, 2], [64, 3], [32, 4], [16, 5], [8, 6], [4, 7], [2, 8]]))

    # Copy 1)TDFlowDir, 2)DEMFill, and 3)StreamNetRas to TDProcDir as uncompressed tiffs (in env settings)
    arcpy.CopyRaster_management(TDFlowDir, TDProcDir + "\\TDFlowDir.tif")
    arcpy.CopyRaster_management(DEMFill, TDProcDir + "\\TDFill.tif")
    arcpy.CopyRaster_management(StrmNetRas, TDProcDir + "\\TDStreamNetRas.tif")

    core_cnt = multiprocessing.cpu_count()
    core_no = core_cnt - 1
    
    # Create a TauDEM Flow Accumulation raster
    callstr = 'mpiexec -n %s AreaD8 ' % core_no + \
              ' -p ' + TDProcDir + "\\TDFlowDir.tif" + \
              ' -ad8 ' + TDProcDir + "\\TDFlowAccumulation.tif -nc"
    call(callstr, shell=True)
    arcpy.CalculateStatistics_management(TDProcDir + "\\TDFlowAccumulation.tif")

    if PourPoints == '':
        arcpy.AddMessage("Generating Stream Reach and Catchments")
        arcpy.AddMessage("No pour point(s) provided")
        
        # Construct command for Stream Reach, Watersheds, etc.
        callstr = 'mpiexec -n %s StreamNet ' % core_no + \
                  ' -fel ' + TDProcDir + "\\TDFill.tif" + \
                  ' -p ' + TDProcDir + "\\TDFlowDir.tif" + \
                  ' -ad8 ' + TDProcDir + "\\TDFlowAccumulation.tif" + \
                  ' -src ' + TDProcDir + "\\TDStreamNetRas.tif" + \
                  ' -ord ' + TDProcDir + "\\TDStreamOrder.tif" + \
                  ' -tree ' + TDProcDir + "\\TDNetworkTree.txt" + \
                  ' -coord ' + TDProcDir + "\\TDNeworkCoord.txt" + \
                  ' -net ' + TDProcDir + "\\TDStreamReach.shp" + \
                  ' -w ' + TDProcDir + "\\TDWatersheds.tif"

        call(callstr, shell=True)
        TDWatersheds = arcpy.CalculateStatistics_management(TDProcDir + "\\TDWatersheds.tif")
        TDStreamReach = arcpy.CalculateStatistics_management(TDProcDir + "\\TDStreamReach.shp")

        # Copy TDStreamReach to OutputStreamReach
        arcpy.CopyFeatures_management(TDStreamReach, OutputStreamReach)

        # Eliminate small segments (less than 30 m^2) in catchments layer - output to OutputCatchments
        catch_poly = arcpy.RasterToPolygon_conversion(TDWatersheds, "catch_poly")
        arcpy.AddField_management(catch_poly, "Area_m2", "FLOAT")
        arcpy.CalculateField_management(catch_poly, "Area_m2", "!shape.area@squaremeters!", "PYTHON")
        
        catch_lyr = arcpy.MakeFeatureLayer_management(catch_poly, "catch_lyr")
        arcpy.SelectLayerByAttribute_management(catch_lyr, "NEW_SELECTION", "\"Area_m2\" <= 30")
        catchments = arcpy.Eliminate_management(catch_lyr, OutputCatchments)
        arcpy.DeleteField_management(catch_poly, ["Area_m2"])

    elif PourPoints <> '':
        arcpy.AddMessage("Pour Point Provided")

        # Copy pour point to TDProcDir folder
        PourPnts = arcpy.CopyFeatures_management(PourPoints, TDProcDir + "\\PourPnts.shp")

        # TauDEM requires an ID field to exist in PourPoints feature class --> add and populate with OID
        arcpy.AddField_management(PourPnts, "ID")
        desc = arcpy.Describe(PourPnts)
        OID = desc.OIDFieldName 
        arcpy.CalculateField_management(PourPnts, "ID", "!%s!" % OID, "PYTHON")

        # Construct command for Stream Reach, Watersheds, etc.
        arcpy.AddMessage("Generating Stream Reach and Catchments")
        callstr = 'mpiexec -n %s StreamNet ' % core_no + \
                  ' -fel ' + TDProcDir + "\\TDFill.tif" + \
                  ' -p ' + TDProcDir + "\\TDFlowDir.tif" + \
                  ' -ad8 ' + TDProcDir + "\\TDFlowAccumulation.tif" + \
                  ' -src ' + TDProcDir + "\\TDStreamNetRas.tif" + \
                  ' -o ' + TDProcDir + "\\PourPnts.shp" + \
                  ' -ord ' + TDProcDir + "\\TDStreamOrder.tif" + \
                  ' -tree ' + TDProcDir + "\\TDNetworkTree.txt" + \
                  ' -coord ' + TDProcDir + "\\TDNeworkCoord.txt" + \
                  ' -net ' + TDProcDir + "\\TDStreamReach.shp" + \
                  ' -w ' + TDProcDir + "\\TDWatersheds.tif"

        call(callstr, shell=True)
        TDWatersheds = arcpy.CalculateStatistics_management(TDProcDir + "\\TDWatersheds.tif")
        TDStreamReach = arcpy.CalculateStatistics_management(TDProcDir + "\\TDStreamReach.shp")

        # Copy TDStreamReach and TDWatersheds
        arcpy.CopyFeatures_management(TDStreamReach, OutputStreamReach)

        # Eliminate small segments (less than 30 m^2) in catchments layer - output to OutputCatchments
        catch_poly = arcpy.RasterToPolygon_conversion(TDWatersheds, "catch_poly")
        arcpy.AddField_management(catch_poly, "Area_m2", "FLOAT")
        arcpy.CalculateField_management(catch_poly, "Area_m2", "!shape.area@squaremeters!", "PYTHON")
        
        catch_lyr = arcpy.MakeFeatureLayer_management(catch_poly, "catch_lyr")
        arcpy.SelectLayerByAttribute_management(catch_lyr, "NEW_SELECTION", "\"Area_m2\" <= 30")
        catchments = arcpy.Eliminate_management(catch_lyr, OutputCatchments)
        arcpy.DeleteField_management(catch_poly, ["Area_m2"])

    if NewWatershed == 'true':
        if OutWatershed == '':
            arcpy.AddError("Generate New Watershed is selected but no output file location is provided...QUITTING!!!")
            sys.exit(0)
        elif OutWatershed <> '':
            arcpy.AddMessage("Generating new watershed boundary...dissolving catchments centered on current watershed boundary")
            catch_lyr = arcpy.MakeFeatureLayer_management(catchments, "catch_lyr")
            arcpy.SelectLayerByLocation_management("catch_lyr", "HAVE_THEIR_CENTER_IN", WBD)
            
            New_WBD = arcpy.Dissolve_management(catch_lyr, OutWatershed)
            arcpy.AddField_management(New_WBD, "ACRES", "Double")
            arcpy.CalculateField_management(New_WBD, "ACRES", "!shape.area@acres!", "PYTHON")
            
    # Cleanup
    arcpy.Delete_management(reach)
    arcpy.Delete_management(strm_ras)
    arcpy.Delete_management(StrmNetRas)
    arcpy.Delete_management(TDFlowDir)
    arcpy.Delete_management(catch_poly)
    del[reach, strm_ras, StrmNetRas, TDFlowDir, catch_poly]
    arcpy.Delete_management(TDProcDir)


if __name__ == "__main__":

    # Define Parameters
    FlowNetwork = arcpy.GetParameterAsText(0)
    D8FlowDir = arcpy.GetParameterAsText(1)
    DEMFill = arcpy.GetParameterAsText(2)
    WBD = arcpy.GetParameterAsText(3)
    LimitField = arcpy.GetParameterAsText(4)
    ClassValue = arcpy.GetParameterAsText(5)
    PourPoints = arcpy.GetParameterAsText(6)
    OutputStreamReach = arcpy.GetParameterAsText(7)
    OutputCatchments = arcpy.GetParameterAsText(8)
    NewWatershed = arcpy.GetParameterAsText(9)
    OutWatershed = arcpy.GetParameterAsText(10)

    # Set extent, snap, workspace
    arcpy.env.extent = D8FlowDir
    arcpy.env.snapRaster = D8FlowDir
    arcpy.env.outputCoordinateSystem = D8FlowDir
    rDesc = arcpy.Describe(D8FlowDir)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.compression = "NONE" 

    # find resolution of input D8FlowDir 
    cellsize = float(arcpy.GetRasterProperties_management(D8FlowDir, "CELLSIZEX").getOutput(0))

    # Create and define folder for TauDEM outputs (in ProcDir) - will be deleted at end of script
    scriptpath = sys.path[0]
    if ' ' in scriptpath:
        arcpy.AddError("space exists in %s that must be removed from path name...QUITTING!!!" % scriptpath)
        sys.exit(0)
    else:
        arcpy.CreateFolder_management(scriptpath, "TDProcDir")
        TDProcDir = scriptpath + "\\TDProcDir"

    # Run Modules
    reach = LimitByField(FlowNetwork, LimitField, ClassValue)
    
    StreamNetworkAndCatchments(reach, D8FlowDir, DEMFill)

