# TileDrainageClassification---------------------------------------------------------------------------------
#
# The tile drainage determination tool estimates which fields are likely to be tile drained based on a
# combination of by-field slope and soils information.
#
# The output is a drainage table containing byfield slope and soils information and a drainage classification (YES, NO, or NonAg).
#
# A field is classified as tile-drained if it meets the criteria of a user-selected query
# The user must select one of two possible queries:
#
# Query 1:  >= 90% of field is less than 5% slope AND/OR >= 10% of field consists of hydric soils
# Query 2:  >= 90% of field is less than 5% slope AND/OR >= 40% of field consists of a dual drainage class

# The By-Field slope statistics should be run prior to the TileDrainage determination tool
#
# Attributes of drainage table:
# 
# PctAllHyd: % of field consisting of hydric soils
# PctPartHyd: % of field consisting of partially hydric soils
# PctDualDrg: % of field consisting of dual drainage class or D class soil
# PctSlp_lt5: % of field consisting of slopes less than 5%
# Drained: YES (tile-drained) NO (non tile-drained), or Non-ag (non-agricultural)
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 04/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time
arcpy.CheckOutExtension("Spatial")


## Modules -----------------------------------------------------------------------

def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True


def CreateSoilsMasks(gSSURGO):
    # FIND THE PERCENTAGE OF EACH FIELD UNDER VARIOUS SOIL CONDITIONS:
    # 1) % hydric
    # 2) % dual drainage class

    arcpy.AddMessage("Creating mask of hydric soils")
    # 1) Generate raster where each pixel represents % of hydric soil - from gSSURGO "hydric" field
    HydricPct = Lookup(gSSURGO, "Hydric")

    arcpy.AddMessage("Creating mask of dual drainage class soils or D soils")
    # 2) Create a raster of "dual drainage class" soils - convert background nodata values to 0, all else to 1
    DualDrainageClass = Con(gSSURGO, 1, 0, "\"HydroGrp\" = 'A/D' OR \"HydroGrp\" = 'B/D' OR \"HydroGrp\" = 'C/D' OR \"HydroGrp\" = 'D'")

    return(HydricPct, DualDrainageClass)


def DrainageClassification(FBIn, HydricPct, DualDrainageClass, SlopeTable):
    arcpy.AddMessage("Creating tile drainage table")
    
    arcpy.AddMessage("Extracting soils information to fields")
    # Create draintable as output of Zonal Stats with "HydricPct": "MEAN" percent hydric
    DrainageTable = ZonalStatisticsAsTable(FBIn, "FBndID", HydricPct, OutDrainageTable, "#", "MEAN")
    
    # Perform zonal stats with Dual Drainage Class; "SUM" = # of pixels occupied by dual drainage class
    DualDrainageTable = ZonalStatisticsAsTable(FBIn, "FBndID", DualDrainageClass, "in_memory" + "\\ddrg_table", "#", "SUM")
    arcpy.JoinField_management(DrainageTable, "FBndID", DualDrainageTable, "FBndID", ["SUM"])

    arcpy.AddMessage("Extracting slope information to fields")
    # Join fields from slope table - to find the % of each field less than 5% slope
    arcpy.JoinField_management(DrainageTable, "FBndID", SlopeTable, "FBndID", ["Pct_lt1", "Pct1_2", "Pct2_5"])

    # Join "isAG" field from FBndID to determine if a field is agricultural
    arcpy.JoinField_management(DrainageTable, "FBndID", FBIn, "FBndID", ["isAG"])   

    # Add fields to DrainageTable
    arcpy.AddField_management(DrainageTable, "PctHyd", "FLOAT")
    arcpy.AddField_management(DrainageTable, "PctDualDrg", "FLOAT")
    arcpy.AddField_management(DrainageTable, "PctSlp_lt5", "FLOAT")
    arcpy.AddField_management(DrainageTable, "Drained", "TEXT")

    # QUERY 1: Tile-drained if >= 90% of the field is <= 5% slope AND >= 10% of field consists of hydric soils
    if QuerySelection == '">= 90% of field is less than 5% slope OR >= 10% of field consists of hydric soils"':
        arcpy.AddMessage("Query 1 selected...field is tile_drained if 90% of field is less than 5% slope OR >= 10% of field consists of hydric soils")
        with arcpy.da.UpdateCursor(DrainageTable, ["MEAN", "SUM", "COUNT", "PctHyd", "PctDualDrg", "Pct_lt1", "Pct1_2", "Pct2_5", "PctSlp_lt5", "Drained", "isAG"]) as cursor:
            for row in cursor:  
                if row[0] == None:
                    row[3] = 0
                else:
                    row[3] = row[0]
                if row[1] == None:
                    row[4] = 0
                else:
                    row[4] = (row[1] / row[2]) * 100
                if row[5] == None:
                    row[8] = 0
                if row[6] == None:
                    row[8] = 0
                if row[7] == None:
                    row[8] = 0
                if row[5] <> None and row[6] <> None and row[7] <> None:
                    row[8] = row[5] + row[6] + row[7]
                if row[10] >= 1 and (row[8] >= 90 or row[3] >= 10):
                    row[9] = "YES"
                elif row[10] >= 1 and row[8] < 90 and row[3] < 10:
                    row[9] = "NO"
                elif row[10] == 0:
                    row[9] = "NonAg"
                cursor.updateRow(row)

    # QUERY 2: Tile-drained if >= 90% of the field is <= 5% slope OR >= 40% of field consists of dual drainage class (A/D, B/D, C/D)        
    elif QuerySelection == '">= 90% of field less than 5% slope OR >= 40% of field consists of dual drainage class (A/D, B/D, C/D) or D class soils"':
        arcpy.AddMessage("Query 2 selected...field is tile drained if 90% of field is less than 5% slope OR >= 40% of field consists of dual drainage class (A/D, B/D, C/D) or D class soils")
        with arcpy.da.UpdateCursor(DrainageTable, ["MEAN", "SUM", "COUNT", "PctHyd", "PctDualDrg", "Pct_lt1", "Pct1_2", "Pct2_5", "PctSlp_lt5", "Drained", "isAG"]) as cursor:
            for row in cursor:
                if row[0] == None:
                    row[3] = 0
                else:
                    row[3] = row[0]
                if row[1] == None:
                    row[4] = 0
                else:
                    row[4] = (row[1] / row[2]) * 100
                if row[5] == None:
                    row[8] = 0
                if row[6] == None:
                    row[8] = 0
                if row[7] == None:
                    row[8] = 0
                elif row[5] <> None and row[6] <> None and row[7] <> None:
                    row[8] = row[5] + row[6] + row[7]
                if row[10] >= 1 and (row[8] >= 90 or row[4] >= 40):
                    row[9] = "YES"
                elif row[10] >= 1 and row[8] < 90 and row[4] < 40:
                    row[9] = "NO"
                elif row[10] == 0:
                    row[9] = "NonAg"
                cursor.updateRow(row)

    arcpy.DeleteField_management(DrainageTable, ["ZONE_CODE", "AREA", "MEAN", "SUM", "Pct_lt1", "Pct1_2", "Pct2_5", "COUNT"])

    # Cleanup
    arcpy.Delete_management(HydricPct)
    arcpy.Delete_management(DualDrainageClass)
    arcpy.Delete_management(DualDrainageTable)
    del[HydricPct, DualDrainageClass, DualDrainageTable]
    arcpy.Delete_management("in_memory")
 
    
if __name__ == "__main__":

    # Define parameters
    FBIn = arcpy.GetParameterAsText(0)
    # Test for a join on FBIn
    basename = arcpy.Describe(FBIn).baseName
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        basename = arcpy.Describe(FBIn).baseName
        arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    gSSURGO = arcpy.GetParameterAsText(1)
    SlopeTable = arcpy.GetParameterAsText(2)
    QuerySelection = arcpy.GetParameterAsText(3)
    OutDrainageTable = arcpy.GetParameterAsText(4)

    # Set environments
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder

    # Run Modules
    HydricPct, DualDrainageClass = CreateSoilsMasks(gSSURGO)

    DrainageClassification(FBIn, HydricPct, DualDrainageClass, SlopeTable)              
