###
# PitFill_HolePunch.py
#
# Created to process state of Iowa Lidar data and create an elevation
#  model that represents the land surface with enhanced hydrologic flow.
#  Enhanced flow is enabled by filling pits, artifacts, and shallow basins.
#
#  The program was written to circumvent ESRI's Spatial Analyst Fill
#  programs and is, in essence, a fill to specified depth method.
#
#  Pits (one-cell-sinks) are ignoed by ESRI's Sink command. These are removed
#   by setting ESRI Sinks to NULL and Filling the rest. NOTE: in some cases,
#   Pits may be filled to a depth greater than the specified maximum fill depth.
#
#  The HolePunching modules iterates through a process that remove deep sinks
#  from the filling process until no filled sinks are deeper than a stated 
#  threshhold. The remaining sinks (punched holes) are set to Null...
#  
#  Once the process is complete, the orginal DEM data are used to replace the
#  punched holes.
#
#-----------------
# original coding: BK Gelder, Iowa State University 01/2011
#  modified:
#  - for use with ArcGIS 10 & arcpy, DE James USDA/ARS 08/2011
#------------------------------------------------------------------------
# 1/19/2012 - removed compression env setting;
#             changed basin depth test to '>' from '>=', if = then process fails
#
# 1/25/2012 - re-write based on BG original
#             Find all filled areas (Depth gt 0), set to 1...
#             original - rDepth_cost = Con(rDepth > 0, 1)  
#             new - rDepth_cost = Con(rDepth > 0, 0, 1)
#
# 02.08.2012 - add zunit test (integer or float) to insure the correct output
#
# 06.13.2012 - add exporting of combined fill regions to support future processing
#              in cutter
#
# 09.19.2014 - add the Pit-filler & make stand alone for a single DEM
#------------------------------------------------------------------------

# Import system modules
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, time

# Set extensions & environments 
arcpy.CheckOutExtension("Spatial")
env.overwriteOutput = True

##-------------------------------------------------------------------------
## ---- Modules ----
def fillOCSinks(rawDEM):
        # Return a raster will all one-cell-sinks filled
        arcpy.AddMessage("Find Pits...")
        sinkFDir = FlowDirection(rawDEM)
        allSinks = Sink(sinkFDir)
                                                                              
        arcpy.AddMessage("Fill everything else...")

        ## Make a No-one-cell-sink DEM
        AllButSinks_DEM = Con(IsNull(allSinks), rawDEM)

        ## Fill the No-one-cell-sink DEM
        absDEM_fill = Fill(AllButSinks_DEM)

        ## Add the Original 'real' sinks back into the filled DEM
        inDEM = Con(IsNull(absDEM_fill), rawDEM, absDEM_fill)

        return(inDEM)


## from djHolePuncher.py
def findMaxDepth(inDEM):
        ## Determine how deep the areas that need fill are
        rFill = Fill(inDEM)

        rDepth = Minus(rFill, inDEM)
     
        MaxDepth = rDepth.maximum
        arcpy.AddMessage("Current maximum depth: " + str(MaxDepth))

        return(rDepth, MaxDepth)


def cutHolesByDepth(inDEM, rDepth, minBsnDpth, cnt):
        #### Punch holes in the bottom of sinks that are deeper than the threshhold
        ####  - inDEM: the current verion of the surface
        ####  - rDepth: the current depth of fill
        ####  - minBsnDepth: the depth of allowed sinks 
        arcpy.AddMessage("Punch holes...")        
        
        ## Find all filled areas (Depth gt 0), set to 1...the surface
        ##   area with no depth require no action -- Costdistance 'Cost' Raster
        rDepth_cost = Con(rDepth, 0, 1, "VALUE > 0")       
        
        ## Find the filled areas deeper than the threshhold, set to 1,
        ##   these are the sinks of interest - Costdistance 'Source' Raster
        rDepth_tsh = Con(rDepth > minBsnDpth, 1)       
        
        ## create the Cost Distance or distance to allowed Sink (LT Threshhold)
        ##   CostDistance is used becoase it relies on connectivity not raw distance
        rCostDist = CostDistance(rDepth_tsh, rDepth_cost)
        
        ## Create a binary surface (True/False) of the CostDistance surface
        ##   0 = w/in threhhold depth, 1 = deeper than threshhold
        rCostDist_tf = Con(rCostDist == 0, 1, 0)
		
        ## Create a Group from each connected Region -- sink regions
        rg8 = RegionGroup(rCostDist_tf, "EIGHT")
            
        ## Only concerned with that portion of the Region deeper than the threshhold
        ## Added stuff to code each fill region uniquely - BKG Jun 11 2012
        rg8lk1 = ExtractByAttributes(rg8, "LINK = 1")
        fillRG8 = Con(IsNull(rg8lk1), 0, rg8lk1)
                  
        ## Find the minimum elevation within the region
        rg8lk1_min = ZonalStatistics(rg8lk1, "VALUE", inDEM,  "MINIMUM")
               
        ## Calculate the inRegion depth
        rg8lk1_dif = Minus(inDEM, rg8lk1_min)
               
        ## Reclassify the depth... if it is 0, then its the minimum so punch a NoData hole
        ##  if not keep and make it the inDEM
        ##  Make sure to include the previously punched old holes 
        
        hole_is0 = Con(IsNull(rg8lk1_dif), 1, Con(rg8lk1_dif == 0, 0, 1))
        oldHoles = Con(IsNull(inDEM), 0, 1) 
        allHoles = BooleanAnd(hole_is0, oldHoles)
            
        ## Count the number of holes punched        
        rows = arcpy.SearchCursor(allHoles, "VALUE = 0")
        for row in rows:
                numTotHoles = row.getValue("COUNT")
                arcpy.AddMessage("Total holes punched: " + str(numTotHoles))
        
        ## Make a new DEM with Holes punched out the bottom of each region      
        newDEMwHoles = Con(allHoles > 0, inDEM)
        
        return(newDEMwHoles)
        
    
def fill_toDepth(inDEM, minBsnDpth):

        rDepth, MaxDepth = findMaxDepth(inDEM)
        
        cnt = 0

        while MaxDepth > minBsnDpth:
                NewDEMwHoles = cutHolesByDepth(inDEM, rDepth, minBsnDpth, cnt)
                
                inDEM = NewDEMwHoles 
                
                rDepth, MaxDepth = findMaxDepth(inDEM)
                
                cnt = cnt + 1

        return(NewDEMwHoles)


def SaveOutput(NewDEMwHoles, inDEM, outDEM):
        arcpy.AddMessage("Saving Output...")
        
        # Fill shallow sinks (<threshhold), holes remain...but no sinks
        FilledDEMwHoles = Fill(NewDEMwHoles)
        
        
        # Fill the punched holes with the original DEM so no Holes remian
        FilledPunched = Con(IsNull(FilledDEMwHoles), inDEM, FilledDEMwHoles)   
        FilledPunched.save(outDEM)
        

        
       
##------------------------------------------------------------------------------
##------------------------------------------------------------------------------


if __name__ == "__main__":
	rawDEM = arcpy.GetParameterAsText(0)
	outDEM = arcpy.GetParameterAsText(1)
	maxFill = arcpy.GetParameterAsText(2)

	minBsnDpth = float(maxFill)	
        
	desc = arcpy.Describe(rawDEM)
	if desc.isInteger and minBsnDpth < 1:
		arcpy.AddMessage("invalid Max Fill Depth for CM Z-units: %s" %(str(minBsnDpth)))
		sys.exit(0)
	if not desc.isInteger and minBsnDpth > 1:
		arcpy.AddMessage("invalid Max Fill Depth for M Z-units: %s" %(str(minBsnDpth)))
		sys.exit(0)
        	                
	# Process through puncher 3m
	     
	arcpy.AddMessage("Hole-punching: " + rawDEM)
	
	# Temporary raster processing directory
	env.scratchWorkspace = env.scratchFolder
	
	if not arcpy.Exists(rawDEM):
		sys.exit("invalid input DEM...")
	        
	inDEM = fillOCSinks(rawDEM)
	
	# Punch the holes
	NewDEMwHoles = fill_toDepth(inDEM, minBsnDpth)
	
	# Save
	SaveOutput(NewDEMwHoles, inDEM, outDEM)
	
	arcpy.AddMessage(outDEM)
	arcpy.BuildPyramids_management(outDEM)


	env.scratchWorkspace = ""



    


