# -----------------------------------------------------------------------------
# updateLU6_newFB.py
# 
#  The Update Edited Field Boundaries tool will rebuild the land-use lookup 
#  tables based on the contents of an edited field boundary feature class, 
#  using the NASS CDL rasters present in the watershed's file geodatabase.
#  Included are:
#   - the CH_<HUC12> table - a summary of NASS CDL data by field for all 
#     land use raster available in the fiel-geodatabase
#   - the LU6_<HUC12> table - a 6 year summary of field land use
#   - crop rotation string (CropRotatn) using a single letter to represent 
#     each year's majpority crop type..additonally, if the majority is Corn
#     or soyBean and the majority is LT 75%, check to see if the second
#     highest class is soyBean or Corn, if true assign as corn/soybean (D)
#   - crop rotation summary (Cropsumry) a string using the single letter
#     of the majority crop followed by the number of occurances of that ctop 
#     over the rotation span
#   - a count of the number of occurances of corn-after-corn 'CC' in the 
#     crop rotation string - CCCount
#   - a count of the number of occurances of the majority crop percent of the 
#     field less than 75% - MixCount
#   - for each year in the span, the majority crop (majYR) and the percent 
#     of the field that it represents (pctYR)
#   - a General Land Use assignment based on the 6-yr crop rotation string and
#     the other derived fields.  
# -----------------------------------------------------------------------------
#  Orginal coding: D.James 08/2012
#    - 07/2013: update to 6-year rotation
#    - 11/2013: update to extend Pasture to include pasture woodlots; add class
#        for flood-prone cropland (1 year of Ag field as water)
#    - 04/2014: updated to ad the full crop history to 2000, table name: CH _inHUC
#    - 04/2014: made the full commitment to in_memory processing, reduce proc time by 60%!
#    - 08/2015: Repurpose the original script to update an edited field boundary 
#               feature class.
#    - BUG! 10/2015 FBndID should have a preceeding 'F' not 'FB'- Line 101
#	 - 03/2016: remove the MixedOnly processing to exclude the c/s mixed landuse
# -----------------------------------------------------------------------------

# Import system modules
import arcpy
from arcpy import env
from arcpy.sa import *

import sys, string, os
from operator import itemgetter, attrgetter

env.overwriteOutput = True

# Set extensions & environments 
arcpy.CheckOutExtension("Spatial")


##--------------------------------------------------------------------------------------
## Modules
## 

def mkOutputFrame(FBedit, StatFrame, TempFrame, inHUC):
	# Create the output frame
	if arcpy.Exists(TempFrame):
		arcpy.Delete_management(TempFrame)
	arcpy.CopyFeatures_management(FBedit, TempFrame)

	
	#Remove all unnecessary fields
	reqFields = ['OBJECTID', 'Shape', 'Shape_Length', 'Shape_Area']
	delFields = []
	field_names = [f.name for f in arcpy.ListFields(TempFrame)]
	for f in field_names:
		if f not in reqFields:
			delFields.append(f)
	arcpy.DeleteField_management(TempFrame, delFields)
	
	# Add core fields
	arcpy.AddField_management(TempFrame, "FBndID", "text", 255)
	arcpy.AddField_management(TempFrame, "Acres", "float")
	arcpy.AddField_management(TempFrame, "isAG", "short")

	# add the LU6 fields
	arcpy.AddField_management(TempFrame, "GenLU", "text", 25)
	arcpy.AddField_management(TempFrame, "CropRotatn", "text", 20)
	arcpy.AddField_management(TempFrame, "CropSumry", "text", 20)
	arcpy.AddField_management(TempFrame, "CCCount", "text", 8)
	arcpy.AddField_management(TempFrame, "MixCount", "text", 8)
	
	# Calculate intitial values, except FBndID
	arcpy.CalculateField_management(TempFrame, "Acres", "!shape.area@acres!", "PYTHON_9.3")
	arcpy.CalculateField_management(TempFrame, "isAG", "0", "PYTHON_9.3")
	arcpy.CalculateField_management(TempFrame, "CropRotatn", "''", "PYTHON_9.3")
	
	# No fields LT 2.5 acres
	arcpy.MakeFeatureLayer_management(TempFrame, "TempLayer")
	arcpy.SelectLayerByAttribute_management("TempLayer", "NEW_SELECTION", '"Acres" < 2.5')
	arcpy.Eliminate_management("TempLayer", StatFrame, "AREA")

	arcpy.MakeFeatureLayer_management(StatFrame, "TempLayer2")
	arcpy.SelectLayerByAttribute_management("TempLayer2", "NEW_SELECTION", '"Acres" < 2.5')
	arcpy.DeleteFeatures_management("TempLayer2")

	# Update the edited FB to unique FBndID -- after the edits
	Desc = arcpy.Describe(StatFrame)
	OID = Desc.OIDFieldName
	calcFBStr = "F%s_!%s!" %(inHUC, OID)
	arcpy.CalculateField_management(StatFrame,"FBndID", "'%s'" %(calcFBStr), "PYTHON_9.3") 

	arcpy.Delete_management("TempLayer")
	arcpy.Delete_management("TempLayer2")
	if arcpy.Exists(TempFrame):
		arcpy.Delete_management(TempFrame)

	

def ProcByYear(StatFrame, FileGDB, inHUC, YrList):
	# Process the new framework through each year
	
	codeblock = """def realDiv(fld,val):
                                return fld / val * 100"""
	expression = "realDiv(float(!COUNT!), float(!COUNT_1!))"                                  

	# for each year, add the majority crop ROTVAL and majority percent
	#  unless of course, there is not data for that year...
	for Year in YrList:
		Yr = Year[-2:]
			
		arcpy.AddMessage("")
		arcpy.AddMessage("Assigning landuse for 20" + Yr + "..."  )		

		YrFld = "maj" + Yr
		PctFld = "pct" + Yr
		MxPctFld = "MixedPct" + Yr
		CDLFld = "wsCDL20" + Yr 
		#CDLroot = r"D:\Data\EDF_ACPF\ACPF_NASS CDL\acpf_NASScdl.gdb\acpf_CDL20"
		wsCDL = Raster("wsCDL20" + Yr)
                
		ZonalSt_CDL = "ZonalSt_CDL" + Yr
		StatRaster = "StatRaster" + Yr
		cmbRaster = "cmbRaster" + Yr
		cmbTable = "cmbTable" + Yr
		
		#----------------------------------------------------------------------------------------------
		## Use the existing NASS CDL data 
		#----------------------------------------------------------------------------------------------
		env.snapRaster = wsCDL
		env.extent = FileGDB + "\\buf" + inHUC
		#bufMask = FileGDB + "\\buf" + inHUC 
		arcpy.AddMessage("Extract land use data...")
		
		#wsCDL = ExtractByMask(CDL_Data, bufMask)
		#if Yr in YrList[-6:]:	
		#wsCDL.save(FileGDB + "\\wsCDL20" + Yr)
		#env.extent = FileGDB + "\\wsCDL20" + Yr
		#env.snapRaster = FileGDB + "\\wsCDL20" + Yr
		env.extent = wsCDL
		env.snapRaster = wsCDL
			
		### Concatenate the crop rotation
		#----------------------------------------------------------------------------------------------
		# Process: Zonal Statistics as Table...
		# Add Majority and Percent fields
		#arcpy.AddMessage("Zonal Statistics")
		ZStat = ZonalStatisticsAsTable(StatFrame, "FBndID", wsCDL, ZonalSt_CDL, "Data") 
		
		arcpy.AddField_management(StatFrame, YrFld, "long")
		arcpy.AddField_management(StatFrame, PctFld, "long")
			
		
		### Determine crop mix percent by FBndID
		#----------------------------------------------------------------------------------------------
		# rasterize the crop frame
		arcpy.AddMessage('Calculating crop mix...')
		
		CellSz = wsCDL.meanCellWidth
		arcpy.FeatureToRaster_conversion(StatFrame, "FBndID", StatRaster, CellSz)
		
		# combine the crop frame raster and the CDL data
		#arcpy.AddMessage('Combining Data...')
		cmbList = [wsCDL, StatRaster]	
		cmbRaster = Combine(cmbList)
		cmbRaster.save(FileGDB + "\\cmbRaster" + Yr)
		
		# convert the raster attributes to a table
		cmbTable = arcpy.TableToTable_conversion(cmbRaster, "in_memory", "cmbTable")
		arcpy.AddField_management(cmbTable, MxPctFld, "long")
		
		# Get the Frame count
		arcpy.AddMessage('Join & Calculate Fields...')
		arcpy.JoinField_management(cmbTable, "STATRASTER" + Yr, StatRaster, "Value", ["COUNT", "FBndID"])
		
		# Calculate the field percent
		arcpy.CalculateField_management(cmbTable, MxPctFld, expression, "PYTHON" ,codeblock)
	
	
		### Assign the Majority to the frame records
		#----------------------------------------------------------------------------------------------
		arcpy.AddMessage('Assign Majority...')
		arcpy.JoinField_management(cmbTable, "FBndID", ZonalSt_CDL, "FBndID", "MAJORITY")
		
		
		# Identify largest Majority Percent crop and delete the rest
		#       Create a table view 
		#	Select rows that are NOT the Majority class
		#	Delete selected rows
		#       Assign the Majority percent to the PctFld
		#arcpy.AddMessage('...Calculating Majority Percent')
		arcpy.MakeTableView_management(cmbTable, "nMajView%s" % Yr, "MAJORITY <> %s" % CDLFld ) 
		arcpy.DeleteRows_management("nMajView%s" % Yr)
		arcpy.Delete_management("nMajView%s" % Yr)
	
		
		# Join the combine table to the Crop History table and calc the Year's Majority and MajPct
		arcpy.JoinField_management(StatFrame, "FBndID", cmbTable, "FBndID", '\"' + MxPctFld + ';MAJORITY\"')
		arcpy.CalculateField_management(StatFrame, PctFld, "!" + MxPctFld + "!" , "PYTHON")
		arcpy.CalculateField_management(StatFrame, YrFld, "!MAJORITY!", "PYTHON")
	
		
	
		# Cleanup
		#----------------------------------------------------------------------------------------------
		arcpy.DeleteField_management(StatFrame, "MAJORITY")
		arcpy.DeleteField_management(StatFrame, "ROTVAL")
		arcpy.DeleteField_management(StatFrame, "MixedPct" + Yr)
		arcpy.DeleteField_management(StatFrame, "Shape_Leng")
		arcpy.Delete_management(ZonalSt_CDL)
		arcpy.Delete_management(StatRaster)
		arcpy.Delete_management(cmbRaster)
		arcpy.Delete_management(cmbTable)
        #arcpy.Delete_management(mixedOnlyTab)


def popSummaries(StatFrame, YrList6, CDL_lkup, LTthresh):
	
	# make a dictionary of the CDL lookup table
	##-------------------------------------------------------------------------
	dictCDL = {}
	for row in arcpy.da.SearchCursor(CDL_lkup,["Value", "ROTVAL"]):
		dictCDL[row[0]] = row[1]
	del row 
	
	#None Keys...
	dictCDL["None"] = 'X'
	dictCDL[""] = 'X'
	
	## Create tableview in memory
	##-------------------------------------------------------------------------
	arcpy.MakeTableView_management(StatFrame, "TView", "", "in_memory")


	## Initialize output fields
	##-------------------------------------------------------------------------
	crsr1 = arcpy.da.UpdateCursor("TView", ["GenLU","CropRotatn","CropSumry","CCCount","MixCount"])
	for row in crsr1:
		row[0] = ""
		row[1] = ""
		row[2] = ""
		row[3] = ""
		row[4] = ""
		crsr1.updateRow(row)
	del row
	del crsr1
	
	
	## Populate the Croprotation field
	##-------------------------------------------------------------------------
	SelectMaj = '"Maj' + YrList6[0] +'" IS NOT NULL and "Maj' + YrList6[1] +'" IS NOT NULL and ' \
	            '"Maj' + YrList6[2] +'" IS NOT NULL and "Maj' + YrList6[3] +'" IS NOT NULL and ' \
	            '"Maj' + YrList6[4] +'" IS NOT NULL and "Maj' + YrList6[5] +'" IS NOT NULL'
	
	rotationfields = ["CropRotatn","Maj"+YrList6[0],"Maj"+YrList6[1],"Maj"+YrList6[2], \
	                  "Maj"+YrList6[3],"Maj"+YrList6[4],"Maj"+YrList6[5]]
	                  
	crsr2 = arcpy.da.UpdateCursor("TView", rotationfields, SelectMaj)
	for row in crsr2:
		row[0] = dictCDL[row[1]] + dictCDL[row[2]] + dictCDL[row[3]] + dictCDL[row[4]] + dictCDL[row[5]] + dictCDL[row[6]]
		crsr2.updateRow(row)
       
	del row
	del crsr2	


	##-------------------------------------------------------------------------
	## Populate the ancillary crop fields:
	##		CropSumry
	##		CCCount
	## 		MixCount
	##-------------------------------------------------------------------------
	
	clist = list('CBDFGHILOPRTUWX')
                   
	theFields = ["CropRotatn", "CropSumry", "CCCount", "MixCount", \
	             "Pct"+YrList6[0],"Pct"+YrList6[1],"Pct"+YrList6[2], \
	             "Pct"+YrList6[3],"Pct"+YrList6[4],"Pct"+YrList6[5]]
	             
	SelectStr =  "\"CropRotatn\" <> ''"

	rows = arcpy.da.UpdateCursor("TView", theFields, SelectStr)
	
	print("Summarizing rotation...")
	
	for row in rows:
		cstrng = ""
	
		#rotstr = str(row.getValue("CropRotatn"))
		rotstr = str(row[0])
		
		# Calculate Crop Summary
		# -----------------------------------------------
		for chr in clist:
			cnt = string.count(rotstr, chr)
			if cnt > 0:
				cstrng = cstrng + chr + str(cnt)
		
		#row.CropSumry = cstrng
		row[1] = cstrng                                  # "CropSumry"
		
		# Calculate Corn-after-Corn count
		# -----------------------------------------------
		ccnt = 0
		for i in range(len(rotstr)-1):
			if 'CC' in rotstr[i:i+2]:
				ccnt += 1
		#row.CCCount = str(ccnt)+':'+str(len(rotstr))
		row[2] = str(ccnt)+':'+str(len(rotstr))          # "CCCount"
	
		# Calculate mixed field count
		# -----------------------------------------------
		thrshld = 75
		cmcnt = 0
		for YRpct in row[4:10]:
			if YRpct < thrshld:
				cmcnt += 1
		row[3] = str(cmcnt)+':'+str(len(YrList6))        # mixcount

		rows.updateRow(row)
	
	del row
	del rows                


	##-------------------------------------------------------------------------
	# Assign General Land Use
	#  Must have Crop Rotation 
	# -----------------------------------------------
	#print("Assigning General Land Use...")
	
	SelectStr =  "\"CropRotatn\" <> ''"
	
	fldList = ["CropRotatn","GenLU","CropSumry","CCCount", "MixCount","Acres","isAG", \
                   "Pct"+YrList6[0],"Pct"+YrList6[1],"Pct"+YrList6[2], \
                   "Pct"+YrList6[3],"Pct"+YrList6[4],"Pct"+YrList6[5]]
                                     
	rowz = arcpy.da.UpdateCursor("TView", fldList, SelectStr)     
	for roow in rowz:
		rotstr = str(roow[0])


		## Extract land use count from rotation string
		beancount = rotstr.count("B")
		corncount = rotstr.count("C")
		CBcount = rotstr.count("D")
		pasturecount = rotstr.count("P")
		othercount = rotstr.count("R")
		forestcount = rotstr.count("F")
		watercount = rotstr.count("T")
		urbancount = rotstr.count("U")
		cloudcount = rotstr.count("X")


        ## Calculate mean percent PCTyr
		tot6 = 0
		mean6 = 100
		for YRpct in roow[7:13]:
			tot6 = tot6 + int(YRpct)
		mean6 = tot6 / 6
			
		
		## Assign General Land Use class
		if roow[5] <= LTthresh:
			roow[1] = "LT %s ac" %(str(LTthresh))
		        
		elif urbancount >= 4:
			roow[1] = "Urban"
		elif watercount >= 4:
			roow[1] = "Water/wetland"
		elif forestcount >= 4:
			roow[1] = "Forest"
		elif pasturecount > 4:
			roow[1] = "Pasture"
			roow[6] = 2
		
		elif roow[3] == "5:6" and roow[4] == "0:6" :           #roow.CCCount and roow.MixCount
			roow[1] = "Continuous Corn"
			roow[6] = 1
		        
		elif roow[2] == "C3B3" and roow[3] == "0:6":           # roow.CropSumry and roow.CCCount
			roow[1] = "Corn/Soybeans"     
			roow[6] = 1
		                      
		elif roow[2] == "C3B3" and roow[3] != "0:6":
			roow[1] = "C/S with Continuous Corn"     
			roow[6] = 1
		
		elif corncount + beancount + CBcount == 6 and roow[3] == "0:6" :
			roow[1] = "Corn/Soybeans"      
			roow[6] = 1
		        
		elif corncount + beancount + CBcount == 6 and roow[3] != "0:6" :
			roow[1] = "C/S with Continuous Corn"    
			roow[6] = 1
		        
		elif corncount + beancount + CBcount + pasturecount == 6 and mean6 > 60:
			roow[1] = "Conservation Rotation"
			roow[6] = 1
		                 
		elif corncount + beancount + CBcount + othercount + pasturecount == 6 and mean6 > 60:
			roow[1] = "Extended Rotation"
			roow[6] = 1
		
		elif corncount + beancount + CBcount + othercount + pasturecount == 6:
			roow[1] = "Mixed Agriculture"
			roow[6] = 1
		
		elif forestcount + pasturecount == 6:
			roow[1] = "Pasture"
			roow[6] = 2
		               
		elif watercount > 0 and  watercount + corncount + beancount + CBcount + othercount + pasturecount == 6:
			roow[1] = "Flood-prone Cropland"
			roow[6] = 1
		
		else:
			roow[1] = "UnAssigned"
		
		rowz.updateRow(roow)
	
	del roow
	del rowz  
            
	arcpy.Delete_management("TView")
                      
	
	
def mkOutputs(StatFrame, FileGDB, inHUC, YrList, FBOut):
	arcpy.AddMessage("Exporting...")

	# rename existing features and tables
	# If the original is the only existing one, rename it to '_orig'
	# If '_orig' already exists, rename to old -- over and over again
	FB_orig = "FB%s_orig" % (inHUC)
	
	LU_table = "LU6_%s" % inHUC
	LU_orig =  "LU6_%s_orig"  % inHUC
	
	CH_table = "CH_%s" % inHUC
	CH_orig =  "CH_%s_orig" % inHUC
#	CH_old =   "CH_%s_old" % inHUC
	
	if arcpy.Exists(FBOut):
		if not arcpy.Exists(FB_orig):
			arcpy.CopyFeatures_management(FBOut, FB_orig)		
			
	if arcpy.Exists(LU_table):
		if not arcpy.Exists(LU_orig):
			arcpy.Rename_management(LU_table, LU_orig, "Table")

	if arcpy.Exists(CH_table):
		if not arcpy.Exists(CH_orig):
			arcpy.Rename_management(CH_table, CH_orig, "Table")
#		elif arcpy.Exists(CH_orig):
#			if arcpy.Exists(CH_old):	arcpy.Delete_management(CH_old)
#			arcpy.Rename_management(CH_table, CH_old, "Table")
		
	# Create the new
	arcpy.CopyFeatures_management(StatFrame, FBOut)		
 	arcpy.TableToTable_conversion(StatFrame, FileGDB, LU_table)
 	arcpy.TableToTable_conversion(StatFrame, FileGDB, CH_table)
 	
 	# Delete appropriate fields
 	FB_delfldList = ["GenLU","CropRotatn","CropSumry","CCCount","MixCount"]	
 	LU6_delfldList = ["Shape_Length", "Shape_Area", "Acres", "isAg"]
 	CH_delfldList = FB_delfldList + LU6_delfldList 

 	arcpy.DeleteField_management(FBOut, FB_delfldList)
 	arcpy.DeleteField_management(LU_table, LU6_delfldList)
 	arcpy.DeleteField_management(CH_table, CH_delfldList)
	
	# Delete the year fields; FBout, LU6
	YrfldList = []
	for Yr in YrList:
		YrFld = "maj" + Yr[-2:]
		PctFld = "pct" + Yr[-2:]
		YrfldList.append(YrFld)
		YrfldList.append(PctFld)

	arcpy.DeleteField_management(FBOut, YrfldList)
	arcpy.DeleteField_management(LU_table, YrfldList)
	
	#arcpy.Delete_management(StatFrame)
	del (YrfldList, FB_delfldList, CH_delfldList)




##------------------------------------------------------------------------------
##------------------------------------------------------------------------------

if __name__ == "__main__":

	FBedit = arcpy.GetParameterAsText(0)
	LTthresh = int(arcpy.GetParameterAsText(1))
	CDL_lkup = arcpy.GetParameterAsText(2)
	
	# LTthresh is the size of the field assign 'LT x Ac.'
	if LTthresh < 5:
		LTthresh = 5

	env.parallelProcessingFactor = "90%"
		
	# Input data	
	#-----------------------------------------------------------------------------
	fbDesc = arcpy.Describe(FBedit)
	FileGDB = os.path.dirname(fbDesc.catalogPath)

	env.workspace = FileGDB
	env.scratchWorkspace = env.scratchFolder

	inHUC = os.path.split(FileGDB)[1][-16:-4]

	FBOut = "FB%s" % inHUC
	if FBedit ==  "%s\\%s" %(FileGDB, FBOut):
		arcpy.AddMessage("The original field boundary feature class is not a valid entry...%s" % FBedit)
		sys.exit(0)
		
	StatFrame = FileGDB + "\\FieldFrame"
	TempFrame = FileGDB + "\\TempFrame"
	
	# Years
	#-----------------------------------------------------------------------------
	#YrList = string.split(inRasters, ';')
	arcpy.AddMessage(env.workspace)
	YrList = arcpy.ListRasters("wsCDL*")
	if len(YrList) < 6:
		arcpy.AddMessage("...at least 6 years of land use data (rasters) are required: %s" %(YrList))
		sys.exit(0)
		
	YrList.sort()
	YrList6 = []
	for year in YrList[-6:]:
		YrList6.append(year[-2:])

	if len(YrList) > 6:
		arcpy.AddMessage("...only the last 6 years will be used for land use assignment: %s" %(YrList6))

			
		
	mkOutputFrame(FBedit, StatFrame, TempFrame, inHUC)
		
	ProcByYear(StatFrame, FileGDB, inHUC, YrList)
		
	popSummaries(StatFrame, YrList6, CDL_lkup, LTthresh)
		
	mkOutputs(StatFrame, FileGDB, inHUC, YrList, FBOut)
				
			
	env.scratchWorkspace = ""

