# WASCOBS
#
# The WASCOB tool identifies potential locations for WASCOBS on watersheds from 2 to 50 acres in size.
# A user-provided embankment height defines the height of the WASCOB (as measured from the bottom center of the drainageway). 
#
# Points are generated every 200 feet along collective flow pathways of 2 to 50 acres in size in agricultural land use fields only (including pasture).
# Points are first removed if the elevation change between itself and the next upstream point is not enough to install a WASCOB at the user-specified
# height without flooding out the upstream location. At each remaining point, a 100 m wide transect line is drawn perpendicular to the mean direction
# of flow in that location. The elevation profile of the transect line is then analyzed to estimate the shape of the drainageway and determine the
# suitability of the location for WASCOB installation. At a minimum, the height of each side of the drainageway wall must be at least the height of
# the embankment to be qualified and tabulated in the attribute table. Transect lines that meet this criteria are appended to the output WASCOB feature
# class. Additionally, WASCOBS that fall within 90 meters of the main stream reach (representing perennial flow) will be omitted from analysis. 
#
# WASCOBS that intersect other WASCOBS are selected, and those with the higher contributing area area maintained
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 05/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import math
import sys, string, os, os.path, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")


def CreateDrainageNetwork(D8FlowAcc):

    # Convert accumulation raster from meters squared to HA - define min and max drainage for sampling
    D8Acc_Acre = ((Times(D8FlowAcc, (cellsize * cellsize))) / 4046)
    MinDrainage = 2
    MaxDrainage = 50

    arcpy.AddMessage("Creating drainage network to be sampled for WASCOB suitability: >= %s acres and <= %s acres" % (MinDrainage, MaxDrainage))
    flownet = Con(D8Acc_Acre, 1, "", "VALUE >= %s and VALUE <= %s" % (MinDrainage, MaxDrainage))

    return(D8Acc_Acre, flownet)


def ReachDelineation(D8FlowDir, DEM, flownet): 
    arcpy.env.compression = "NONE"
    arcpy.AddMessage("Generating reach file from drainage network")

    # Fill the DEM
    DEMFill = Fill(DEM)

    # Reclassify D8FlowDir to TauDEM values 
    TDFlowDir = Reclassify(D8FlowDir, "VALUE",
                           RemapRange([[1, 1], [128, 2], [64, 3], [32, 4], [16, 5], [8, 6], [4, 7], [2, 8]]))

    # Copy TDFlowDir, DEMFill, and flow_ras to TDProcDir as uncompressed tiffs
    arcpy.CopyRaster_management(TDFlowDir, TDProcDir + "\\TDflowdir.tif")
    arcpy.CopyRaster_management(DEMFill, TDProcDir + "\\TDfill.tif")
    arcpy.CopyRaster_management(flownet, TDProcDir + "\\TDflownet.tif")

    core_cnt = multiprocessing.cpu_count()
    core_no = core_cnt - 1

    # Create a TauDEM Flow Accumulation raster
    callstr = 'mpiexec -n %s AreaD8 ' % core_no + \
              ' -p ' + TDProcDir + "\\TDflowdir.tif" + \
              ' -ad8 ' + TDProcDir + "\\TDflow_acc.tif -nc"
    call(callstr, shell=True)
    arcpy.CalculateStatistics_management(TDProcDir + "\\TDflow_acc.tif")

    # Create TauDEM stream reach and catchments
    callstr = 'mpiexec -n %s StreamNet ' % core_no + \
              ' -fel ' + TDProcDir + "\\TDfill.tif" + \
              ' -p ' + TDProcDir + "\\TDflowdir.tif" + \
              ' -ad8 ' + TDProcDir + "\\TDflow_acc.tif" + \
              ' -src ' + TDProcDir + "\\TDflownet.tif" + \
              ' -ord ' + TDProcDir + "\\TDstreamorder.tif" + \
              ' -tree ' + TDProcDir + "\\TDnetworktree.txt" + \
              ' -coord ' + TDProcDir + "\\TDnetworkcoord.txt" + \
              ' -net ' + TDProcDir + "\\TDreach_file.shp" + \
              ' -w ' + TDProcDir + "\\TDcatchments.tif"
    call(callstr, shell=True)
    
    # Calculate statistics
    TD_reach = arcpy.CalculateStatistics_management(TDProcDir + "\\TDreach_file.shp")

    arcpy.Delete_management(DEMFill)
    arcpy.Delete_management(TDFlowDir)
    arcpy.Delete_management(flownet)
    del[DEMFill, TDFlowDir, flownet]

    arcpy.ClearEnvironment("compression")

    return(TD_reach)
    

def joinCheck(lyr):
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def FieldBoundaryMask(TD_reach, FBIn, WBD, RunoffRiskTable, SubsetField, SubsetValue):

    if RunoffRiskTable == '':
        # Create mask of just agricultural fields    
        lyr = arcpy.MakeFeatureLayer_management(FBIn, "lyr", "\"isAG\" >= 1")    
        AgFB = arcpy.CopyFeatures_management(lyr, "in_memory" + "\\AgFB")

        # Clip TD_reach by WBD
        bnd_clip = arcpy.Clip_analysis(TD_reach, WBD, "in_memory" + "\\bnd_clip")

        # Clip by AgFB
        reach_file = arcpy.Clip_analysis(bnd_clip, AgFB, "reach_file")
        arcpy.RepairGeometry_management(reach_file)
        
        # Trim lines shorter than the distance of 1/2 WASCOB
        arcpy.TrimLine_edit(reach_file, "50 meters", "DELETE_SHORT")

        arcpy.Delete_management(AgFB)
        arcpy.Delete_management(bnd_clip)
        del[AgFB, bnd_clip]

    elif RunoffRiskTable  <> '':
        # Create mask of agricultural fields that have a runoff risk specified by subset field and subset value
        if SubsetField == '':
            arcpy.AddError("Runoff Risk table was provided but no field was selected to subset results...QUITTING!!!")
            sys.exit(0)
            
        elif SubsetField <> '':
            arcpy.AddMessage("Runoff Risk table provided")
            
            # Add field delimiters to the SubsetField
            Field_type = arcpy.ListFields(RunoffRiskTable, "%s" % SubsetField)[0].type
            Field_del = arcpy.AddFieldDelimiters(arcpy.Describe(RunoffRiskTable).path, "%s" % SubsetField)

            # Values from the selected field will be chosen and added to a list
            if str(Field_type) == 'String':
                ValueList = []
                for value in SubsetValue.split(';'):
                    if value not in ValueList:
                        ValueList.append(value)

            # A query is created to select fields that contain values in the list
            ValueList = ["'%s'" % value for value in ValueList]
            whereClause = "\"isAG\" >= 1 and %s IN(%s)" % (Field_del, ', '.join(map(str, ValueList)))
            arcpy.AddMessage("Subsetting fields to those with %s values of: %s" % (SubsetField, ValueList))

            # Make a copy of FBIn
            FB_copy = arcpy.CopyFeatures_management(FBIn, "in_memory" + "\\FB_copy")
                        
            # Join SubsetField from runoff risk
            arcpy.JoinField_management(FB_copy, "FBndID", RunoffRiskTable, "FBndID", ["%s" % SubsetField])
            
            # Apply query slection to FB_copy
            lyr = arcpy.MakeFeatureLayer_management(FB_copy, "lyr", whereClause)
            AgFB = arcpy.CopyFeatures_management(lyr, "in_memory" + "\\AgFB")

            # Clip TD_reach by WBD
            bnd_clip = arcpy.Clip_analysis(TD_reach, WBD, "in_memory" + "\\bnd_clip")

            # Clip by AgFB
            reach_file = arcpy.Clip_analysis(bnd_clip, AgFB, "reach_file")
            arcpy.RepairGeometry_management(reach_file)
            
            # Trim lines shorter than the distance of 1/2 WASCOB
            arcpy.TrimLine_edit(reach_file, "50 meters", "DELETE_SHORT")

            arcpy.Delete_management(FB_copy)
            arcpy.Delete_management(AgFB)
            arcpy.Delete_management(bnd_clip)
            del[FB_copy, AgFB, bnd_clip]

    arcpy.Delete_management(TDProcDir)
    arcpy.Delete_management("in_memory")

    return(reach_file)        


def DirectionStats(reach_file):

    # Find the mean directional angle of each reach in the reach file - WASCOBs will be created perpendicular to this
    Dir_Stats = arcpy.DirectionalMean_stats(reach_file, "Dir_Stats", "", "LINKNO")

    return(Dir_Stats)


def CreatePointsAlongDrainageNetwork(reach_file):
    arcpy.AddMessage("Creating sample points along drainage network at 200 foot spacing")
    
    # CREATE POINTS ALONG THE STREAM NETWORK (distance defined by spacing distance)
    create_pts = arcpy.CreateFeatureclass_management("in_memory", "create_pts", "Point", "", "", "", reach_file)

    # Identify the geometry field
    desc = arcpy.Describe(reach_file)
    shapefieldname = desc.ShapeFieldName
    
    # Create search cursor
    rows = arcpy.SearchCursor(reach_file)
    row = rows.next()
    
    iCurPts = arcpy.InsertCursor(create_pts)
    # Enter while loop for each feature/row
    while row:
        # Create the geometry object
        feat = row.getValue(shapefieldname)
        partnum = 0
        partcount = feat.partCount

        # Enter while loop for each part in the feature (if a singlepart feature this will occur only once)
        while partnum < partcount:
            part = feat.getPart(partnum)
            pnt = part.next()
            pntcount = 0

            ## statnPtsSpacing is the distance between Station Points along the line
            statnPtsSpacing = 60.0
            ## statnSctnLeftovers is the unallocated distance between Station Points along the line when you move to a new part
            statnSctnLeftovers = 0.00
                
            newPt = arcpy.CreateObject("Point")
            # Enter while loop for each vertex          
            while pnt:
                if pntcount > 0:
                    ## Calculate the distance (hypotenuse) and angle between previous and current point
                    pnt_dx = pnt.X - prevPtX
                    pnt_dy = pnt.Y - prevPtY
                    hypotenuse = math.sqrt(pnt_dx**2 + pnt_dy**2)
                    angle = math.atan2(pnt_dy, pnt_dx)
                    perp = angle + math.pi
                    
                    statnPtsInSctn = (hypotenuse-(statnSctnLeftovers * statnPtsSpacing))/statnPtsSpacing + 1
                    statnPtsInSctnInt = int(statnPtsInSctn)

                    ## Calculate statnPtsSpacing distance between starting point and next point
                    if angle >= -math.pi/2 and angle < math.pi/2: ## -90 to 90 degrees
                        statnPtDX = statnPtsSpacing*math.cos(angle)
                        statnPtDY = statnPtsSpacing*math.sin(angle)
                    if angle >= 0 and angle < math.pi/2: ## 0 to 90 degrees
                        statnPtDX = statnPtsSpacing*math.cos(angle)
                        statnPtDY = statnPtsSpacing*math.sin(angle)
                    if angle < 0 and angle >= -math.pi/2: ## 0 to -90 degrees
                        statnPtDX = statnPtsSpacing*math.cos(angle)
                        statnPtDY = statnPtsSpacing*math.sin(angle)
                    if angle >= math.pi/2 and angle <= math.pi: ## 90 to 180 degrees
                        statnPtDX = -statnPtsSpacing*math.cos(math.pi-angle)
                        statnPtDY = statnPtsSpacing*math.sin(math.pi-angle)
                    if angle < -math.pi/2 and angle >= -math.pi: ## -90 to -180 degrees
                        statnPtDX = -statnPtsSpacing*math.cos(math.pi-angle)
                        statnPtDY = statnPtsSpacing*math.sin(math.pi-angle)
                        
                    ## Calculate for all sections
                    for i in range(0, statnPtsInSctnInt):
                    ## Set point X & Y
                        newPt.X = prevPtX + (statnSctnLeftovers + i)*statnPtDX
                        newPt.Y = prevPtY + (statnSctnLeftovers + i)*statnPtDY
                        iRowPts = iCurPts.newRow()
                        iRowPts.Shape = newPt
                        iCurPts.insertRow(iRowPts)

                    ## Calculate how much distance was left between last station point in line part and next point in line feature
                    statnSctnRemainingFraction = (hypotenuse-(statnSctnLeftovers * statnPtsSpacing))%statnPtsSpacing/statnPtsSpacing
                    ## Print "Section Remaining Fraction = " + str(statnSctnRemainingFraction)
                    ## Calculate how much distance beetween station points was left over
                    statnSctnLeftovers = 1-statnSctnRemainingFraction 
                    ## Print "Section Leftovers = " + str(statnSctnLeftovers)

                prevPtX = pnt.X
                prevPtY = pnt.Y
                pnt = part.next()
                pntcount += 1
                
            partnum += 1
            
        row = rows.next()

    return(create_pts)


def CreateTransects(D8Acc_Acre, create_pts, D8FlowAcc, DEM, reach_file, Dir_Stats, StreamReach, FBIn, ZFactor):

    # Define WASCOB Impoundment Height in same elevation units as DEM
    ImpoundmentHeight = float(WASCOBImpoundmentHeight) / float(ZFactor)
    TooIncised = float(Times(2, ImpoundmentHeight))

    # Add and calculate a WASCOBID to points - same as OID
    desc = arcpy.Describe(create_pts)
    OID = desc.OIDFieldName 
    arcpy.AddField_management(create_pts, "WASCOBID", "LONG")
    arcpy.CalculateField_management(create_pts, "WASCOBID", "!%s!" % OID, "PYTHON")

    # Convert reach_file to raster by LINKNO field
    reach_ras = arcpy.FeatureToRaster_conversion(reach_file, "LINKNO", "reach_ras", cellsize)

    # Atribute each point
    arcpy.AddMessage("Attributing points with upstream drainage area, elevation, and reach ID")

    # Atribute each point with the flow accumulation (in acres) at that point
    AccStats = ZonalStatisticsAsTable(create_pts, "WASCOBID", D8Acc_Acre, "in_memory" + "\\AccStats", "", "MEAN")
    arcpy.JoinField_management(create_pts, "WASCOBID", AccStats, "WASCOBID", ["MEAN"])
    arcpy.AddField_management(create_pts, "ContAreaAC", "DOUBLE")
    arcpy.CalculateField_management(create_pts, "ContAreaAC", "!MEAN!", "PYTHON")
    arcpy.DeleteField_management(create_pts, ["MEAN"])

    # Atribute each point with the elevation at each point
    ElevStats = ZonalStatisticsAsTable(create_pts, "WASCOBID", DEM, "in_memory" + "\\ElevStats", "", "MEAN")
    arcpy.JoinField_management(create_pts, "WASCOBID", ElevStats, "WASCOBID", ["MEAN"])
    arcpy.AddField_management(create_pts, "Elevation", "LONG")
    arcpy.CalculateField_management(create_pts, "Elevation", "!MEAN!", "PYTHON")
    arcpy.DeleteField_management(create_pts, ["MEAN"])

    # Atribute each point with the reach id ("LINKNO") that if falls upon
    ReachStats = ZonalStatisticsAsTable(create_pts, "WASCOBID", reach_ras, "in_memory" + "\\ReachStats", "", "MEAN")
    arcpy.JoinField_management(create_pts, "WASCOBID", ReachStats, "WASCOBID", ["MEAN"])
    arcpy.AddField_management(create_pts, "Reach", "LONG")
    arcpy.CalculateField_management(create_pts, "Reach", "!MEAN!", "PYTHON")
    arcpy.DeleteField_management(create_pts, ["MEAN"])

    # Delete points that have a contributing area < MinDrainage (in acres) or > MaxDrainage (in acres) or NULL or do not fall on a reach
    arcpy.MakeFeatureLayer_management(create_pts, "erase_pts")
    arcpy.SelectLayerByAttribute_management("erase_pts", "NEW_SELECTION", "\"Reach\" IS NULL OR \"ContAreaAC\" < 2 OR \"ContAreaAC\" > 50 OR \"ContAreaAC\" IS NULL")
    arcpy.DeleteFeatures_management("erase_pts")

    # Sort by contributing area --> upstream points are first
    sort_pts = arcpy.Sort_management(create_pts, "in_memory" + "\\sort_pts", [["ContAreaAC", "ASCENDING"]])

    # Create ReachList ----> append each unique "LINKNO" to ReachList as a string
    rows = arcpy.SearchCursor(sort_pts)
    ReachList = []
    for row in rows:
        ReachID = str(row.getValue("Reach"))
        if ReachID not in ReachList:
            ReachList.append(ReachID)

    arcpy.AddMessage("Deleting points with an elevation drop between itself and the next upstream point less than the WASCOB impoundment height")
    for Reach in ReachList:
        # Delete points where, if impoundment was installed, it would flood the next upstream point
        UpElev = 100000000
        with arcpy.da.UpdateCursor(sort_pts, ["Elevation"], "\"Reach\" = %s" % Reach) as cursor:
            for row in cursor:
                CurrentElev = (row[0] + ImpoundmentHeight)
                if CurrentElev < UpElev:
                    UpElev = row[0]
                else:
                    cursor.deleteRow()

    # Add and populate "EmbankHgt" - represents the height of the WASCOB in the same elevation units as the DEM
    arcpy.AddField_management(sort_pts, "EmbankHgt", "FLOAT")
    arcpy.CalculateField_management(sort_pts, "EmbankHgt", "%s" % ImpoundmentHeight, "PYTHON")

    arcpy.AddMessage("creating WASCOBS")
    # Join "CompassA" field from Dir_Stats - angle that is the mean direction of flow along each reach
    arcpy.JoinField_management(sort_pts, "Reach", Dir_Stats, "LINKNO", ["CompassA"])

    # Add and populate a "PerpAng" field - angle that is perpendicular to the mean direction of flow along each reach
    arcpy.AddField_management(sort_pts, "PerpAng", "FLOAT")
    arcpy.CalculateField_management(sort_pts, "PerpAng", "!CompassA! - 90", "PYTHON")

    # Add xy coordinates to each point
    arcpy.AddXY_management(sort_pts)

    # Add 2 distance fields to sort_pts; represents a left and a right offset (+50 meters and -50 meters)
    arcpy.AddField_management(sort_pts, "loffset", "SHORT")
    arcpy.CalculateField_management(sort_pts, "loffset", 50, "PYTHON")
    arcpy.AddField_management(sort_pts, "roffset", "SHORT")
    arcpy.CalculateField_management(sort_pts, "roffset", -50, "PYTHON")

    # Use bearing distance to line tool to generate perpendicular lines (do twice - once for loffset and roffset)
    loffset = arcpy.BearingDistanceToLine_management(sort_pts, "in_memory" + "\\loffset", "POINT_X", "POINT_Y", "loffset", "METERS", "PerpAng", "DEGREES", "", "WASCOBID", sort_pts)
    roffset = arcpy.BearingDistanceToLine_management(sort_pts, "in_memory" + "\\roffset", "POINT_X", "POINT_Y", "roffset", "METERS", "PerpAng", "DEGREES", "", "WASCOBID", sort_pts)

    # Convert left bank lines and right bank lines to rasters
    lbank_ras = arcpy.FeatureToRaster_conversion(loffset, "WASCOBID", "in_memory" + "\\loffset_ras", cellsize)
    rbank_ras = arcpy.FeatureToRaster_conversion(roffset, "WASCOBID", "in_memory" + "\\roffset_ras", cellsize)

    # Zonal Stats of left bank and right bank rasters - to find the elevation range along each bank
    left_stats = ZonalStatisticsAsTable(lbank_ras, "Value", DEM, "in_memory" + "\\left_stats", "", "RANGE")
    right_stats = ZonalStatisticsAsTable(rbank_ras, "Value", DEM, "in_memory" + "\\right_stats", "", "RANGE")

    # Merge the left and right bank lines - represents 100 meter long WASCOB line ---> dissolve 
    off_merge = arcpy.Merge_management([loffset, roffset], "in_memory" + "\\off_merge")
    WASCOB_diss = arcpy.Dissolve_management(off_merge, "in_memory" + "\\WASCOB_diss", ["WASCOBID"])

    # Join attributes from sort_pts based on common WASCOBID
    arcpy.JoinField_management(WASCOB_diss, "WASCOBID", sort_pts, "WASCOBID", ["ContAreaAC", "Elevation", "EmbankHgt"])

    # Join the range field from left_stats and right_stats - represents elevation range of each bank
    arcpy.JoinField_management(WASCOB_diss, "WASCOBID", left_stats, "VALUE", ["RANGE"])
    arcpy.AddField_management(WASCOB_diss, "lbank_hgt", "FLOAT")
    arcpy.CalculateField_management(WASCOB_diss, "lbank_hgt", "!RANGE!", "PYTHON")
    arcpy.DeleteField_management(WASCOB_diss, ["RANGE"])
    
    arcpy.JoinField_management(WASCOB_diss, "WASCOBID", right_stats, "VALUE", ["RANGE"])
    arcpy.AddField_management(WASCOB_diss, "rbank_hgt", "FLOAT")
    arcpy.CalculateField_management(WASCOB_diss, "rbank_hgt", "!RANGE!", "PYTHON")
    arcpy.DeleteField_management(WASCOB_diss, ["RANGE"])

    # Delete WASCOBS where either bank wall is less than the height of the impoundment (i.e. the drainage is too shallow)
    arcpy.AddMessage("Deleting WASCOBS if either drainageway wall (right or left bank) is lower than the height of the WASCOB impoundment of %s vertical map units" % ImpoundmentHeight)
    erase = arcpy.MakeFeatureLayer_management(WASCOB_diss, "erase")
    arcpy.SelectLayerByAttribute_management(erase, "NEW_SELECTION", "\"lbank_hgt\" < %s OR \"rbank_hgt\" < %s OR \"lbank_hgt\" > %s OR \"rbank_hgt\" > %s" % (ImpoundmentHeight, ImpoundmentHeight, TooIncised, TooIncised))
    arcpy.DeleteFeatures_management(erase)

    arcpy.AddMessage("Cleanup...removing WASCOBS that overlap or intersect")
    
    # Find WASCOBS that intersect other WASCOBS - delete those with smaller drainange areas
    WASCOB_lyr = arcpy.MakeFeatureLayer_management(WASCOB_diss, "WASCOB_lyr")

    # Loop through each row in WASCOBS layer
    with arcpy.da.SearchCursor(WASCOB_diss, ["WASCOBID", "ContAreaAC"]) as cursor:
        for row in cursor:
            WASCOBID = row[0]
            
            # Make feature layer of each WASCOB - current WASCOB
            curr_WASCOB = arcpy.MakeFeatureLayer_management(WASCOB_diss, "curr_WASCOB", "\"WASCOBID\" = %s" % WASCOBID)
            
            # Select WASCOBS that intersect with current WASCOB  
            arcpy.SelectLayerByLocation_management(WASCOB_lyr, "INTERSECT", curr_WASCOB, "", "NEW_SELECTION")

            # Perfom a count to see how many intersections exist
            count = int(arcpy.GetCount_management(WASCOB_lyr).getOutput(0))

            # If count > 1, sort WASCOBS by contributing area
            while count > 1:
                for x in sorted(arcpy.da.SearchCursor(WASCOB_lyr, ["WASCOBID"])):
                    eraseid = x[0]
                    erase = arcpy.MakeFeatureLayer_management(WASCOB_lyr, "erase", "\"WASCOBID\" = %s" % eraseid)
                    arcpy.DeleteRows_management(erase)

                    # Perform count again
                    count = int(arcpy.GetCount_management(WASCOB_lyr).getOutput(0))

    # Convert field boundaries to a line
    FB_line = arcpy.FeatureToLine_management(FBIn, "in_memory" + "\\FB_line")

    # Delete WASCOBS that are within 90 meters of the stream reach
    arcpy.SelectLayerByLocation_management(WASCOB_lyr, "WITHIN_A_DISTANCE", StreamReach, "90 METERS", "NEW_SELECTION")

    # Add those WASCOBS that intersect a field boundary
    arcpy.SelectLayerByLocation_management(WASCOB_lyr, "INTERSECT", FB_line, "", "ADD_TO_SELECTION")

    # Delete
    arcpy.DeleteFeatures_management(WASCOB_lyr)

    # Near analysis to identify WASCOBS within 10 meters of each other
    arcpy.Near_analysis(WASCOB_diss, WASCOB_diss, "10 meters")

    # Select those within a distance less than 10 meters
    WASCOB_near = arcpy.MakeFeatureLayer_management(WASCOB_diss, "WASCOB_near", "\"NEAR_DIST\" <> -1")

    # Delete identical
    arcpy.DeleteIdentical_management(WASCOB_near, "NEAR_DIST")

    # Delete near analysis fields
    arcpy.DeleteField_management(WASCOB_diss, ["NEAR_FID", "NEAR_DIST"])

    # Copy to output
    arcpy.CopyFeatures_management(WASCOB_diss, OutWASCOBS)

    arcpy.Delete_management(reach_ras)
    arcpy.Delete_management(AccStats)
    arcpy.Delete_management(ElevStats)
    arcpy.Delete_management(ReachStats)
    arcpy.Delete_management(reach_file)
    arcpy.Delete_management(create_pts)
    arcpy.Delete_management(sort_pts)
    arcpy.Delete_management(loffset)
    arcpy.Delete_management(roffset)
    arcpy.Delete_management(lbank_ras)
    arcpy.Delete_management(rbank_ras)
    arcpy.Delete_management(left_stats)
    arcpy.Delete_management(right_stats)
    arcpy.Delete_management(off_merge)
    arcpy.Delete_management(Dir_Stats)
    arcpy.Delete_management(WASCOB_diss)
    arcpy.Delete_management(FB_line)
    del[reach_ras, AccStats, ElevStats, ReachStats, reach_file, create_pts, sort_pts, loffset, roffset, lbank_ras, rbank_ras, left_stats, right_stats, off_merge, Dir_Stats, WASCOB_diss, FB_line]
    arcpy.Delete_management("in_memory")


if __name__ == "__main__":

    # Define Parameters
    FBIn = arcpy.GetParameterAsText(0)
    # Test for a join on FBIn
    basename = arcpy.Describe(FBIn).baseName
    hasJoin = joinCheck(FBIn)
    if hasJoin:
        arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    DEM = arcpy.GetParameterAsText(1)
    D8FlowAcc = arcpy.GetParameterAsText(2)
    D8FlowDir = arcpy.GetParameterAsText(3)
    StreamReach = arcpy.GetParameterAsText(4)
    WBD = arcpy.GetParameterAsText(5)
    RunoffRiskTable = arcpy.GetParameterAsText(6)
    # Test for a join on RunoffRisk table
    if RunoffRiskTable <> "":
        # Test for a join on RunoffRisk table
        basename = arcpy.Describe(RunoffRiskTable).baseName
        hasJoin = joinCheck(RunoffRiskTable)
        if hasJoin:
            arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
            sys.exit(0)    
    SubsetField = arcpy.GetParameterAsText(7)
    SubsetValue = arcpy.GetParameterAsText(8)
    WASCOBImpoundmentHeight = arcpy.GetParameterAsText(9)
    ZFactor = arcpy.GetParameterAsText(10)
    OutWASCOBS = arcpy.GetParameterAsText(11)
    	
    # Set extent, snap, workspace
    arcpy.env.extent = DEM
    arcpy.env.snapRaster = DEM
    arcpy.env.outputCoordinateSystem = DEM
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder

    # Set cellsize
    cellsize = float(arcpy.GetRasterProperties_management(D8FlowAcc, "CELLSIZEX").getOutput(0))
    
    # Create and define folder for TauDEM outputs (in ProcDir) - will be deleted at end of script
    scriptpath = sys.path[0]
    arcpy.CreateFolder_management(scriptpath, "TDProcDir")
    TDProcDir = scriptpath + "\\TDProcDir"

    # Run Modules
    D8Acc_Acre, flownet = CreateDrainageNetwork(D8FlowAcc)
    
    TD_reach = ReachDelineation(D8FlowDir, DEM, flownet)

    reach_file = FieldBoundaryMask(TD_reach, FBIn, WBD, RunoffRiskTable, SubsetField, SubsetValue)

    Dir_Stats = DirectionStats(reach_file)

    create_pts = CreatePointsAlongDrainageNetwork(reach_file)

    CreateTransects(D8Acc_Acre, create_pts, D8FlowAcc, DEM, reach_file, Dir_Stats, StreamReach, FBIn, ZFactor)

