# The WASCOB Basin tool delineates the area in which water would pond upstream of each WASCOB. 
#
# The elevation of each WASCOB (defined by the Embankment Height + Elevation at each site ----> derived from
# EmbankHgt and Elevation fields in the attribute table of the input WASCOB feature class) is "burned" into
# the input filled DEM. By using a filled DEM, no depressions should initially be present in the DEM. Once WASCOBS
# have been burned into the filled DEM, the filled DEM is then put through a "fill" process to identify the depressions
# created as a result of adding the WASCOBs.
#
# The output is a WASCOB basin depth raster (WASCOBbasin + inHUC), which represents the ponded area behind each WASCOB. 
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 01/2015
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time
from subprocess import call
import multiprocessing
arcpy.CheckOutExtension("Spatial")


def joinCheck(lyr):
    
    fields = arcpy.Describe(lyr).fields
    basename = arcpy.Describe(lyr).baseName
    for field in fields:
        if field.name.find(basename) == 0:
            return True
    return False


def WASCOBWatersheds(WASCOBS, DEMFill, D8FlowDir, ZFactor):

    arcpy.AddMessage("Burning WASCOBS into filled DEM...")
    # Convert WASCOBS to a raster based on WASCOBID
    WASCOB_ras = arcpy.FeatureToRaster_conversion(WASCOBS, "WASCOBID", "in_memory" + "\\WASCOB_ras", cellsize)

    # Find watersheds of all WASCOBS
    WASCOB_wshed = Watershed(D8FlowDir, WASCOB_ras, "VALUE")

    # Convert WASCOBS to a raster based on "EmbankHgt"
    WASCOB_elev = arcpy.FeatureToRaster_conversion(WASCOBS, "EmbankHgt", "in_memory" + "\\WASCOB_elev", cellsize)

    # Find the minimum elevation along each WASCOB
    WASCOB_min = ZonalStatistics(WASCOB_ras, "VALUE", DEMFill, "MINIMUM")

    # Add WASCOB_elev to WASCOB_min raster
    WASCOB_hgt = Plus(WASCOB_elev, WASCOB_min)

    # Burn WASCOB_elev into filled DEM raster
    NewDEM = Con(IsNull(WASCOB_elev), DEMFill, WASCOB_hgt)
    arcpy.AddMessage("New filled DEM created with WASCOBs burned in...")

    # Fill the new DEM
    arcpy.AddMessage("Filling sinks")
    NewDEMFill = Fill(NewDEM)

    # Subtract unfilled DEM from filled DEM - each cell represents the sink depth at that location
    arcpy.AddMessage("Delineating basins behind WASCOBS")
    FillReg = Minus(NewDEMFill, DEMFill)

    # Convert FillReg (which represents the volume of storage at every sink cell in the DEM) to volume in acre feet
    FillRegMet = Times(FillReg, float(ZFactor))
    FillRegCM = Times(FillRegMet, (cellsize * cellsize))
    FillRegAF = Times(FillRegCM, .000810713)

    # Convert values of all sink cells to the WASCOBID of the watershed that it falls in
    Sinks_ID = Con(FillReg, WASCOB_wshed, "", "VALUE > 0")

    # Convert to polygon based on gridcode (equal to WASCOBID) - add and populate "WASCOBID" field
    basins_poly = arcpy.RasterToPolygon_conversion(Sinks_ID, "basins_poly")
    arcpy.AddField_management(basins_poly, "WASCOBID", "LONG")
    arcpy.CalculateField_management(basins_poly, "WASCOBID", "!gridcode!", "PYTHON")
    arcpy.DeleteField_management(basins_poly, ["gridcode", "id"])

    # Dissolve
    basins_diss = arcpy.Dissolve_management(basins_poly, OutBasins, ["WASCOBID"])

    # Overlay basins_diss with FillRegAF to identify the volume (in acre feet) of potential storage behind each WASCOB
    storage_stats = ZonalStatisticsAsTable(basins_diss, "WASCOBID", FillRegAF, "in_memory" + "\\storage_stats", "#", "SUM")

    arcpy.JoinField_management(basins_diss, "WASCOBID", storage_stats, "WASCOBID", ["SUM"])
    arcpy.AddField_management(basins_diss, "StorageAF", "FLOAT")
    arcpy.CalculateField_management(basins_diss, "StorageAF", "!SUM!", "PYTHON")
    arcpy.DeleteField_management(basins_diss, ["SUM"])

    # Add and populate an area (in hectares) field                             
    arcpy.AddField_management(basins_diss, "BasAreaHA", "FLOAT")
    arcpy.CalculateField_management(basins_diss, "BasAreaHA", "!shape.area@hectares!", "PYTHON")

    if DepthRaster == 'true':
        arcpy.AddMessage("Output Basin Depth Raster selected")
        if OutBasinsRas == '':
            arcpy.AddError("Output Basin Depth Raster is selected but no output file location is provided...QUITTING!!!")
            sys.exit(0)
        elif OutBasinsRas <> '':
            arcpy.AddMessage("Generating output basin depth raster")
    
            # Convert values of all sink cells to the fillreg raster 
            Sinks_elev = Con(FillReg, FillReg, "", "VALUE > 0")

            # Copy to output file location
            arcpy.CopyRaster_management(Sinks_elev, OutBasinsRas)

            # Build pyramids
            arcpy.BuildPyramids_management(OutBasinsRas)

            arcpy.Delete_management(Sinks_elev)
            del[Sinks_elev]

    arcpy.Delete_management(WASCOB_ras)
    arcpy.Delete_management(WASCOB_wshed)
    arcpy.Delete_management(WASCOB_elev)
    arcpy.Delete_management(WASCOB_min)
    arcpy.Delete_management(WASCOB_hgt)
    arcpy.Delete_management(NewDEM)
    arcpy.Delete_management(NewDEMFill)
    arcpy.Delete_management(FillReg)
    arcpy.Delete_management(FillRegMet)
    arcpy.Delete_management(FillRegCM)
    arcpy.Delete_management(FillRegAF)
    arcpy.Delete_management(Sinks_ID)
    arcpy.Delete_management(basins_poly)
    arcpy.Delete_management(storage_stats)
    del[WASCOB_ras, WASCOB_wshed, WASCOB_elev, WASCOB_min, WASCOB_hgt, NewDEM, NewDEMFill, FillReg, FillRegMet, FillRegCM, FillRegAF, Sinks_ID, basins_poly, storage_stats]
    arcpy.Delete_management("in_memory")


if __name__ == "__main__":

    # Define Parameters
    WASCOBS = arcpy.GetParameterAsText(0)
    hasJoin = joinCheck(WASCOBS)
    if hasJoin:
        basename = arcpy.Describe(WASCOBS).baseName
        arcpy.AddError("Found a join on %s  that must be removed before running tool...QUITTING!!!!" % basename)
        sys.exit(0)
    DEMFill = arcpy.GetParameterAsText(1)
    D8FlowDir = arcpy.GetParameterAsText(2)
    OutBasins = arcpy.GetParameterAsText(3)
    ZFactor = arcpy.GetParameterAsText(4)
    DepthRaster = arcpy.GetParameterAsText(5)
    OutBasinsRas = arcpy.GetParameterAsText(6)
    	
    # Set extent, snap, workspace
    arcpy.env.snapRaster = DEMFill
    arcpy.env.extent = DEMFill
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder

    # Set Cellsize
    cellsize = float(arcpy.GetRasterProperties_management(DEMFill, "CELLSIZEX").getOutput(0))
    
    # Run Modules
    WASCOBWatersheds(WASCOBS, DEMFill, D8FlowDir, ZFactor)
