# The water table depth tool uses an input unfilled DEM, D8 Flow Direction grid, and Stream Reach feature class (polyline)
#
# Elevation differences are found between each grid cell in the DEM and the stream-channel grid cell that will receive overland flow from that cell.
# These elevation differences are reclassified into depth categories and used as a proxy for water table depth.
#
# Process: The Stream Reach polyline is converted to a raster and expanded by one cell to identify "bank" grid cells, which are assigned false,
# sky-high elevation values. The input DEM is then "filled", and filled elevation values are extracted to the channel grid cells only.
# Flow direction is calculated using just the bank cells and the channel cells. Sky-high elevation values for bank grid cells forces flow from the
# bank cells directly into the channel, while "filled" channel elevation values prevent in-channel sinks and ensures that flow is forced downstream
# once it enters the channel. The new flow direction values are burned into the original flow direction grid to create an "Adjusted Flow Direction"
# grid. 
#
##--------------------------------------------------------------------------------
## Original coding: S. Porter - NLAE 09/2014
##--------------------------------------------------------------------------------
#
##--------------------------------------------------------------------------------
# Import System Modules
##--------------------------------------------------------------------------------
import arcpy
from arcpy import env
from arcpy.sa import *
import sys, string, os, os.path, time
arcpy.CheckOutExtension("Spatial")


def RelativeElevation(StreamReach, D8FlowDir, DEM):
    arcpy.AddMessage("Running Relative Elevation Module")
    
    # Fill the DEM
    DEMFill = Fill(DEM)
    
    # Dissolve the StreamReach
    rch_diss = arcpy.Dissolve_management(StreamReach, "rch_diss", "", "", "", "")
 
    # Convert to raster
    desc = arcpy.Describe(rch_diss)
    OID = desc.OIDFieldName
    reach_ras = arcpy.PolylineToRaster_conversion(rch_diss, "%s" % OID, "reach_ras", "", "", cellsize)

    # Convert stream values to 1 and background values to NODATA
    stream_reach_ras = Con(reach_ras, 1, "", "VALUE >= 0")
 
    # Expand stream_reach_ras by 1 pixel 
    stream_reach_exp = Expand(stream_reach_ras, "1", "1")
    
    # Extract elevation from filled DEM (DEMFill) to stream_reach_ras
    stream_reach_elev = Con(stream_reach_ras, DEMFill, "", "VALUE = 1")
    
    # Create pseudo_elev ---> banks receive a value of 100000, channel cells remain value of stream_reach_elev, everything else is NODATA
    pseudo_elev = Con(IsNull(stream_reach_ras), stream_reach_exp * 100000, stream_reach_elev)

    # Find focal minimum of pseudo_elev using 3*3 window; channel cells and bank cells will receive the same value that will = the filled channel elevation values
    # One cell beyond the bank cells will have a value of 100000
    bank_focal_stats = FocalStatistics(pseudo_elev, NbrRectangle(3, 3, "CELL"), "MINIMUM")

    # Recalculate flow direction using pseudo elevation raster - forces flow direction from bank cell into channel, then down the channel
    bank_dir = FlowDirection(pseudo_elev, "NORMAL")

    # Exctract attributes from bank_dir that correspond to the arcgis directional values
    bank_flow_dir = ExtractByAttributes(bank_dir, "\"VALUE\" IN (1, 2, 4, 8, 16, 32, 64, 128)")

    # Create a bank id - banks receive a value of 1, channel cells receive a value of NODATA, everything else is 0
    bank_ID = Con(IsNull(stream_reach_ras), Con(IsNull(stream_reach_exp), 0, 1), "")

    # Substitute bank_focal_stats for bank_id; banks cells will receive value of filled channel elevation values, channel cells will be NODATA - everything else is 0
    bank_elev = Con(bank_ID, bank_focal_stats, 0, "VALUE = 1")

    # Substitute bank_flow_dir back into original flow direction raster
    AdjFlowDir = Con(IsNull(bank_flow_dir), D8FlowDir, bank_flow_dir)

    # Copy to output file location
    arcpy.CopyRaster_management(AdjFlowDir, OutAdjFlowDir)
    arcpy.AddMessage("Adjusted Flow Direction raster created")

    # Build pyramids
    arcpy.BuildPyramids_management(OutAdjFlowDir)

    # Find the length to the outlet of the watershed (which is the channel) using bank_elev as a weight raster
    # Every cell in output raster will receive the elevation value of the channel cell into which it will flow
    # EXCEPTION: If a cell flows through 2 bank cells before entering the channel (diagonal flow), it will receive the sum of their elevations - this is why the following step is required!!
    flow_length_1 = FlowLength(AdjFlowDir, "DOWNSTREAM", bank_elev)

    # Find the lengh to the outlet of watershed using the bank_id as the weight raster 
    # the result is that every cell in the raster will recieve a 1 if it flows through just 1 bank cell prior to entering channel (sum of 2 cells if 2, diagonals, etc.)
    flow_length_2 = FlowLength(AdjFlowDir, "DOWNSTREAM", bank_ID)

    # Divide flow_length_2 by flow_length_1
    # result is that every cell will receive the elevation value of the channel cell into which it will flow
    intersect_elev = Con(bank_elev, flow_length_1 / flow_length_2, bank_elev, "VALUE = 0")

    # Subtract the filled channel elevation values from the actual elevation of each grid cell
    RelativeElev = Minus(DEM, intersect_elev)

    # Save to output file location
    RelativeElev.save("RelativeElev")
    if OutRelElev <> '':
        arcpy.CopyRaster_management(RelativeElev, OutRelElev)

        # Build pyramids
        arcpy.BuildPyramids_management(OutRelElev)

    # Cleanup
    arcpy.Delete_management(rch_diss)
    arcpy.Delete_management(reach_ras)
    arcpy.Delete_management(stream_reach_ras)
    arcpy.Delete_management(stream_reach_exp)
    arcpy.Delete_management(stream_reach_elev)
    arcpy.Delete_management(pseudo_elev)
    arcpy.Delete_management(bank_focal_stats)
    arcpy.Delete_management(bank_dir)
    arcpy.Delete_management(bank_ID)
    arcpy.Delete_management(bank_elev)
    arcpy.Delete_management(flow_length_1)
    arcpy.Delete_management(flow_length_2)
    arcpy.Delete_management(intersect_elev)
    arcpy.Delete_management(AdjFlowDir)
    arcpy.Delete_management(DEMFill)
    del[rch_diss, reach_ras, stream_reach_ras, stream_reach_exp, stream_reach_elev, pseudo_elev, bank_focal_stats, bank_dir, \
        bank_flow_dir, bank_ID, bank_elev, flow_length_1, flow_length_2, intersect_elev, AdjFlowDir, DEMFill]

    return(RelativeElev)

    
def WaterTableDepth(RelativeElev, ZFactor):
    arcpy.AddMessage("Creating Water Table Depth Classes")

    # Multiply RelativeElev by 1000 - need integer format for reclassify tool below
    RelElev_1000 = Int(Times(RelativeElev, 1000))
    
    # Determine max and min relative elevation values
    MinWTD = arcpy.GetRasterProperties_management(RelElev_1000, "Minimum")
    MaxWTD = arcpy.GetRasterProperties_management(RelElev_1000, "Maximum")

    # Reclassify relative elevation into Water Table Depth classes - each remap range has been multiplied by 1000 to work with RelElev_1000
    WaterTableDepth = Reclassify(RelElev_1000, "Value",
                                 RemapRange([[MinWTD, 0.00, "1"],
                                             [0, ((150 / (float(ZFactor) * 100)) * 1000), "2"],
                                             [((150 / (float(ZFactor) * 100)) * 1000), ((300 / (float(ZFactor) * 100)) * 1000), "3"],
                                             [((300 / (float(ZFactor) * 100)) * 1000), ((500 / (float(ZFactor) * 100)) * 1000), "4"],
                                             [((500 / (float(ZFactor) * 100)) * 1000), MaxWTD, "5"]]))
    
    # Save to output file location
    WTD = arcpy.CopyRaster_management(WaterTableDepth, OutWTD)

    # Add WaterTableDepth field and populate with labels
    arcpy.AddField_management(WTD, "WaterTableDepth", "STRING")
    rows = arcpy.UpdateCursor(WTD)
    for row in rows:
        if row.Value == 1:
            row.WaterTableDepth = "Channel"
        else:
            if row.Value == 2:
                row.WaterTableDepth = "lt_1_5m"
            else:
                if row.Value == 3:
                    row.WaterTableDepth = "1_5to3m"
                else:
                    if row.Value == 4:
                        row.WaterTableDepth = "3to5m"
                    else:
                        if row.Value == 5:
                            row.WaterTableDepth = "gt_5m"                          
        rows.updateRow(row)

    # Build pyramids
    arcpy.BuildPyramids_management(OutWTD)

    # Cleanup
    arcpy.Delete_management(WaterTableDepth)
    arcpy.Delete_management(RelativeElev)
    arcpy.Delete_management(RelElev_1000)
    del[WaterTableDepth, RelativeElev, RelElev_1000]


if __name__ == "__main__":

    # Define Parameters
    DEM = arcpy.GetParameterAsText(0)
    D8FlowDir = arcpy.GetParameterAsText(1)
    StreamReach = arcpy.GetParameterAsText(2)
    ZFactor = arcpy.GetParameterAsText(3)
    OutAdjFlowDir = arcpy.GetParameterAsText(4)
    OutRelElev = arcpy.GetParameterAsText(5)
    OutWTD = arcpy.GetParameterAsText(6)
    
    # Set snap
    arcpy.env.snapRaster = DEM
    arcpy.env.extent = DEM
    rDesc = arcpy.Describe(DEM)
    arcpy.env.cellSize = rDesc.meanCellHeight
    arcpy.env.overwriteOutput = True
    arcpy.env.scratchWorkspace = env.scratchFolder
    
    # Determine cell size of input DEM
    cellsize = float(arcpy.GetRasterProperties_management(DEM, "CELLSIZEX").getOutput(0))

    # Modules
    RelativeElev = RelativeElevation(StreamReach, D8FlowDir, DEM)
    
    WaterTableDepth(RelativeElev, ZFactor)
